/*
* 	MPI reference code manually developed and optimized
*	for_each code reditributing the load
*	Allocate only memory needed for the local part 
*	Asynchronous sends and receives
*
* v2.1
*/


#include<stdio.h>
#include<math.h>
#include<stdlib.h>
#include<mpi.h>
#include <limits.h>

#define max(x,y)    ((x) > (y)? (x) : (y))
#define min(x,y)    ((x) < (y)? (x) : (y))

#define	mpiTestError(ok,cad)	\
    if ( ok != MPI_SUCCESS ) {	\
        fprintf(stderr,"RunTime-Error, Rank %d: %s - %d\n", mpiRank, cad, ok);\
        exit(-1);				\
    }

int rows; 
int mpiRank;
int mpiNProc;
double clock, clockReduce,clockSeq, clockSeq2, clockReduceSeq, clockCom, clockCom2, clockReduceCom;

static inline void function_find (int *exit, double par) {
  if (par > 499950000+0.75*rows/10){
       *exit = 1;
    }
  else {
      *exit = 0;
    }
}

/*
 * MAIN PROGRAM
 */
int main(int argc, char *argv[]) {
	int ok;
	int i,j;
#ifdef DEBUG
	setbuf(stderr, NULL);
	setbuf(stdout, NULL);
#endif

	/* 0. READ PARAMETERS */
	int  beginR,endR;

	if ( argc < 4 ) {
		fprintf(stderr, "\nUsage: %s <numRows> <begin> <end>  \n", argv[0]);
		exit(EXIT_FAILURE);
	}
	rows = atoi( argv[1] );
	beginR = atoi( argv[2] );
	endR = atoi( argv[3] );



    	/* 1. INITIALIZE MPI */
    	ok = MPI_Init( &argc, &argv);
	mpiTestError( ok, "Initializing MPI" );

   	 /* 2. GET BASIC GROUP PARAMETERS */
   	 MPI_Comm_rank(MPI_COMM_WORLD, &mpiRank);
   	 MPI_Comm_size(MPI_COMM_WORLD, &mpiNProc);

	/* 3. START CLOCK */
	MPI_Barrier( MPI_COMM_WORLD );
	clock = MPI_Wtime();

	/* 4. CARTESIAN TOPOLOGY */
	MPI_Comm cartesianComm;
	int procs = mpiNProc;

	/* 4.1. CHECK: NO SUPPORT FOR PARTITIONING LESS ELEMENTS THAN PROCS IN A DIMENSION */
	if ( procs > rows  ) {
		if ( mpiRank == 0 ) fprintf(stderr, "Error: This code does not support partitioning less rows/columns than processors in that axis\n");
		MPI_Finalize();
	}

	int periods=0;
	ok = MPI_Cart_create(MPI_COMM_WORLD,1 , &procs, &periods, 0, &cartesianComm);
	mpiTestError( ok, "Creating cartesian toology" );

	int myCoords;
	ok = MPI_Cart_coords(cartesianComm, mpiRank, 1, &myCoords);
	mpiTestError( ok, "Getting cartesian coordinates" );

	int rankUp=-1000, rankDown=-1000, rankLeft=-1000, rankRight=-1000;
	int foo;

	if ( myCoords != 0 ) {
		ok = MPI_Cart_shift(cartesianComm, 0, -1, &foo, &rankUp);
		mpiTestError( ok, "Shifting cartesian coordinates up" );
	}

	if ( myCoords != procs-1 ) {
		ok = MPI_Cart_shift(cartesianComm, 0, +1, &foo, &rankDown);
		mpiTestError( ok, "Shifting cartesian coordinates down" );
	}



	/* 5. COMPUTE THE LOCAL PART OF THE MATRIX */
	double procRatio;
	procRatio = (double)(rows)/procs;

	int begin;
	int end;

#ifdef DEBUG
fprintf(stderr,"CTRL %d, myCoords %d\n", mpiRank, myCoords);
#endif

	begin = (int)floor( myCoords*procRatio ); 
	end = (int)floor( (myCoords+((rows > procs) ? 1 : 0))*procRatio - ((rows > procs) ? 1 : 0) );
	
#ifdef DEBUG
fprintf(stderr,"CTRL %d, matrix[%d:%d]\n", mpiRank, begin, end);
#endif


#define	element(mat, idx1)		(mat[(idx1)])

   	// 6. DEFINE MAIN DATA STRUCTURES 
	int size =  end-begin+1;
	double *matrix;
	matrix = (double *)calloc( (size_t)(size), sizeof(double) );

	if ( matrix==NULL ) {
        fprintf(stderr,"RunTime-Error, Rank %d: Allocating data structures\n", mpiRank);
		exit(2);
	}

	// 7. INITIAL MATRIX VALUES 
	for (i=0; i<size; i++) {
		srand48(i+begin);
		matrix[i] = (double)((int)ceil(drand48()*(double)rows));
	}

	// Init clock
	clockSeq = MPI_Wtime();


	// 8. REDITRIBUTE THE DATA AMONG BEGINR AND ENDR IN A DIFFERENT ARRAY
	int rowsD= (endR-beginR+1);
	double procRatioNewD= (double)(rowsD)/procs;
	int beginD = beginR+ (int)floor( myCoords*procRatioNewD ); 
	int endD = beginR+(int)floor( (myCoords+((rowsD > procs) ? 1 : 0))*procRatioNewD - ((rowsD > procs) ? 1 : 0) );
	int sizeD=(endD-beginD+1);
	// ALLOCATE THE AUXILIAR MATRIX
	double *matrixD;
	matrixD = (double *)calloc( (size_t)(sizeD), sizeof(double) );
	if ( matrixD==NULL ) {
        fprintf(stderr,"RunTime-Error, Rank %d: Allocating data structures\n", mpiRank);
		exit(2);
	}
#ifdef DEBUG
fprintf(stderr,"CTRL %d, matrixD[%d:%d]\n", mpiRank, beginD, endD);
#endif

#ifdef DEBUG_1
fprintf(stderr,"%d- SENDS\n",mpiRank);
#endif

        // PERFORM THE COMMUNICATION 
	double clockComm = MPI_Wtime();

	
        MPI_Request *request ;
	MPI_Status *statusV;
	int numCom=0;
        request = (MPI_Request *)calloc( (size_t)(2*procs), sizeof(MPI_Request) );
        statusV = (MPI_Status *)calloc( (size_t)(2*procs), sizeof(MPI_Status) );
	for(i=0;i < 2*procs; i++){
		request[i]=MPI_REQUEST_NULL;
	}

	int p,beginpD,endpD;

        // TRAVERSE THE PROCESSES TO KNOW WHAT LOCAL HAS TO SEND
	for(p=0;p < procs; p++){
           // do not send to myself
           //if(p!=mpiRank){
		// CALCULATE THE BEGIND AND ENDD OF THE p PROCESS
		beginpD= beginR + (int)floor( p*procRatioNewD );
		endpD= beginR + (int)floor( (p+((rowsD > procs) ? 1 : 0))*procRatioNewD - ((rowsD > procs) ? 1 : 0) );

#ifdef DEBUG_1
fprintf(stderr,"CTRL %d-%d,matrix[%d:%d]  matrixpD[%d:%d]\n",mpiRank, p,begin,end, beginpD, endpD);
#endif
		//Perform communication if the redistributed range overlap the original range
		if(beginpD<=end && endpD>=begin ){

			//send from the maximum of begin an beginpD; and the minimum of end and endpD
			int initS= max(begin,beginpD);
			int countS= min(end,endpD) - initS +1;
			ok = MPI_Isend( &element(matrix, initS-begin), countS, MPI_DOUBLE, p, 0, MPI_COMM_WORLD, &request[p] );
                        mpiTestError( ok, "Send p" );
#ifdef DEBUG_1
			printf("\n %d Send %d elements to %d: from %d to %d\n",mpiRank,countS, p, initS,initS+countS-1);
#endif
		}
	   // }	
	}

 
#ifdef DEBUG_1
fprintf(stderr,"%d- RECEIVES\n", mpiRank);
#endif
        // TRAVERSE THE PROCESSES TO KNOW WHAT LOCAL HAS TO RECEIVE
        int beginp, endp;
	for(p=0;p < procs; p++){
           // do not receive from myself
         //  if(p!=mpiRank){
		// CALCULATE THE BEGIN AND END OF THE p PROCESS
		beginp=  (int)floor( p*procRatio );
		endp=  (int)floor( (p+((rows > procs) ? 1 : 0))*procRatio - ((rows > procs) ? 1 : 0) );
#ifdef DEBUG_1
fprintf(stderr,"CTRL %d-%d, matrixD[%d:%d] matrixp[%d:%d]\n", mpiRank, p, beginD,endD, beginp, endp);
#endif
		//Perform communication if the redistributed range overlap the original range
		if(beginD<=endp && endD>=beginp ){
			//receive from the maximum of  an ; and the minimum of  and 
			int initS= max(beginD,beginp);
			int countS= min(endD,endp) - initS +1;
 			ok = MPI_Irecv( &element(matrixD, 0), countS, MPI_DOUBLE, p, 0, MPI_COMM_WORLD, &request[procs+p] );
                        mpiTestError( ok, " Recv from p" );
#ifdef DEBUG_1
			printf("\n %d Receive %d elements from %d: from %d to %d\n",mpiRank,countS, p, initS,initS+countS-1);
#endif
		}
	   // }	
	}

	// 15. END Communications 
	MPI_Waitall(2*procs , request, statusV );
	
	double clockComm2 = MPI_Wtime()-clockComm;


#ifdef DEBUG
	printf("\n Antes computacion\n");
	for (i=0; i<sizeD; i++) {
		printf("%lf ", element(matrixD, i));	   
	}
#endif


	// 9. COMPUTE
	// For Each function
	int k;
	for (i=0; i<sizeD; i++) {
	    for(k=0; k<rows; k++){
		if(element(matrixD, i)>500){
		 element(matrixD, i) =element(matrixD, i)+ (k*k/100) ;
	       }
	   }
	}


#ifdef DEBUG
	printf("\n %d- Despues computacion\n",mpiRank);
	for (i=0; i<sizeD; i++) {
		printf("%d- %lf \n",mpiRank*sizeD+i, element(matrixD, i));	   
	}
#endif

        // 10. REDISTRIBUTE THE COMPUTED DATA TO THE ORIGINAL ARRAY

#ifdef DEBUG_2
fprintf(stderr,"SECOND REDISTRIBUTION \n \n");
#endif

	clockComm = MPI_Wtime();
	
	for(i=0;i < 2*procs; i++){
		request[i]=MPI_REQUEST_NULL;
	}

	for(p=0;p < procs; p++){
           // do not send to myself
          // if(p!=mpiRank){
		// CALCULATE THE BEGIND AND ENDD OF THE p PROCESS
		beginp=  (int)floor( p*procRatio );
		endp=  (int)floor( (p+((rows > procs) ? 1 : 0))*procRatio - ((rows > procs) ? 1 : 0) );
#ifdef DEBUG_2
fprintf(stderr,"2-CTRL %d, matrixp[%d:%d] matrixD[%d:%d]\n", mpiRank, beginp, endp, beginD,endD);
#endif
		//Perform communication if the redistributed range overlap the original range
		if(beginD<=endp && endD>=beginp ){
			request[p]=MPI_REQUEST_NULL;
			//send from the maximum of begin an beginpD; and the minimum of end and endpD
			int initS= max(beginD,beginp);
			int countS= min(endD,endp) - initS +1;
			ok = MPI_Isend( &element(matrixD, 0), countS, MPI_DOUBLE, p, 0, MPI_COMM_WORLD, &request[p] );
                        mpiTestError( ok, "2-Send p" );
			numCom++;
#ifdef DEBUG_2
			printf("\n 2- %d Send %d elements to %d: from %d to %d\n",mpiRank,countS, p, initS,initS+countS-1);
#endif
		}
	    //}	
	}



	for(p=0;p < procs; p++){
           // do not receive from myself
           //if(p!=mpiRank){
		// CALCULATE THE DISTRIBUTED BEGIN AND END OF THE p PROCESS
		beginpD= beginR + (int)floor( p*procRatioNewD );
		endpD= beginR + (int)floor( (p+((rowsD > procs) ? 1 : 0))*procRatioNewD - ((rowsD > procs) ? 1 : 0) );

#ifdef DEBUG_2
fprintf(stderr,"2R-CTRL %d, matrixpD[%d:%d] matrix[%d:%d]\n", mpiRank, beginpD, endpD,begin,end);
#endif
		//Perform communication if the redistributed range overlap the original range
		if(begin<=endpD && end>=beginpD ){
			//receive from the maximum of  an ; and the minimum of  and 
			int initS= max(begin,beginpD);
			int countS= min(end,endpD) - initS +1;
 			ok = MPI_Irecv( &element(matrix, initS-begin), countS, MPI_DOUBLE, p, 0, MPI_COMM_WORLD, &request[procs+p] );
                        mpiTestError( ok, "2R-Recv from p" );
#ifdef DEBUG_2
			printf("\n2R- %d Receive %d elements from %d: from %d to %d\n",mpiRank,countS, p, initS,initS+countS-1);
#endif
		}
	    //}	
	}


	// 15. END Communications 
	MPI_Waitall(2*procs , request, statusV );
	
	clockComm2 = MPI_Wtime()-clockComm+ clockComm2;
	

#ifdef DEBUG
	printf("\n %d- Despues redistribucion\n", mpiRank);
	for (i=0; i<size; i++) {
		printf("%d- %lf \n",mpiRank*size+i, element(matrix, i));	   
	}
#endif


// Obtain time
	clockSeq2 = MPI_Wtime()-clockSeq;
	double clockT=0, clockTComm;



// FIND
	double clockFind = MPI_Wtime();

	// 8. REDITRIBUTE THE DATA AMONG BEGINR AND ENDR IN A DIFFERENT ARRAY
	endR = endR +10;
	beginR = beginR -10; 
	rowsD= (endR-beginR+1);
	procRatioNewD= (double)(rowsD)/procs;
	beginD = beginR+ (int)floor( myCoords*procRatioNewD ); 
	endD = beginR+(int)floor( (myCoords+((rowsD > procs) ? 1 : 0))*procRatioNewD - ((rowsD > procs) ? 1 : 0) );
	sizeD=(endD-beginD+1);
	// ALLOCATE THE AUXILIAR MATRIX
	free(matrixD);
	matrixD = (double *)calloc( (size_t)(sizeD), sizeof(double) );
	if ( matrixD==NULL ) {
        fprintf(stderr,"RunTime-Error, Rank %d: Allocating data structures\n", mpiRank);
		exit(2);
	}
#ifdef DEBUG
fprintf(stderr,"CTRL %d, matrixD[%d:%d]\n", mpiRank, beginD, endD);
#endif

#ifdef DEBUG_1
fprintf(stderr,"%d- SENDS\n",mpiRank);
#endif

        // PERFORM THE COMMUNICATION 

	
	numCom=0;
        request = (MPI_Request *)calloc( (size_t)(2*procs), sizeof(MPI_Request) );
        statusV = (MPI_Status *)calloc( (size_t)(2*procs), sizeof(MPI_Status) );
	for(i=0;i < 2*procs; i++){
		request[i]=MPI_REQUEST_NULL;
	}

        // TRAVERSE THE PROCESSES TO KNOW WHAT LOCAL HAS TO SEND
	for(p=0;p < procs; p++){
           // do not send to myself
           //if(p!=mpiRank){
		// CALCULATE THE BEGIND AND ENDD OF THE p PROCESS
		beginpD= beginR + (int)floor( p*procRatioNewD );
		endpD= beginR + (int)floor( (p+((rowsD > procs) ? 1 : 0))*procRatioNewD - ((rowsD > procs) ? 1 : 0) );

#ifdef DEBUG_1
fprintf(stderr,"CTRL %d-%d,matrix[%d:%d]  matrixpD[%d:%d]\n",mpiRank, p,begin,end, beginpD, endpD);
#endif
		//Perform communication if the redistributed range overlap the original range
		if(beginpD<=end && endpD>=begin ){

			//send from the maximum of begin an beginpD; and the minimum of end and endpD
			int initS= max(begin,beginpD);
			int countS= min(end,endpD) - initS +1;
			ok = MPI_Isend( &element(matrix, initS-begin), countS, MPI_DOUBLE, p, 0, MPI_COMM_WORLD, &request[p] );
                        mpiTestError( ok, "Send p" );
#ifdef DEBUG_1
			printf("\n %d Send %d elements to %d: from %d to %d\n",mpiRank,countS, p, initS,initS+countS-1);
#endif
		}
	   // }	
	}

 
#ifdef DEBUG_1
fprintf(stderr,"%d- RECEIVES\n", mpiRank);
#endif
        // TRAVERSE THE PROCESSES TO KNOW WHAT LOCAL HAS TO RECEIVE
	for(p=0;p < procs; p++){
           // do not receive from myself
         //  if(p!=mpiRank){
		// CALCULATE THE BEGIN AND END OF THE p PROCESS
		beginp=  (int)floor( p*procRatio );
		endp=  (int)floor( (p+((rows > procs) ? 1 : 0))*procRatio - ((rows > procs) ? 1 : 0) );
#ifdef DEBUG_1
fprintf(stderr,"CTRL %d-%d, matrixD[%d:%d] matrixp[%d:%d]\n", mpiRank, p, beginD,endD, beginp, endp);
#endif
		//Perform communication if the redistributed range overlap the original range
		if(beginD<=endp && endD>=beginp ){
			//receive from the maximum of  an ; and the minimum of  and 
			int initS= max(beginD,beginp);
			int countS= min(endD,endp) - initS +1;
 			ok = MPI_Irecv( &element(matrixD, 0), countS, MPI_DOUBLE, p, 0, MPI_COMM_WORLD, &request[procs+p] );
                        mpiTestError( ok, " Recv from p" );
#ifdef DEBUG_1
			printf("\n %d Receive %d elements from %d: from %d to %d\n",mpiRank,countS, p, initS,initS+countS-1);
#endif
		}
	   // }	
	}

	// 15. END Communications 
	MPI_Waitall(2*procs , request, statusV );

	int Flag=0;
 	int index = INT_MAX;
	int indexTotal = INT_MAX;
	for (i=0; i<sizeD && Flag==0; i++) {
	   function_find(&Flag,matrixD[i]);
	   if(Flag==1){
		index= beginD -beginR+ i;
	    }
	}
	MPI_Allreduce( &index, &indexTotal, 1, MPI_INT, MPI_MIN, MPI_COMM_WORLD );


	// Obtaining Time
	double clockFind2 = MPI_Wtime()-clockFind;
	double clockFindT=0;
	MPI_Allreduce( &clockFind2, &clockFindT, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD );
	MPI_Allreduce( &clockSeq2, &clockT, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD );
	MPI_Allreduce( &clockComm2, &clockTComm, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD );



#ifdef RESULT
	// Accumulate Result
	double sum=0, sumTotal=0;
	for (i=0; i<size; i++) {
		sum= sum+element(matrix, i);
	   
	}
	MPI_Allreduce( &sum, &sumTotal, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD );
	if( mpiRank ==0){
		printf("\n Accumulated Sum: %lf \n SQRT: %lf\n", sumTotal, sqrt(sumTotal));
		printf("\n FIND : %d \n", indexTotal);
	}
#endif
	if( mpiRank ==0){
		printf("\nClock for_Each: %lf \n", clockT);
		printf("\nClock find: %lf\n", clockFindT);
		printf("Clock comm: %lf \n", clockTComm);
	}


    // 21. END PROGRAM 

	free( matrix );
//	if ( mpiRank == 0 ) free( matrixResult );
	MPI_Comm_free( &cartesianComm );
	MPI_Finalize();

	/* 22. RETURN CORRECT */
	return 0;
}
