/*
* Hitmap for each and find_if using the operators of the HLPP'16 conference
* 
*
* @author Ana Moretón Fernández
* @version v0.1
*
* (c) 2017, Grupo Trasgo, Universidad de Valladolid
*/

#include <hitmap.h>
#include <limits.h>
#include "../operadores.h"

#define swap(a,b)	{ double c=a; a=b; b=c; }

HitClock        mainClock;
HitClock        findClock;
int tam; 

/*******************************************************
 ********** Example: for_each STL library **************
 ******************************************************/

typedef void (*compare_cb)(double *exit);
compare_cb _TT_function_for_each;

int for_each (HitTile_double m, int begin, int end)
{

  int j;
#ifdef DEBUG
  printf("\n Begin:%d End:%d CArd M: %d\n", begin, end,hit_tileDimCard(m,0));
  for(j=0;j<hit_tileDimCard(m,0);j++){
	printf(" %lf", hit_tileElemAt( m, 1,j));
  }
#endif

  HitTile_double new = HIT_TILE_NULL_STATIC_TRASGO;

  //  ARRAY REDISTRIBUTION 
  {
    HitShape range1 = hit_shape (1, hit_sig (begin, end, 1)); 
    HitShape range2 = hit_shape (1, hit_sig (0, end-begin, 1)); 
    ArrayRemapRange(m ,new, range1, range2 , double, 1);
  }

#ifdef DEBUG
  printf("\n BEFORE FOR EACH\n");
  for(j=0;j<hit_tileDimCard(new,0);j++){
	printf(" %lf", hit_tileElemAt( new, 1,j));
  }
#endif

  for(j=0;j<hit_tileDimCard(new,0);j++){
	_TT_function_for_each(&hit_tileElemAt( new, 1,j));
  }
#ifdef DEBUG
  printf("\n NEW FOR EACH\n");
  for(j=0;j<hit_tileDimCard(new,0);j++){
	printf(" %lf", hit_tileElemAt( new, 1,j));
  }
#endif

  //  ARRAY REDISTRIBUTION 
  {
    HitShape range1 = hit_shape (1, hit_sig (begin, end, 1)); 
    HitShape range2 = hit_shape (1, hit_sig (0, end-begin, 1)); 
    ArrayRemapRange(new,m, range2, range1 , double, 0);
  }

  hit_tileFree(new);

  return 0;
}


/*******************************************************
 ********** Example: find_if STL library **************
 ******************************************************/
typedef void (*compare_find)(int *exit, double par);
compare_find function_find_if;



int find_if (int *totalFlag, HitTile_double matrix, int begin, int end)
{

  HitTile_double new = HIT_TILE_NULL_STATIC_TRASGO;

  int index = INT_MAX;
  //  ARRAY REDISTRIBUTION 
  {
    HitShape range1 = hit_shape (1, hit_sig (begin, end, 1)); 
    HitShape range2 = hit_shape (1, hit_sig (0, end-begin, 1)); 
    ArrayRemapRange(matrix,new, range1, range2 , double, 1);
  }

  int i, Flag=0;
  int n=hit_tileDimCard(new,0)-1;
  for (i=0; i<=n && Flag==0; i++) {
   double tmp = hit_tileElemAtNoStride(new,1,i); 
   function_find_if(&Flag,tmp);
   if(Flag==1){
	index=hit_tileDimBegin( new, 0 ) + i;
    }
  }
  HitTile FlagTile = hit_tileSingle (index, int);
  HitTile totalFlagTile = hit_tileSingle (*totalFlag, int);

  hit_comDoOnce (hit_comReduce (new.Layout,
				HIT_RANKS_NULL,
				&FlagTile,
				&totalFlagTile, HIT_INT, HIT_OP_MIN_INT));
  hit_tileFree(new);
  return 0;
}

/*******************************************************/
/*           Main program - For each and Find          */
/*******************************************************/
static inline void function_FE (double *exit)
{

  int k=0;
  for(k=0; k<tam; k++){
      if((*exit )>500){
	  (*exit ) = (*exit ) + (k*k/100) ;
      }
   }
}

static inline void function_find (int *exit, double par) {
  *exit = 0;
  if (par > 499950000+0.75*tam/10){
       *exit = 1;
    }
}

int main (int argc, char *argv[])
{
#ifdef DEBUG
setbuf(stdout, NULL);
setbuf(stderr, NULL);
#endif

  	hit_comInit (&argc, &argv);

	if ( argc < 4 ) {
		fprintf(stderr, "\nUsage: %s <numRows> <begin> <end>  \n", argv[0]);
		exit(EXIT_FAILURE);
	}
	/* Tam */
 	tam = atoi( argv[1] );
	int beginR = atoi( argv[2] );
	int endR = atoi( argv[3] );

    	/* 1. CREATE VIRTUAL TOPOLOGY */
    	HitTopology topo = hit_topology( plug_topPlain );

        /* 2. DECLARE FULL MATRIX WITHOUT MEMORY */
        HitTile_double mat;
        hit_tileDomain( &mat, double, 1, tam );

    	/* 3. COMPUTE PARTITION */
        HitShape parallelShape = hit_tileShape( mat );
        HitLayout matLayout = hit_layout( plug_layBlocks, topo, parallelShape );
	HitShape parShape = hit_layShape(matLayout);

	/* DATA STRUCTURES */
	HitTile_double m;
	hit_tileSelect( &m, &mat, parShape );		
	hit_tileAlloc( &m);
	//hit_tileDomainShapeAlloc(&m, double, parShape); 
	(m).Layout = matLayout;

	/* 4. INIT VECTOR*/
	int x ;
	hit_tileForDimDomain(m, 0, x){
			srand48(x+hit_layDimBegin (matLayout, 0));
			double aux=(double)((int)ceil(drand48()*(double)tam));
			hit_tileElemAt(m, 1, x ) = aux; 		
	}

#ifdef DEBUG
	{
	int j;
	printf("\n INIT for each\n");
	for(j=0;j<hit_tileDimCard(m,0);j++){
		printf("%lf ",hit_tileElemAt(m, 1, j ) );
	}
	}
#endif

        hit_clockSynchronizeAll();
        hit_clockStart( mainClock );

	/* 5. FOR EACH */
	_TT_function_for_each=function_FE;
	for_each (m, beginR, endR);
 
	hit_clockStop( mainClock );

        hit_clockSynchronizeAll();
        hit_clockStart( findClock );

	function_find_if = function_find;
	int totalFlag=0;
     	find_if (&totalFlag, m, beginR - 10, endR+10);

	hit_clockStop( findClock );

	{
	int j; double sum=0, sumTotal=0;
	for(j=0;j<hit_tileDimCard(m,0);j++){
		sum= sum+hit_tileElemAt(m, 1, j );
	}
	MPI_Allreduce( &sum, &sumTotal, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD );
	if( hit_Rank ==0){
		printf("\n Accumulated Sum: %lf \n SQRT: %lf\n", sumTotal, sqrt(sumTotal));
	}
	}
#ifdef DEBUG
	{
	int j;
	printf("\n END for each, find %d\n", totalFlag);
	for(j=0;j<hit_tileDimCard(m,0);j++){
		printf("%lf ",hit_tileElemAt(m, 1, j ) );
	}
	}
#endif
	hit_clockReduce( matLayout, findClock );
	hit_clockReduce( matLayout, mainClock );
	if (hit_Rank==0){
			printf("\n FIND : %d \n", totalFlag);
			printf("\nClock for_Each: %lf\n", mainClock.max );
			printf("\nClock find: %lf\n", findClock.max );
	}


	/* 6. FREE */
  	hit_layFree (matLayout);
  	hit_topFree (topo);
  	hit_tileFree (m);

  	hit_comFinalize ();
  	return 0;

}


