/*
* MPI QuickSort 
* Code based on the program of the 
* 
*
* @author Ana Moretón Fernández
* @version v0.1
*
* (c) 2017, Grupo Trasgo, Universidad de Valladolid
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
//#include <sunmath.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <time.h>
#include <mpi.h>
#include "hqs.h"

#define	mpiTestError(ok,cad)	\
    if ( ok != MPI_SUCCESS ) {	\
        fprintf(stderr,"RunTime-Error, Rank %d: %s - %d\n", mpiRank, cad, ok);\
        exit(-1);				\
    }

#define ceild(n,d)  ceil(((double)(n))/((double)(d)))
#define floord(n,d) floor(((double)(n))/((double)(d)))
#define max(x,y)    ((x) > (y)? (x) : (y))
#define min(x,y)    ((x) < (y)? (x) : (y))


/*******************************************************************/
/* --- merge --- merge and sort two sorted arrays into third array */
/*******************************************************************/
void merge(double a[ ], double b[ ], double c[ ], int m, int n) {
   int  i,      /* Loop control for array a */
        j,      /* Loop control for array b */
        k;      /* Loop control for array c */

   i = 0; 
   j = 0; 
   k = 0;

   while ( ( i < m ) && ( j < n ) ) {
      if ( a[ i ] < b[ j ] ) {
         c[ k ] = a[ i ];
         k++; 
         i++;
      } else {
         c[ k ] = b[ j ];
         k++; 
         j++;
      };
   }


   /* Pick up any remainder */
   while ( i < m ) {
      c[ k ] = a[ i ];
      k++; 
      i++;
   };
   while ( j < n ) {
      c[ k ] = b[ j ];
      k++; 
      j++;
   };

}


/********************************************************/
/*               Swap: swap two values                  */
/********************************************************/
void Swap(double *i, double *j) {
double 	t;		/* Temporary variable for swap */

   t = *i;
   *i = *j;
   *j = t;
}


/*****************************************************************************/
/* Median3: Find the median of left, center, and right; return the median    */
/* as the pivot, and "hide" the pivot as the next-to-last (right-1) element. */
/*****************************************************************************/
void Median3 (double a[], int left, int right, double *pivot) {
int center;   	/* Temporary variable for determining median */

   center = ( left + right ) / 2;

   if ( a[ left ] > a[ center ] ) {
       Swap ( &a[ left ], &a[ center ] );
   }
   if ( a[ left ] > a[ right ] ) {
       Swap ( &a[ left ], &a[ right ] );
   }
   if ( a[ center ] > a[ right ] ) {
       Swap ( &a[ center ], &a[ right ] );
   }

   *pivot = a[ center ];
   Swap ( &a[ center ], &a[ right - 1 ] );
}


/**********************************************************/
/*      QuickSort: the basic QuickSort algorithm.         */
/*      NOTE: Requires more than 2 items to sort properly */
/**********************************************************/
void QuickSort(double a[], int left, int right) {
   int 
	i, 		/* Loop control variable         */
	j; 		/* Loop control variable         */
double	pivot;		/* Quicksort pivot 	        */

   if ( left < right ) {
      Median3 (a, left, right, &pivot); /* Find the pivot. */

      /* Do the partition. */
      i = left;
      j = right - 1;
      if ( i < j ) {
        do {
           do {
              i++;
           } while ( a[ i ] < pivot );
           do {
              j-- ;
           } while ( a[ j ] > pivot );
           Swap( &a[ i ], &a[ j ] );
        } while ( j > i );

          Swap( &a[ i ], &a[ j ] );  		/* Undo last swap. */
          Swap( &a[ i ], &a[ right - 1 ]);  	/* Restore pivot. */
        };

      QuickSort( a, left, i - 1 );
      QuickSort( a, i + 1, right );
   }
}

/*********************************************************************/
/* -----------------------     InitData    ---------------------     */
/*********************************************************************/
void InitData(int numberToSort, double items[]){


   int i;
   double aux;
   for ( i = 0; i < numberToSort; i++ ) {
      //items[i] = (double)(numberToSort-i)/10;
	srand48(i);
	aux=(double)((int)ceil(drand48()*(double)numberToSort));
	//printf("%lf ", aux);
	items[i]=aux;
   }

#ifdef DEBUG
  {
   printf( "Data generation complete.  Sorting begins.\n\n" );
   fflush( stdout );

   printf("\n data generated\n");
   int j;
   for(j=0;j<numberToSort;j++){

	printf("%lf ",items[j] );
   }
  }
#endif

}
/*********************************************************************/
/* --- SendData--- send subarray and length to each node             */
/*********************************************************************/
void SendData(int numberToSort, int numberOfNodes, int *length0, double items[]) {
   int 	
	i,			/* Loop control variable                  */
 	extras,			/* Number of processors that get one more */
	length,			/* Lengths of subarrays for each node     */
	length1;		/* Lengths of subarrays for each node     */

  /* Distribute array data to nodes */
   /* Load balance, at least initially */
   length = numberToSort / numberOfNodes;
   extras = numberToSort - length * numberOfNodes;
   length1 = length + 1;

   /* Send initial data to be sorted to each process */
   for ( i = 1; i < numberOfNodes; i++ ) { 
      if ( i >= extras ) {
         MPI_Send( &items[ length * i + extras ], length, MPI_DOUBLE, 
                   i, DATA, MPI_COMM_WORLD );
      } else {
         MPI_Send( &items[ length * i + ( i < extras ) * i ], 
                   length1, MPI_DOUBLE, i, DATA, MPI_COMM_WORLD );
      }
   }     


   /* Length for processor 0 */
   if ( extras )
      *length0 = length1;
   else
      *length0 = length;
}


/*********************************************************/
/*                      HQSnode                          */
/*********************************************************/
int HQSnode( double *items, long int numberToSort ) {
   double
        *half,			/* Subarray for merging            	  */
        *lower,			/* Array for subdividing problem   	  */
        
	localMedian, 		/* Median of all sorted substrings 	  */
	globalMedian, 		/* Median of all sorted substrings 	  */
        medians[ MAX_NODES ],	/* Array of medians		   	  */
        *upper;			/* Array for subdividing problem   	  */

   double  
	startTime, 
        endTime;		/* variables to handle MPI timer results  */

   int	dimension,		/* Dimension of hypercube          	  */
        error=0, 
        first,			/* Number of first node in subcube 	  */
        halfLength,		/* Length of subarray for merging   	  */
        i,			/* Loop control variable	   	  */
        j,			/* Loop control variable	   	  */
        k,			/* Loop control variable	   	  */
        last,			/* Number of last node in subcube 	  */
	length,			/* Length of this node's list		  */
        lowerLength,		/* Length of subdivision	   	  */
        myCubeNumber,		/* Current subcube number		  */
        myNodeNumber,		/* My node number 			  */
        myRank,                	/* My rank in current subcube             */


        numberOfNodes,		/* Number of nodes in hypercube		  */
        numberOfProcessors,	/* Number of processors			  */
        numberOfCubes,		/* Number of cubes in subdivision of 
				   hypercube	  			  */
        partner,		/* Number of node to exchange with in this 
				   subdivision of hypercube		  */
        cubeSize,		/* Number of nodes in subcube		  */
        upperLength;		/* Length of subdivision	   	  */

   MPI_Status
	Status;			/* Status returned from MPI_Recv function */

   MPI_Comm
	subcubeCommunicator;	/* The following variable is needed to do 
				   subgroup communication  		  */
	

   struct rusage;		/* Structure to access time		  */

    MPI_Comm_rank( MPI_COMM_WORLD,&myNodeNumber );
    MPI_Comm_size( MPI_COMM_WORLD,&numberOfProcessors );

#ifdef DEBUG
    printf( "Hello from process %d out of %d\n", 
            myNodeNumber, numberOfProcessors );
    fflush( stdout );
#endif

    /* Initialize variables */
    dimension = floor( log2( (double) numberOfProcessors ) );
    numberOfNodes = 1 << dimension;
    if( (upper = malloc( sizeof(double) * ( numberToSort/2 + 1 ) ) ) == 0 ||
        (half  = malloc( sizeof(double) * ( numberToSort/2 + 1 ) ) ) == 0 ||
        (lower = malloc( sizeof(double) * ( numberToSort/2 + 1 ) ) ) == 0 ){

        if ( myNodeNumber == HOST ) {
            printf(stderr,"Unable to allocate enough memory, exiting...\n\n");
	}
     };

    /* Generate numbers to sort and split the work among the processors */
    if ( myNodeNumber == HOST ) {

       /* SendData will return length and array for the HOST node */
       SendData( numberToSort, numberOfNodes, &length, items );

    } else if ( myNodeNumber < numberOfNodes ) {

       /* If this processor is actually participating */
       MPI_Recv( items, MAX_ARRAY, MPI_DOUBLE, HOST, DATA, MPI_COMM_WORLD, 
 	         &Status );
       MPI_Get_count( &Status, MPI_DOUBLE, &length );

    } else {

       /* If this processor is NOT participating */
       length = 0;

    }
    
#ifdef DEBUG
	{
	printf("\n INCIO\n");
	for(j=0;j<length;j++){

		printf("%lf ",items[j] );
	}

	}
#endif

    /*  NOTE: Timing is system dependent */
    /* Start timing */
    MPI_Barrier( MPI_COMM_WORLD );
    startTime = MPI_Wtime( );

    /* Do only if this processor is participating */
    if ( myNodeNumber < numberOfNodes ) {

        /* 2) Sort using quicksort */
        QuickSort( items, 0, length-1 );

    }

    for ( i = dimension; i > 0; i-- ) {

        /* Do only if this processor is participating */
        if ( myNodeNumber < numberOfNodes ) {

            /* 3) Broadcast median to rest of subcube */

            /* Determine which subcube this processor is in */
            numberOfCubes = 1 << ( dimension - i );
            cubeSize = numberOfNodes / numberOfCubes;
            myCubeNumber = myNodeNumber >> i;

            /* Calculate median */
            if ( length % 2 ) { 			/* if odd */
                localMedian = items[ length / 2 ];
            } else {					/* if even */
                localMedian = ( items[ ( length / 2 ) - 1 ] 
                              + items[ length / 2 ] ) / 2;
            };

            first = myCubeNumber * cubeSize;
            last  = first + cubeSize;
        }

        /* Set up communication subgroup */

        if ( myNodeNumber < numberOfNodes ) {

            /* Do only if this processor is participating */
            MPI_Comm_split( MPI_COMM_WORLD, myCubeNumber, myNodeNumber, 
	      	            &subcubeCommunicator );
            MPI_Comm_rank( subcubeCommunicator, &myRank );

        } else {

            /* Do if this processor is NOT participating  - Note use of 
               MPI_UNDEFINED */
            MPI_Comm_split( MPI_COMM_WORLD, MPI_UNDEFINED, myNodeNumber, 
		            &subcubeCommunicator );
        }

        /* Do only if this processor is participating */
        if ( myNodeNumber < numberOfNodes ) {

            /* Broadcast median to each node in subcube (except me) */
            medians[ myRank ] = localMedian;
            for ( j = 0; j < cubeSize; j++ ) {
                MPI_Bcast( &medians[j], 1, MPI_INT, j, subcubeCommunicator );
            }


            /* Determine median of medians */
            QuickSort( medians, 0, cubeSize-1 );

            /* There are always an even number of medians */
            globalMedian = ( medians[ ( cubeSize / 2 ) - 1 ] 
                           + medians[ cubeSize / 2 ] ) / 2;


            /* 4) Create lower and upper 'halves' of items */
            j = 0; 
            k = 0;
            while ( ( items[ j ] <= globalMedian ) && ( j < length ) ) {
               lower[ k ] = items[ j ];
               k++; 
               j++;
            };
            lowerLength = k; 
            k = 0;
            while ( j < length ) {
               upper[ k ] = items[ j ];
               k++; 
               j++;
            };
            upperLength = k;

            /* 5) Send lower/upper 'half' to partner */

            /* Determine partner by inverting bits of node number */
            partner = ( numberOfNodes - 1 ) - myNodeNumber;
            partner = ( myNodeNumber ^ ( ~( ~0 << 1 ) << ( i-1 ) ) );

            /* Exchange info with partner */
            if ( myNodeNumber > partner ) {

               /* Send lower 'half' to partner */
               MPI_Send( lower, lowerLength, MPI_DOUBLE, partner, HALF, 
                         MPI_COMM_WORLD );

               /* Receive upper 'half' from partner */
               MPI_Recv( half, MAX_ARRAY, MPI_DOUBLE, partner, HALF, 
                         MPI_COMM_WORLD, 
	      	        &Status );
               MPI_Get_count( &Status, MPI_DOUBLE, &halfLength );

            } else {

               /* Receive lower 'half' from partner */
               MPI_Recv( half, MAX_ARRAY, MPI_DOUBLE, partner, HALF, 
                         MPI_COMM_WORLD, 
		        &Status );
               MPI_Get_count( &Status, MPI_DOUBLE, &halfLength );

               /* Send upper 'half' to partner */
               MPI_Send( upper, upperLength, MPI_DOUBLE, partner, HALF, 
                         MPI_COMM_WORLD );
            };


            /* 6) Merge lower/upper array with 'half' array into items array */
            if (myNodeNumber > partner) {
               merge( upper, half, items, upperLength, halfLength );
               length = upperLength + halfLength;
            } else {
               merge( lower, half, items, lowerLength, halfLength );
               length = lowerLength + halfLength;
            };
        };
   };

#ifdef DEBUG
	{
	printf("\n FINAL Quick\n");
	for(j=0;j<length;j++){

		printf("%lf ",items[j] );
	}

	}
#endif

   /*NOTE:  Timing is system dependent */
   /* End timing */
   MPI_Barrier( MPI_COMM_WORLD );
   endTime = MPI_Wtime( ) - startTime;
   double clockFinalT=0;
   MPI_Allreduce( &endTime, &clockFinalT, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD );
   if( myNodeNumber ==0){
     	printf("\nClock main: %lf\n", clockFinalT );
    }
   return 0;
}

/*******************************************************/
/*      Main program - starts up on each machine       */
/*******************************************************/
int main( int argc, char *argv[ ] ) {

    /*  1. INIT MPI */
    MPI_Init( &argc, &argv );

    /* 2. ALLOCATE DATA STRUCTURES AND GET PARAMETERS */
    int error=0; 
    if ( argc != 2 ) {
    /* Check command line arguments; "random" is not really necessary;
       it's just used to show a concept */
   
        printf( "Usage: %s  <number to sort>\n", argv[0] );
        fflush( stdout );
        
        error = 1;
    } 
 
    /* Abort if there's an error on the command line */
    if ( error ) {
        return (1);
    }

    long int numberToSort = atol( argv[1] );
    double *items;			/* Array of info to be sorted 	   	  */

    if ( numberToSort < 2 || numberToSort > MAX_ARRAY ) {
        printf( "The number to sort must be between 2 and %d\n", 
                   MAX_ARRAY );
        error = 1;
        /* Abort if there's an error on the command line */
        if ( error ) {
            return (1);
        }
    } else if( (items = malloc( sizeof(double) * numberToSort ) ) == 0 ){
            printf(stderr,"Unable to allocate enough memory, exiting...\n\n");
            error = 1;
            /* Abort if there's not enough memory */
            if ( error ) {
              return (1);
            }
    }

    /* 3. INIT DATA STRUCTURES */
    InitData(numberToSort,items);

    /* 4. QUICKSORT */
    error= HQSnode( items, numberToSort );

    int myNodeNumber; MPI_Comm_rank( MPI_COMM_WORLD,&myNodeNumber );
    {
     int j; double suma=0, sumaTotal=0;
     for(j=0;j<numberToSort;j++){
	suma = suma + items[j]*j;
	}
   
     MPI_Allreduce( &suma, &sumaTotal, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD );
     if( myNodeNumber ==0) printf("\n SUM: %lf \n SQRT: %lf\n", sumaTotal, sqrt(sumaTotal));
    }
    /* 5. END MPI */
    MPI_Finalize( );
    return 0; 
}
