/*
* Hitmap QuickSort using the operators of the HLPP'16 conference
* 
*
* @author Ana Moretón Fernández
* @version v0.1
*
* (c) 2017, Grupo Trasgo, Universidad de Valladolid
*/
#include <hitmap.h>
#include "../operadores.h"

#define swap(a,b)	{ double c=a; a=b; b=c; }

HitClock        mainClock;


/* C. FUNCTION TO COMPUTE SEQUENTIAL PIVOTING */
int sequentialPivoting_seq( double pivot, HitTile_double data ) {

	int lo = 0;
	int hi = hit_tileCard( data )-1;
	double save = hit_tileElemAt1( data, lo );

	while( lo < hi ) { 
		/* SEARCH BACKWARD FOR LESSER THAN PIVOT */
		while ( hit_tileElemAt1( data, hi ) >= pivot && ( lo < hi )) { hi--; }
		if ( lo != hi ) 
		{
			hit_tileElemAt1( data, lo ) = hit_tileElemAt1( data, hi );
			lo++;
		}

		/* SEARCH FORWARD FOR GREATER THAN PIVOT */
		while ( hit_tileElemAt1( data, lo ) <= pivot && lo < hi ) { lo++; } 
		if (lo != hi)
		{
			hit_tileElemAt1( data, hi ) = hit_tileElemAt1( data, lo );
			hi--;
		}
	} 
	hit_tileElemAt1( data, lo ) = save;

	/* RETURN VALUE */
	if ( save <= pivot ) return lo + 1;
	else return lo;
}


/* D. COMPLETE SEQUENTIAL QUICKSORT */
int sequentialQuickSort_seq( HitTile_double data, int begin, int end ) {

	if ( begin >= end ) return 0;

	int lo = begin;
	int hi = end;

	double pivot = hit_tileElemAt1( data, begin );

	while( lo < hi ) { 
		/* SEARCH BACKWARD FOR LESSER THAN PIVOT */
		while ( hit_tileElemAt1( data, hi ) >= pivot && ( lo < hi )) { hi--; }
		if ( lo != hi ) 
		{
			hit_tileElemAt1( data, lo ) = hit_tileElemAt1( data, hi );
			lo++;
		}

		/* SEARCH FORWARD FOR GREATER THAN PIVOT */
		while ( hit_tileElemAt1( data, lo ) <= pivot && lo < hi ) { lo++; } 
		if (lo != hi)
		{
			hit_tileElemAt1( data, hi ) = hit_tileElemAt1( data, lo );
			hi--;
		}
	} 
	hit_tileElemAt1( data, lo ) = pivot;

	/* RECURSION */
	sequentialQuickSort_seq( data, begin, lo );
	sequentialQuickSort_seq( data, lo+1, end );

	/* RETURN VALUE */
	return lo;
}

static inline void
sequentialPivoting (double pivot, HitTile_double data, int
		    *leftCard, int *rightCard, HitTile_int mask)
{
  int i;
  *leftCard = sequentialPivoting_seq (pivot, data);
  *rightCard = hit_tileCard (data) - 1;
  for (i = 0; i < hit_tileDimCard (data, 0); i++)
    {
      if (i < *leftCard)
	{
	  hit_tileElemAt1 (mask, i) = 0;
	}
      else
	{
	  hit_tileElemAt1 (mask, i) = 1;
	}
    }
}

static inline void
stopParallel (HitTile_double data, int *sal)
{
  if (*sal != 1)
    {
      if (hit_layNumActives (data.Layout) == 1)
	{
	  sequentialQuickSort_seq (data, 0, hit_tileCard (data) - 1);
	  *sal = 1;
	}
    }
}

static inline void
stopRecursion (HitTile_double data, int *sal)
{
  *sal = 0;
  /* 1. STOP RECURSION */
  /* 1.a. TRIVIAL CASE: NON ACTIVE IN THE LAYOUT */
  if (!hit_layImActive (data.Layout))
    *sal = 1;
  /* 1.b. TRIVIAL CASE: ONE DATA ELEMENT IF THE WHOLE LAYOUT */
  if (hit_shapeCard (hit_layFullShape (data.Layout)) <= 1)
    *sal = 1;
  /* 1.c. TRIVIAL SEQUENTIAL CASES FOR ONLY ONE ACTIVE PROCESS */
  if (hit_layNumActives (data.Layout) == 1)
    {
      if (hit_tileCard (data) <= 1)
	*sal = 1;
      if (hit_tileCard (data) == 2)
	{
	  if (hit_tileGet (data, 1, 0) > hit_tileGet (data, 1, 1))
	    swap (hit_tileElemAt (data, 1, 0), hit_tileElemAt (data, 1, 1));
	  *sal = 1;
	}
    }
}



int
parallelQuickSort (HitTile_double * data)
{



  /** QUICK SORT */


  double pivote;

  int leftCard;
  int rightCard;
  int totalLeftCard;
  int totalRightCard;

  int salida = 0;

  HitTile_int mask = HIT_TILE_NULL_STATIC_TRASGO;

  stopRecursion (*data, &salida);

  stopParallel (*data, &salida);


  if (salida != 1)
    {

	///////////////////////////////////
	//ARRAY
	//////////////////////////////////

     	hit_tileDomainShapeAlloc (&(mask), int, hit_tileShape ((*data)));
      	(mask).Layout = (*data).Layout;

	pivote = hit_tileGet (*data, 1, 0);


     	HitTile NULLTile = HIT_TILE_NULL;
     	HitTile pivoteTile = hit_tileSingle (pivote, double);

     	hit_comDoOnce (hit_comReduce ((*data).Layout,
				    HIT_RANKS_NULL,
				    &NULLTile,
				    &pivoteTile,
				    HIT_DOUBLE, HIT_OP_SUM_DOUBLE));

     	pivote = pivote / hit_layNumActives ((*data).Layout);

	sequentialPivoting (pivote, *data, &leftCard, &rightCard, mask);


      	HitTile leftCardTile = hit_tileSingle (leftCard, int);
      	HitTile totalLeftCardTile = hit_tileSingle (totalLeftCard, int);

      	hit_comDoOnce (hit_comReduce ((*data).Layout,
				    HIT_RANKS_NULL,
				    &leftCardTile,
				    &totalLeftCardTile,
				    HIT_INT, HIT_OP_SUM_INT));
	totalRightCard= hit_shapeCard (hit_layFullShape ((*data).Layout)) - totalLeftCard;

	double weights[2] ={totalLeftCard, totalRightCard};
		
	HitLayout grouping; 
      	HitTile_double (*_TT_Newtiledata_3) = malloc (2 * sizeof (HitTile_double));	
        int (*_TT_indiceGlobal_data_3) = (int *) malloc (hit_tileDims (*data) * (2 + 1) * sizeof (int));	

	// . ARRAY PARTITION 
	ArrayDivide_dev(grouping, 2, weights, data, _TT_Newtiledata_3, mask, _TT_indiceGlobal_data_3, double );

		// . RECURSIVE CALL 
		if (hit_layHasElem (grouping, 0))
	      	  parallelQuickSort (&(_TT_Newtiledata_3[0]));
	    	
	  	if (hit_layHasElem (grouping, 1))
	    	  parallelQuickSort (&(_TT_Newtiledata_3[1]));

	// . MERGE 
	ArrayMerge_dev(grouping,2, data, _TT_Newtiledata_3, _TT_indiceGlobal_data_3, double);

	 hit_layFree (grouping);

	 hit_tileFree (mask);
    }

    return 0;
}

/*******************************************************/
/*               Main program - QuickSort              */
/*******************************************************/
int main (int argc, char *argv[])
{

  	hit_comInit (&argc, &argv);

	if ( argc != 2 ) {
                fprintf(stderr, "\nUsage: %s <numRows> \n", argv[0]);
                exit(EXIT_FAILURE);
        }

	/* numberToSort */
 	int numberToSort = atoi( argv[1] );

    	/* 1. CREATE VIRTUAL TOPOLOGY */
    	HitTopology topo = hit_topology( plug_topPlain );

        /* 2. DECLARE FULL MATRIX WITHOUT MEMORY */
        HitTile_double mat;
        hit_tileDomain( &mat, double, 1, numberToSort );

    	/* 3. COMPUTE PARTITION */
        HitShape parallelShape = hit_tileShape( mat );
        HitLayout matLayout = hit_layout( plug_layBlocks, topo, parallelShape );
	HitShape parShape = hit_layShape(matLayout);

	/* DATA STRUCTURES */
	HitTile_double m;
	hit_tileDomainShapeAlloc(&m, double, parShape); 
	(m).Layout = matLayout;

	/* 4. INIT VECTOR*/
	int x ;
	hit_tileForDimDomain(m, 0, x){
			srand48(x+hit_layDimBegin (matLayout, 0));
			double aux=(double)((int)ceil(drand48()*(double)numberToSort));
			hit_tileElemAt(m, 1, x ) = aux; 		
	}

#ifdef DEBUG
	{
	int j;
	printf("\n INIT Quick\n");
	for(j=0;j<hit_tileDimCard(m,0);j++){
		printf("%lf ",hit_tileElemAt(m, 1, j ) );
	}
	}
#endif

        hit_clockSynchronizeAll();
        hit_clockStart( mainClock );

	/* 5. QUICKSORT */
  	parallelQuickSort (&m);

	hit_clockStop( mainClock );
	hit_clockReduce( matLayout, mainClock );

	{
	  int j; double suma=0, sumaTotal=0;
	  for(j=0;j<hit_tileDimCard(m,0);j++){
	  suma = suma + hit_tileElemAt(m, 1, j )*(hit_layDimBegin (matLayout, 0)+j);
	}   
	MPI_Allreduce( &suma, &sumaTotal, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD );
	if (hit_Rank==0) printf("\n SUM: %lf \n SQRT: %lf\n", sumaTotal, sqrt(sumaTotal));
	}

#ifdef DEBUG
	{
	int j;
	printf("\n END Quick\n");
	for(j=0;j<hit_tileDimCard(m,0);j++){
		printf("%lf ",hit_tileElemAt(m, 1, j ) );
	}
	}
#endif

	if (hit_Rank==0){
			printf("\nClock main: %lf\n", mainClock.max );
	}


	/* 6. FREE */
  	hit_layFree (matLayout);
  	hit_topFree (topo);
  	hit_tileFree (m);

  	hit_comFinalize ();
  	return 0;

}
