/*
* MPI QuickSort 
* Code based on the program of the 
* 
*
* @author Ana Moretón Fernández
* @version v0.1
*
* (c) 2017, Grupo Trasgo, Universidad de Valladolid
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
//#include <sunmath.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <time.h>
#include <mpi.h>
#include "hqs.h"

#define	mpiTestError(ok,cad)	\
    if ( ok != MPI_SUCCESS ) {	\
        fprintf(stderr,"RunTime-Error, Rank %d: %s - %d\n", mpiRank, cad, ok);\
        exit(-1);				\
    }

#define ceild(n,d)  ceil(((double)(n))/((double)(d)))
#define floord(n,d) floor(((double)(n))/((double)(d)))
#define max(x,y)    ((x) > (y)? (x) : (y))
#define min(x,y)    ((x) < (y)? (x) : (y))


/*********************************************************************/
/* -----------------------     InitData    ---------------------     */
/*********************************************************************/
void InitData(int numberToSort, double items[]){

   int i;
   double aux;
   for ( i = 0; i < numberToSort; i++ ) {
	srand48(i);
	aux=(double)((int)ceil(drand48()*(double)numberToSort));
	if(i < 5){ aux=2 ; }	
	items[i]=aux;
   }

#ifdef DEBUG
  {
   printf( "Data generation complete.  Sorting begins.\n\n" );
   fflush( stdout );

   printf("\n data generated\n");
   int j;
   for(j=0;j<numberToSort;j++){

	printf("%lf ",items[j] );
   }
  }
#endif

}

void SendData(int numberToSort, int numberOfNodes, int *length0, double items[]) {
   int 	
	i,			/* Loop control variable                  */
 	extras,			/* Number of processors that get one more */
	length,			/* Lengths of subarrays for each node     */
	length1;		/* Lengths of subarrays for each node     */

  /* Distribute array data to nodes */
   /* Load balance, at least initially */
   length = numberToSort / numberOfNodes;
   extras = numberToSort - length * numberOfNodes;
   length1 = length + 1;

   /* Send initial data to be sorted to each process */
   for ( i = 1; i < numberOfNodes; i++ ) { 
      if ( i >= extras ) {
         MPI_Send( &items[ length * i + extras ], length, MPI_DOUBLE, 
                   i, DATA, MPI_COMM_WORLD );
      } else {
         MPI_Send( &items[ length * i + ( i < extras ) * i ], 
                   length1, MPI_DOUBLE, i, DATA, MPI_COMM_WORLD );
      }
   }     


   /* Length for processor 0 */
   if ( extras )
      *length0 = length1;
   else
      *length0 = length;
}


/** 
   Unique Copy: copy in an output vector the non-consecutive neither duplicated data.
   Parameters: Input vector, how is distributed the input vector,  lenght in local process for the output vector, total number of elements 
   Return: output vector
*/
double * unique_copy(double *m, int *lenght_in,int *lenght_out,  int N_1){

	MPI_Comm cartesianComm;
	int mpiRank;
	int mpiNProc;
   	MPI_Comm_rank(MPI_COMM_WORLD, &mpiRank);
   	MPI_Comm_size(MPI_COMM_WORLD, &mpiNProc);

	int procs = mpiNProc;
	int ok;
	int i,j;
	int beginR=0;
	int endR= N_1-1;
	int begin2=0;

	int periods=0;
	ok = MPI_Cart_create(MPI_COMM_WORLD,1 , &procs, &periods, 0, &cartesianComm);
	mpiTestError( ok, "Creating cartesian toology" );

	int myCoords;
	ok = MPI_Cart_coords(cartesianComm, mpiRank, 1, &myCoords);
	mpiTestError( ok, "Getting cartesian coordinates" );

	int rankUp=-1000, rankDown=-1000, rankLeft=-1000, rankRight=-1000;
	int foo;

	if ( myCoords != 0 ) {
		ok = MPI_Cart_shift(cartesianComm, 0, -1, &foo, &rankUp);
		mpiTestError( ok, "Shifting cartesian coordinates up" );
	}

	if ( myCoords != procs-1 ) {
		ok = MPI_Cart_shift(cartesianComm, 0, +1, &foo, &rankDown);
		mpiTestError( ok, "Shifting cartesian coordinates down" );
	}


	/* 5. COMPUTE THE LOCAL PART OF THE MATRIX */
	double procRatio;
	procRatio = (double)(N_1)/procs;

	int beginD;
	int endD;

#ifdef UNIQUE
fprintf(stderr,"CTRL %d, myCoords %d\n", mpiRank, myCoords);
#endif

	beginD = (int)floor( myCoords*procRatio ); 
	endD = (int)floor( (myCoords+((N_1 > procs) ? 1 : 0))*procRatio - ((N_1 > procs) ? 1 : 0) );
	
#ifdef UNIQUE
fprintf(stderr,"CTRL %d, matrix[%d:%d]\n", mpiRank, beginD, endD);
#endif


#define	element(mat, idx1)		(mat[(idx1)])

        // 6. DEFINE MAIN DATA STRUCTURES 

	// Reservo un dato mas para traer el valor del proceso de la izquierda
	int size =  endD-beginD+1;
	double *matrix, *matrix_aux;
 	double *matrix2;
	matrix = (double *)calloc( (size_t)(size+1), sizeof(double) );
	matrix_aux = (double *)calloc( (size_t)(size), sizeof(double) );
	matrix2 = (double *)calloc( (size_t)(size), sizeof(double) );


	if ( matrix==NULL || matrix2==NULL || matrix_aux==NULL) {
        fprintf(stderr,"RunTime-Error, Rank %d: Allocating data structures\n", mpiRank);
		exit(2);
	}


	// 7. INITIAL MATRIX VALUES 
	for (i=0; i<size; i++) {
	   element(matrix_aux, i) = 0;
	   element(matrix2, i) = 1.5;
	   element(matrix, i) = -2.25;
	}

#ifdef UNIQUE
	printf("\n Init\n");
	for (i=0; i<lenght_in[mpiRank]; i++) {
		printf("%lf ", element(m, i));	   
	}
#endif



	// 1-Aº Redistribuyo a un array nuevo con los datos entre begin y end
	
        MPI_Request *requestI ;
	MPI_Status *statusVI;
	int numCom=0;
        requestI = (MPI_Request *)calloc( (size_t)(2*procs), sizeof(MPI_Request) );
        statusVI = (MPI_Status *)calloc( (size_t)(2*procs), sizeof(MPI_Status) );

	if ( requestI==NULL || statusVI ==NULL ) {
        fprintf(stderr,"RunTime-Error, Rank %d: Allocating request\n", mpiRank);
		exit(2);
	}
	for(i=0;i < 2*procs; i++){
		requestI[i]=MPI_REQUEST_NULL;
	}


	int p,beginpD,endpD;
	 int begin=0;
	//BEgin y end iniciales
	for(i=0;i<mpiRank;i++){
		 begin=begin+lenght_in[i];
	}

	 int end=begin+lenght_in[mpiRank]-1;


        // TRAVERSE THE PROCESSES TO KNOW WHAT LOCAL HAS TO SEND
	for(p=0;p < procs; p++){
           // do not send to myself
           //if(p!=mpiRank){
		// CALCULATE THE BEGIND AND ENDD OF THE p PROCESS
		beginpD= beginR + (int)floor( p*procRatio );
		endpD= beginR + (int)floor( (p+((N_1 > procs) ? 1 : 0))*procRatio - ((N_1 > procs) ? 1 : 0) );

#ifdef UNIQUE
fprintf(stderr,"\nCTRL %d-%d,matrix[%d:%d]  matrixpD[%d:%d]\n",mpiRank, p,begin,end, beginpD, endpD);
#endif
		//Perform communication if the redistributed range overlap the original range
		if(beginpD<=end && endpD>=begin ){

			//send from the maximum of begin an beginpD; and the minimum of end and endpD
			int initS= max(begin,beginpD);
			int countS= min(end,endpD) - initS +1;
			ok = MPI_Isend( &element(m, initS-begin), countS, MPI_DOUBLE, p, 0, MPI_COMM_WORLD, &requestI[p] );
                        mpiTestError( ok, "Send p" );
#ifdef UNIQUE
			printf("\n %d Send %d elements to %d: from %d to %d, %d\n",mpiRank,countS, p, initS,initS+countS-1, initS-begin);
#endif
		}
	   // }	
	}

 
#ifdef UNIQUE
fprintf(stderr,"%d- RECEIVES\n", mpiRank);
#endif
        // TRAVERSE THE PROCESSES TO KNOW WHAT LOCAL HAS TO RECEIVE
        int beginp, endp;
	for(p=0;p < procs; p++){
           // do not receive from myself
         //  if(p!=mpiRank){
		// CALCULATE THE BEGIN AND END OF THE p PROCESS
		beginp= 0; //(int)floor( p*procRatio );
		for(i=0;i<p;i++){
			 beginp=beginp+lenght_in[i];
		}
		endp= beginp+lenght_in[p]-1; // (int)floor( (p+((rows > procs) ? 1 : 0))*procRatio - ((rows > procs) ? 1 : 0) );
#ifdef UNIQUE
fprintf(stderr,"\nCTRL %d-%d, matrixD[%d:%d] matrixp[%d:%d]\n", mpiRank, p, beginD,endD, beginp, endp);
#endif
		//Perform communication if the redistributed range overlap the original range
		if(beginD<=endp && endD>=beginp ){
			//receive from the maximum of  an ; and the minimum of  and 
			int initS= max(beginD,beginp);
			int countS= min(endD,endp) - initS +1;
 			ok = MPI_Irecv( &element(matrix, initS-beginD+1), countS, MPI_DOUBLE, p, 0, MPI_COMM_WORLD, &requestI[procs+p] );
                        mpiTestError( ok, " Recv from p" );
#ifdef UNIQUE
			printf("\n %d Receive %d elements from %d: from %d to %d, %d\n",mpiRank,countS, p, initS,initS+countS-1,  initS-begin+1);
#endif
		}
	   // }	
	}

	// 15. END Communications 
	MPI_Waitall(2*procs , requestI, statusVI );


	free(requestI);
	free(statusVI);


#ifdef UNIQUE
	printf("\n Init matrix\n");
	for (i=0; i<=size; i++) {
		printf("%lf ", element(matrix, i));	   
	}
#endif


	// 1-Bº communicacion trayendome los datos de los vecinos
	MPI_Request request_i[2] ={MPI_REQUEST_NULL, MPI_REQUEST_NULL} ;
	MPI_Status statusV_i[2];
	if(mpiRank!= procs-1){
		ok = MPI_Isend( &element(matrix, size), 1, MPI_DOUBLE, rankDown, 0, MPI_COMM_WORLD, &request_i[0] );
                mpiTestError( ok, "Send p" );
	}

	if ( mpiRank != 0 ) {
 		ok = MPI_Irecv( &element(matrix, 0), 1, MPI_DOUBLE, rankUp, 0, MPI_COMM_WORLD, &request_i[1] );
                mpiTestError( ok, " Recv from p" );
	}
	MPI_Waitall(2 , request_i, statusV_i );

	// Numero de datos por proceso
	int *numDatperProc=NULL, *numDatperProcT=NULL;
	numDatperProc = (int *)calloc( (size_t)(procs), sizeof(int) );
	numDatperProcT = (int *)calloc( (size_t)(procs), sizeof(int) );
	for(i=0;i<procs;i++){
		numDatperProc[i]=0;
		numDatperProcT[i]=-1;
	}


#ifdef UNIQUE
	printf("\n Busqueda iguales\n");
	for (i=0; i<=size; i++) {
		printf("%lf ", element(matrix, i));	   
	}
	printf("\n");
#endif


	// 2º copio en un bufer los datos mios que tengo que copiar en el output
	int k;
	for (i=1; i<size+1; i++) {
		// si es diferente lo copi
		if(element(matrix, i)!=element(matrix, i-1)){
			//Lo copio
	  		element(matrix_aux, numDatperProc[mpiRank]) = element(matrix, i);
			// aumento el contador de numero de datos
			numDatperProc[mpiRank]++;	
		}	
	}


#ifdef UNIQUE
	printf("\n cuales copio\n");
	for (i=0; i<numDatperProc[mpiRank]; i++) {
		printf("%lf ", element(matrix_aux, i));	   
	}
	printf("\n");
#endif

	// 3º Reduce el numero total de datos a copiar por cada proceso 
   	 MPI_Allreduce(numDatperProc,numDatperProcT,procs,MPI_INT,MPI_SUM,MPI_COMM_WORLD);

#ifdef UNIQUE
	printf("\n datos per proc\n");
	for (i=0; i<procs; i++) {
		printf("%d ", numDatperProcT[i]);	   
	}
	printf("\n");
#endif


	//SAco los datos previos que tengo para todos los procs para conocer donde copio
	int *numDatPrev=NULL;
	numDatPrev = (int *)calloc( (size_t)(procs), sizeof(int) );
	numDatPrev[0]=0; 
	//numDatPrev[1]=numDatperProcT[0];
	for(p=1;p<procs;p++){
		numDatPrev[p]=numDatperProcT[p-1]+ numDatPrev[p-1];
	}

#ifdef UNIQUE
	printf("\n datos previos per proc\n");
	for (i=0; i<procs; i++) {
		printf("%d ", numDatPrev[i]);	   
	}
	printf("\n");
#endif



	// 4º Sabiendo el numero de datos a copiar por cada proceso, calculo donde ya quien voy a copiar mis datos y les envio
	
        MPI_Request *request ;
	MPI_Status *statusV;
        request = (MPI_Request *)calloc( (size_t)(2*procs), sizeof(MPI_Request) );
        statusV = (MPI_Status *)calloc( (size_t)(2*procs), sizeof(MPI_Status) );
	for(i=0;i < 2*procs; i++){
		request[i]=MPI_REQUEST_NULL;
	}


#ifdef UNIQUE
fprintf(stderr,"%d- SENDS\n", mpiRank);
#endif
	// trozo donde voy a mandar mis datos de matrix2
	int beginDmy=begin2+numDatPrev[mpiRank];
	int endDmy=begin2+numDatPrev[mpiRank]+numDatperProc[mpiRank]-1; 
	// TRAVERSE THE PROCESSES TO KNOW WHAT LOCAL HAS TO SEND
	for(p=0;p < procs; p++){

		// CALCULATE THE BEGIND AND ENDD OF THE p PROCESS
		beginp= begin2 + (int)floor( p*procRatio );
		endp= begin2 + (int)floor( (p+((N_1 > procs) ? 1 : 0))*procRatio - ((N_1 > procs) ? 1 : 0) );

#ifdef UNIQUE
fprintf(stderr,"CTRL %d-%d, trozo[%d:%d]  matrixp[%d:%d]\n",mpiRank, p,beginDmy,endDmy, beginp, endp);
#endif
		// Intersect el trozo que tengo con donde tiene que ir
		if(beginDmy<=endp && endDmy>=beginp ){

			//send from the maximum of begin an beginpD; and the minimum of end and endpD
			int initS= max(beginp,beginDmy);
			int countS= min(endp,endDmy) - initS +1;
			ok = MPI_Isend( &element(matrix_aux, initS-beginDmy), countS, MPI_DOUBLE, p, 0, MPI_COMM_WORLD, &request[p] );
                        mpiTestError( ok, "Send p" );
#ifdef UNIQUE
			printf("\n %d Send %d elements to %d: from %d to %d, %d\n",mpiRank,countS, p, initS,initS+countS-1, initS-beginDmy);
#endif
		}
	}
	

	// 5º Sabiendo el numero de datos a copiar por cada proceso, calculo donde y de quien voy a recibir mis datos y les recibo
// trozo donde voy a mandar mis datos de matrix2
 

#ifdef UNIQUE
fprintf(stderr,"\n\n\n%d- RECEIVES\n", mpiRank);
#endif

	// TRAVERSE THE PROCESSES TO KNOW WHAT LOCAL HAS TO SEND
	for(p=0;p < procs; p++){

		// CALCULATE THE BEGIND AND ENDD OF THE p PROCESS
		beginpD= begin2 + numDatPrev[p]; 
		endpD= beginpD+numDatperProcT[p]-1;

#ifdef UNIQUE
fprintf(stderr,"CTRL %d-%d, matrix[%d:%d]  trozop[%d:%d]\n",mpiRank, p,beginD,endD, beginpD, endpD);
#endif
		// Intersect el trozo que tengo con donde tiene que ir
		if(beginpD<=endD && endpD>=beginD ){

			//send from the maximum of begin an beginpD; and the minimum of end and endpD
			int initS= max(beginpD,beginD);
			int countS= min(endpD,endD) - initS +1;
			ok = MPI_Irecv( &element(matrix2, initS-beginD), countS, MPI_DOUBLE, p, 0, MPI_COMM_WORLD, &request[procs+p] );
                        mpiTestError( ok, "Send p" );
#ifdef UNIQUE
			printf("\n %d Receive %d elements from %d: from %d to %d, %d\n",mpiRank,countS, p, initS,initS+countS-1, initS-beginD);
#endif
		}
	}
	
	MPI_Waitall(2*procs , request, statusV );

	free(request);
	free(statusV);


#ifdef UNIQUE
	printf("\n Salida\n");
	for (i=0; i<size; i++) {
		printf("%lf ", element(matrix2, i));	   
	}
#endif

	// Return the size allocated in local process
	*lenght_out=size;

#ifdef UNIQUE
	// Accumulate Result
	double sum=0, sumTotal=0;
	for (i=0; i<size; i++) {
		sum= sum+(beginD+i)*element(matrix2, i);
	   
	}
	MPI_Allreduce( &sum, &sumTotal, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD );
	if( mpiRank ==0){
		printf("\n Accumulated Sum: %lf \n", sumTotal);
	}
#endif


	// Free and return output vector
	free( matrix );
	free( matrix_aux );

	free(numDatPrev);
	free(numDatperProc);
	free(numDatperProcT);
	MPI_Comm_free( &cartesianComm );

	return matrix2; 


}


/*******************************************************/
/*      Main program - starts up on each machine       */
/*******************************************************/
int main( int argc, char *argv[ ] ) {

    /*  1. INIT MPI */
    MPI_Init( &argc, &argv );

    /* 2. ALLOCATE DATA STRUCTURES AND GET PARAMETERS */
    int error=0; 
    if ( argc != 2 ) {
    /* Check command line arguments; "random" is not really necessary;
       it's just used to show a concept */
   
        printf( "Usage: %s  <number to sort>\n", argv[0] );
        fflush( stdout );
        
        error = 1;
    } 
 
    /* Abort if there's an error on the command line */
    if ( error ) {
        return (1);
    }

    long int numberToSort = atol( argv[1] );
    double *items;			/* Array of info to be sorted 	   	  */

    if ( numberToSort < 2 || numberToSort > MAX_ARRAY ) {
        printf( "The number to sort must be between 2 and %d\n", 
                   MAX_ARRAY );
        error = 1;
        /* Abort if there's an error on the command line */
        if ( error ) {
            return (1);
        }
    } else if( (items = malloc( sizeof(double) * numberToSort ) ) == 0 ){
            printf(stderr,"Unable to allocate enough memory, exiting...\n\n");
            error = 1;
            /* Abort if there's not enough memory */
            if ( error ) {
              return (1);
            }
    }

    /* 3. INIT DATA STRUCTURES */
    int myNodeNumber, numberOfProcessors, length;
    MPI_Comm_rank( MPI_COMM_WORLD,&myNodeNumber );
    MPI_Comm_size( MPI_COMM_WORLD,&numberOfProcessors );
    InitData(numberToSort,items);
    int dimension = floor( log2( (double) numberOfProcessors ) );
    int numberOfNodes = 1 << dimension;
    MPI_Status Status;	
    /* Generate numbers to sort and split the work among the processors */
    if ( myNodeNumber == HOST ) {
       /* SendData will return length and array for the HOST node */
       SendData( numberToSort, numberOfNodes, &length, items );

    } else if ( myNodeNumber < numberOfNodes ) {

       /* If this processor is actually participating */
       MPI_Recv( items, MAX_ARRAY, MPI_DOUBLE, HOST, DATA, MPI_COMM_WORLD, 
 	         &Status );
       MPI_Get_count( &Status, MPI_DOUBLE, &length );
    } else {
       /* If this processor is NOT participating */
       length = 0;
    }

    // Obtaining which part is allocated in each process 
    int *lenghts=NULL,*lenghtsT=NULL;
    lenghts = (int *)calloc( (size_t)(numberOfProcessors), sizeof(int) );
    lenghtsT = (int *)calloc( (size_t)(numberOfProcessors), sizeof(int) );
    int i;
    for (i=0;i<numberOfProcessors; i++){
	lenghts[i]=0;
	lenghtsT[i]=0;
    }
    lenghts[myNodeNumber]=length;
    MPI_Allreduce(lenghts,lenghtsT,numberOfProcessors,MPI_INT,MPI_SUM,MPI_COMM_WORLD);
    double *m2=NULL;
    int lenght_out= length; 
    int begin =  (int)floor( myNodeNumber*((double)numberToSort/numberOfProcessors ));

#ifdef DEBUG
	{
	printf("\n Init Unique\n");
	int j;
	for(j=0;j<length;j++){
		printf("%lf ",items[j] );
	}
	}
#endif

    /* Start timing */
    MPI_Barrier( MPI_COMM_WORLD );
    double startTime = MPI_Wtime( );

    /// Unique copy 
    m2=unique_copy(items, lenghtsT,  &lenght_out, numberToSort);

   /* End timing */
   MPI_Barrier( MPI_COMM_WORLD );
   double endTime = MPI_Wtime( ) - startTime;
   double clockFinalT=0;
   MPI_Allreduce( &endTime, &clockFinalT, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD );
   if( myNodeNumber ==0){
     	printf("\nClock main: %lf\n", clockFinalT );
    }

   {
   int j; double suma=0, sumaTotal=0;
   for(j=0;j<lenght_out;j++){
	suma = suma + m2[j];
	}
   
   MPI_Allreduce( &suma, &sumaTotal, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD );
   if( myNodeNumber ==0) printf("\n SUM: %lf \n SQRT: %lf\n", sumaTotal, sqrt(sumaTotal));
  }

#ifdef DEBUG
	{
	printf("\n FINAL Unique\n");
	int j; 
	for(j=0;j<lenght_out;j++){
		printf("%lf ",m2[j] );
	}

	}
#endif

    /* 5. END MPI */
    MPI_Finalize( );
    return 0; 
}
