/*
* Hitmap Unique_copy using the operators of the HLPP'16 conference
* 
*
* @author Ana Moretón Fernández
* @version v0.1
*
* (c) 2017, Grupo Trasgo, Universidad de Valladolid
*/
#include <hitmap.h>
#include "../operadores.h"

// Communication function
#include "comGPluto.h"


HitClock        mainClock;

int N_1=0;

void distribute_seq(HitTile_double aux, int *left, HitTile_int data){
	int i,j=0;
	hit_tileForDimDomain( aux, 0, i ){
		if( hit_tileElemAt( data, 1,i)==0){
			j++;
		}	
	}
	*left=j;
}

int remove_tile (HitTile_double * m, int begin, int end, HitTile_int data)
{
  HitTile_double m2 = HIT_TILE_NULL_STATIC_TRASGO;
  HitTile_double aux = HIT_TILE_NULL_STATIC_TRASGO;
  HitShape auxShape;

  int left=0, totalLeft=0;
  double weights = 0;

  // MAP
  HitShape _TT_insideShapemap1 = hit_shape (1, hit_sig (begin, end-1, 1));
  HitTopology _TT_topomap1 = hit_topology (plug_topPlain);
  HitLayout map1 = hit_layout (plug_layBlocks, _TT_topomap1, _TT_insideShapemap1);

  auxShape = hit_layShape (map1);
  hit_tileDomainShapeAlloc (&(aux), double, auxShape);
  (aux).Layout = map1;

  //ARRAY REDISTRIBUTION
  {
    HitShape range1 = hit_shape (1, hit_sig (begin, end-1, 1)); 
    ArrayRemapRange((*m),aux, range1, range1 , double, 0);
  }

  distribute_seq (aux, &left, data);

  HitTile leftTile = hit_tileSingle (left, int);
  HitTile totalLeftTile = hit_tileSingle (totalLeft, int);

  hit_comDoOnce (hit_comReduce (map1,
				HIT_RANKS_NULL,
				&leftTile,
				&totalLeftTile, HIT_INT, HIT_OP_SUM_INT));

  weights = hit_tileDimCard (*m, 0);

  /////////////////////////////////////////////////////
  ///////         ARRAY PARTITION
  /////////////////////////////////////////////////////
  int num_maxMask_aux_2 = 2;
  HitTile_double (*_TT_Newtileaux_2) = malloc (num_maxMask_aux_2 * sizeof (HitTile_double));
  int (*_TT_indiceGlobal_aux_2) = (int *) malloc (hit_tileDims (aux) * (num_maxMask_aux_2 + 1) *  sizeof (int));
  HitLayout grouping;
  ArrayDivide_dev(grouping, 2, &weights, &aux, _TT_Newtileaux_2, data,_TT_indiceGlobal_aux_2, double ); 
  ArrayMerge_dev(grouping,1, &aux, _TT_Newtileaux_2, _TT_indiceGlobal_aux_2, double);

  // Map
  HitShape _TT_insideShapemapAux2;
  _TT_insideShapemapAux2 = hit_shapeSubset (hit_layFullShape (((*m)).Layout),
					    hit_shape (1, hit_sig (0, totalLeft - 1, 1)));
  HitLayout mapAux2 = hit_layout (plug_layBlocks, _TT_topomap1, _TT_insideShapemapAux2);

  // Last redistribution

  HitShape mShape;
  mShape = hit_layShape (mapAux2);
  hit_tileDomainShapeAlloc (&(*m), double, mShape);
  (*m).Layout = mapAux2;

  {
    HitShape range2 = hit_shape (1, hit_sig (0, totalLeft , 1)); 
    HitShape range1 = hit_shape (1, hit_sig (begin, begin+totalLeft , 1)); 
    ArrayRemapRange(aux,(*m), range1, range2 , double, 0);
  }
  hit_layFree (mapAux2);
  hit_layFree (grouping);
  hit_layFree (map1);
  hit_topFree (_TT_topomap1);

  hit_tileFree (aux);

  return 0;
}



void mark_seq (HitTile_double aux, HitTile_int marcas){
	int i;
	hit_tileElemAtArrayCoords(marcas,1,hit_tileDimBegin(aux, 0))=1;
	for( i=hit_tileDimBegin(aux, 0)+1;i<=hit_tileDimEnd(aux,0);i=i+hit_tileDimStride(aux,0) ){
		hit_tileElemAtArrayCoords(marcas,1,i)=0;
		if(hit_tileElemAtArrayCoords( aux, 1,i-1) == hit_tileElemAtArrayCoords( aux, 1,i)){
			hit_tileElemAtArrayCoords(marcas,1,i)=1;
		}
	}
}


int
unique_copy (HitTile_double m, int begin, int end, HitTile_double * m2, int begin2)
{

  HitTile_double new = HIT_TILE_NULL_STATIC_TRASGO;
  HitShape newShape;

  HitTile_int marks = HIT_TILE_NULL_STATIC_TRASGO;
  HitShape marksShape;

  int N;

  // Layout 
  HitLayout mapData;
  HitShape domainData = hit_shape (1, hit_sig (begin, end-1, 1));
  HitTopology topo = hit_topology (plug_topPlain);
  mapData = hit_layout (plug_layBlocks, topo, domainData);

  // Mask Array 
  marksShape = hit_layShape (mapData);
  marksShape = hit_shapeDimExpand (marksShape, 0, HIT_SHAPE_BEGIN, -1);
  /*hit_tileSelect( &marks, &domainData, marksShape );		
  hit_tileAlloc( &marks);*/
  hit_tileDomainShapeAlloc (&(marks), int, marksShape);
  (marks).Layout = mapData;

  // New Array 
  newShape =  hit_layShape (mapData);
  newShape = hit_shapeDimExpand (newShape, 0, HIT_SHAPE_BEGIN, -1);
  /*hit_tileSelect( &new, &domainData, newShape );		
  hit_tileAlloc( &new);*/
  hit_tileDomainShapeAlloc (&(new), double, newShape);
  (new).Layout = mapData; 

#ifdef DEBUG
  printf("Begin: %d End: %d \n", begin, end);
  dumpShape(newShape);
  dumpShape(hit_tileShape(new));
#endif

  int zero=0,j; 
  for(j=0;j<hit_tileDimCard(new,0);j++){
	hit_tileElemAt(new, 1, j ) =0; 
	hit_tileElemAt(marks, 1, j ) =0; 
   }
 /* hit_tileFill( &marks, &zero );
  hit_tileFill( &new, &zero );*/

  //  ARRAY REDISTRIBUTION 
  {
    HitShape range1 = hit_shape (1, hit_sig (begin, end-1, 1)); 
    HitShape range2 = hit_shape (1, hit_sig (begin, end-1, 1)); 
    ArrayRemapRange(m,new, range1, range2, double, 0);
  }

  HitPattern _TT_neighSync_first3 = hit_pattern (HIT_PAT_UNORDERED);
  _TT_neighSync_first3 = com_first3 (new, mapData);
  hit_patternDo (_TT_neighSync_first3);



  mark_seq(new,marks); 

#ifdef DEBUG
	{
	int j;
	printf("\n NEW Unique\n");
	for(j=0;j<hit_tileDimCard(new,0);j++){
		printf("%lf ",hit_tileElemAt(new, 1, j ) );
	}
	printf("\n Marks Unique\n");
	for(j=0;j<hit_tileDimCard(new,0);j++){
		printf("%d ",hit_tileElemAt(marks, 1, j ) );
	}
	}
#endif

  int left,totalLeft; 
  distribute_seq (new, &left, marks);
  HitTile leftTile = hit_tileSingle (left, int);
  HitTile totalLeftTile = hit_tileSingle (totalLeft, int);

  hit_comDoOnce (hit_comReduce (new.Layout,
				HIT_RANKS_NULL,
				&leftTile,
				&totalLeftTile, HIT_INT, HIT_OP_SUM_INT));


  double weights = hit_tileDimCard (new, 0);

  /////////////////////////////////////////////////////
  ///////         ARRAY PARTITION
  /////////////////////////////////////////////////////
  int num_maxMask_aux_2 = 2;
  HitTile_double (*_TT_Newtileaux_2) = malloc (num_maxMask_aux_2 * sizeof (HitTile_double));
  int (*_TT_indiceGlobal_aux_2) = (int *) malloc (hit_tileDims (new) * (num_maxMask_aux_2 + 1) *  sizeof (int));

  HitLayout grouping;
  ArrayDivide_dev(grouping, 2, &weights, &new, _TT_Newtileaux_2, marks,_TT_indiceGlobal_aux_2, double ); 
#ifdef DEBUG
	{
	int j;
	printf("\n _TT_Newtileaux_2 Unique 0 \n");
	for(j=0;j<hit_tileDimCard(_TT_Newtileaux_2[0],0);j++){
		printf("%lf ",hit_tileElemAt(_TT_Newtileaux_2[0], 1, j ) );
	}
	printf("\n _TT_Newtileaux_2 Unique 1 \n");
	for(j=0;j<hit_tileDimCard(_TT_Newtileaux_2[1],0);j++){
		printf("%lf ",hit_tileElemAt(_TT_Newtileaux_2[1], 1, j ) );
	}
	}
#endif

#ifdef DEBUG
	printf("\n TotalLeft %d \n", totalLeft);
#endif
 
  N=totalLeft-1; 
  //  ARRAY REDISTRIBUTION 
  {
    HitShape range1 = hit_shape (1, hit_sig (begin2, begin2 + N , 1)); 
    HitShape range2 = hit_shape (1, hit_sig (0, N, 1)); 
    ArrayRemapRange(_TT_Newtileaux_2[0],(*m2), range2, range1 , double, 0);
  }

#ifdef DEBUG
	{
	int j;
	printf("\n m Unique\n");
	for(j=0;j<hit_tileDimCard((*m2),0);j++){
		printf("%lf ",hit_tileElemAt((*m2), 1, j ) );
	}
	}
#endif

  hit_layFree (mapData);
  hit_topFree (topo);


  hit_tileFree (marks);
  hit_tileFree (_TT_Newtileaux_2[0]);
  hit_tileFree (_TT_Newtileaux_2[1]);
  free(_TT_indiceGlobal_aux_2);
  free(_TT_Newtileaux_2);

  return 0;
}




/*******************************************************/
/*             Main program - Unique_copy              */
/*******************************************************/
int main (int argc, char *argv[])
{

  	hit_comInit (&argc, &argv);

	if ( argc != 2 ) {
                fprintf(stderr, "\nUsage: %s <numRows> \n", argv[0]);
                exit(EXIT_FAILURE);
        }

	/* tam */
 	int tam = atoi( argv[1] );

    	/* 1. CREATE VIRTUAL TOPOLOGY */
    	HitTopology topo = hit_topology( plug_topPlain );

        /* 2. DECLARE FULL MATRIX WITHOUT MEMORY */
        HitTile_double mat;
        hit_tileDomain( &mat, double, 1, tam );

    	/* 3. COMPUTE PARTITION */
        HitShape parallelShape = hit_tileShape( mat );
        HitLayout matLayout = hit_layout( plug_layBlocks, topo, parallelShape );
	HitShape parShape = hit_layShape(matLayout);

	/* DATA STRUCTURES */
	HitTile_double m, m2;
	hit_tileDomainShapeAlloc(&m, double, parShape); 
	hit_tileDomainShapeAlloc(&m2, double, parShape); 
	(m).Layout = matLayout;
	(m2).Layout = matLayout;

	/* 4. INIT VECTOR*/
	int x ;
	hit_tileForDimDomain(m, 0, x){
			srand48(x+hit_layDimBegin (matLayout, 0));
			double aux=(double)((int)ceil(drand48()*(double)tam));
			hit_tileElemAt(m, 1, x ) = aux; 	
			hit_tileElemAt(m2, 1, x ) = 1.5; 	
			if( (x+hit_layDimBegin (matLayout, 0)) < 5){ hit_tileElemAt(m, 1, x )=2 ; }	
	}

#ifdef DEBUG
	{
	int j;
	printf("\n INIT Unique\n");
	for(j=0;j<hit_tileDimCard(m,0);j++){
		printf("%lf ",hit_tileElemAt(m, 1, j ) );
	}
	}
#endif

        hit_clockSynchronizeAll();
        hit_clockStart( mainClock );

	/* 5. UNIQUE */
	unique_copy (m, 0, tam, &m2, 0);

	hit_clockStop( mainClock );
	hit_clockReduce( matLayout, mainClock );
	{
	int j;
	double suma=0, sumaTotal=0;
	for(j=0;j<hit_tileDimCard(m,0);j++){
		suma = suma +hit_tileElemAt(m2, 1, j );
	}
	MPI_Allreduce( &suma, &sumaTotal, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD );
	if (hit_Rank==0) printf("\n SUM: %lf \n SQRT: %lf\n", sumaTotal, sqrt(sumaTotal));
	}

#ifdef DEBUG
	{
	int j;
	printf("\n END Unique\n");
	for(j=0;j<hit_tileDimCard(m,0);j++){
		printf("%lf ",hit_tileElemAt(m2, 1, j ) );
	}
	}
#endif

	if (hit_Rank==0){
			printf("\nClock main: %lf\n", mainClock.max );
	}


	/* 6. FREE */
  	hit_layFree (matLayout);
  	hit_topFree (topo);
  	hit_tileFree (m);
  	hit_tileFree (m2);

  	hit_comFinalize ();
  	return 0;

}


