  

/* The different kind of kernel that can be used*/
//#define HIT_COMPILE_CUDA
//#define HIT_COMPILE_CORE

  
#include <hitmap.h>
#include <math.h>
#include <time.h>
#include "c_randi8.h"

#include <hit_acel.h>
#include <hit_partition.h>

#include "config.h"

#define N_MATRIX 8192


hit_tileNewType( int );
      

#define	MATRIX_AROWS		N_MATRIX				//N=M=Costante config.h
#define	MATRIX_ACOLUMNS		N_MATRIX
#define	MATRIX_BCOLUMNS		N_MATRIX

/* Clocks*/
HitClock	mainClock;
HitClock	productClock;
HitClock	comClockMax;
HitClock	comClockAve;
HitClock	sendRecivLaunch;
HitClock	sendRecivLaunchSUM;
HitClock	productClockSUM;


void cannonsMM( int n, int m, int p );


/*KERNEL 2: matrix multiplication (cannon implementation) with CPU CORES*/
kernelDefinitionCORE(CORE,matrix_mul, HitTile_int  A, HitTile_int B, HitTile_int C);


/* A. MAIN: INVOCATION OF CANNON'S ALGORITHM */
int main(int argc, char *argv[]){
	hit_comInit( &argc, &argv );

	cannonsMM( MATRIX_AROWS, MATRIX_ACOLUMNS, MATRIX_BCOLUMNS );

	hit_comFinalize();
	return 0;
}


/* B. FUNCTION TO COMPUTE THE PRODUCT OF TWO LOCAL SUBMATRICES */
void matrixBlockProduct( HitTile_int A, HitTile_int B, HitTile_int C ) {
	int i,j,k;
#ifdef OPTIMIZE_PRODUCT
	int *datas[3] = { C.data, A.data, B.data };
	int rowSizes[3] ={  C.origAcumCard[1],A.origAcumCard[1],B.origAcumCard[1] };
#  ifdef OPTIMIZE_LOOPS
	int loopLimits[3] = { hit_tileDimCard(C,0), hit_tileDimCard(C,1), hit_tileDimCard(A,1) };
	for (i=0; i<loopLimits[0]; i++) {
		for (j=0; j<loopLimits[1]; j++) {
			for (k=0; k<loopLimits[2]; k++) {
#   else
	for (i=0; i<hit_tileDimCard(C,0); i++) {
		for (j=0; j<hit_tileDimCard(C,1); j++) {
			for (k=0; k<hit_tileDimCard(A,1); k++) {
#   endif
#else
	for (i=0; i<hit_tileDimCard(C,0); i++) {
		for (j=0; j<hit_tileDimCard(C,1); j++) {
			for (k=0; k<hit_tileDimCard(A,1); k++) {
#endif

#ifdef OPTIMIZE_PRODUCT
					datas[0][i*rowSizes[0]+j] =
						datas[0][i*rowSizes[0]+j] + datas[1][i*rowSizes[1]+k] * datas[2][k*rowSizes[2]+j] ;
#else
					hit_tileElemAtNoStride(C,2,i,j) = 
							hit_tileElemAtNoStride(C,2,i,j) 
							+ hit_tileElemAtNoStride(A,2,i,k) * hit_tileElemAtNoStride(B,2,k,j);
#endif
			}
		}
	}
}


/* C. INITIALIZE MATRICES */
void initMatrices( int m, int p, 
					HitTile_int tileA, HitTile_int tileB, HitTile_int tileC,
					HitLayout layA, HitLayout layB ) {
	/* 1. INIT C = 0 */
	int zero=0; 
	hit_tileFill( &tileC, &zero );

#ifdef READ_AB
	hit_tileFileRead( &tileA, "matrixA.dat" );
	hit_tileFileRead( &tileB, "matrixB.dat" );
#else
	int i,j;

	/* 2. INIT A(i,j) = ( i*numCol + j ) % 1000 */
	hit_tileFill( &tileA, &zero );
	hit_layForDimDomain( layA, 0, i ) 
		hit_layForDimDomain( layA, 1, j )  
			hit_tileElemAtArrayCoords( tileA, 2, i,j ) = ( i * m + j ) % 1000;

	/* 3. INIT B: DIFFERENT COMPILE-TIME OPTIONS */
	hit_tileFill( &tileB, &zero );
#ifdef B_IDENTITY

	/* 3.a. INIT B = Identity */
	hit_layForDimDomain( layB, 0, i )  
		hit_layForDimDomain( layB, 1, j )  
			hit_tileElemAtArrayCoords( tileB, 2, i,j ) = (i==j) ?  1.0 : 0.0;
#else
	/* 3.b. INIT B = Random with seed */
	srand48( 364543 );
	for (i=0; i<m; i++) {
		for (j=0; j<p; j++) {
			if ( i>=hit_layDimBegin(layB,0) && i<=hit_layDimEnd(layB,0) && j>=hit_layDimBegin(layB,1) && j<=hit_layDimEnd(layB,1) ) 
				hit_tileElemAtArrayCoords( tileB, 2, i,j ) = (int)(1000*drand48());
			else drand48();
		}
	}
#endif
#ifdef OUTPUT_NUNCA
	/* 4. OUTPUT: WRITE A,B ON FILES */
	int err;
	if ( ( err=hit_tileFileWriteText( &tileA, "A.dtxt", HIT_FILE_INT, 14, 4 ))) {
		fprintf(stderr,"CTRL [%d] Error writing file. Error: %d\n", hit_Rank, err);
	}
	if ( ( err=hit_tileFileWriteText( &tileB, "B.dtxt", HIT_FILE_INT, 14, 4 ))) {
		fprintf(stderr,"CTRL [%d] Error writing file. Error: %d\n", hit_Rank, err);
	}
#endif
#endif
}


/* D. PARALLEL CANNON'S ALGORITHM FOR MATRIX MULTIPLICATION */
void cannonsMM( int n, int m, int p )  {
	/* 0. INIT CLOCKS */
	hit_clockSynchronizeAll();
	hit_clockStart( mainClock );
	hit_clockReset( productClock );


	

    /* 1. CREATE VIRTUAL TOPOLOGY */
    HitTopology topo = hit_topology( plug_topSquare );		//Copy Y

	/* 2. DECLARE FULL MATRICES WITHOUT MEMORY */
	HitTile_int A, B, C;
//	hit_tileDomain( &A, sizeof(int), HIT_NONHIERARCHICAL, 2, n, m );		//Copy Y hit_tileDomain( &A, int, 2, n, m );
//	hit_tileDomain( &B, sizeof(int), HIT_NONHIERARCHICAL, 2, m, p );
//	hit_tileDomain( &C, sizeof(int), HIT_NONHIERARCHICAL, 2, n, p );

	hit_tileDomain( &A, int, 2, n, m );
	hit_tileDomain( &B, int, 2, m, p );
	hit_tileDomain( &C, int, 2, n, p );

    /* 3. COMPUTE PARTITIONS */


#ifndef RATIOMACRO

	HitLayout layA = hit_layout(plug_layBlocks, topo, hit_tileShape( A ));			//Copy Y copiar el concepto, 3 layoutsssss
	HitLayout layB = hit_layout(plug_layBlocks, topo, hit_tileShape( B ));
    	HitLayout layC = hit_layout(plug_layBlocks, topo, hit_tileShape( C ));			

#else

	HitLayout layA = hit_layout(plug_layBlocksBalance, topo, hit_tileShape( A ), 0, RATIO);			//Copy Y copiar el concepto, 3 layoutsssss
	HitLayout layB = hit_layout(plug_layBlocks, topo, hit_tileShape( B ));
    	HitLayout layC = hit_layout(plug_layBlocksBalance, topo, hit_tileShape( C ), 0, RATIO);			
#endif

	hit_layWrapNeighbors( &layA );
	hit_layWrapNeighbors( &layB );


	/* The maximum shape of the three layouts*/
//	HitShape maxMaxShape= hit_shapeMaxMaxShape(hit_layMaxShape(layA), hit_layMaxShape(layB), hit_layMaxShape(layC), HIT_TOPOLOGY_INFO.selfRank);


		HitPartition partitionPC;						//ThreadBlock access pattern version		


 		/* Creating the partition from threadBlock access pattern*/

/*
		partitionPC= hit_partition(plug_partBlocksPattern,shapeAux, 3, //hit_layMaxShape(layC)
						hit_shape2(hit_sigONE_POS(), hit_sigALL_POS()), 
							hit_shape2(hit_sigALL_POS(), hit_sigONE_POS()), hit_shapeNULL_STATIC());	


		partitionPA= hit_partition(plug_partBlocksPattern, hit_layMaxShape(layA), 3, //hit_layMaxShape(layC)
						hit_shape2(hit_sigONE_POS(), hit_sigALL_POS()), 
							hit_shape2(hit_sigALL_POS(), hit_sigONE_POS()), hit_shapeNULL_STATIC());	


		partitionPB= hit_partition(plug_partBlocksPattern, hit_layMaxShape(layB), 3, //hit_layMaxShape(layC)
						hit_shape2(hit_sigONE_POS(), hit_sigALL_POS()), 
							hit_shape2(hit_sigALL_POS(), hit_sigONE_POS()), hit_shapeNULL_STATIC());	

*/


//	partitionPC=hit_partition(plug_partNaive, hit_layShape(layC), 3,hit_shapeTHISTHIS, hit_shapeTHISALL, hit_shapeALLTHIS);
//	partitionPA=hit_partition(plug_partNaive, hit_layMaxShape(layA));
//	partitionPB=hit_partition(plug_partNaive, hit_layMaxShape(layB));





	/* 4. CREATE AND ALLOCATE TILES */
	HitTile_int tileA, tileB, tileC;// copyTileA, copyTileB;
	hit_tileSelectNoBoundary( &tileA, &A, hit_layMaxShape(layA));
	hit_tileSelectNoBoundary( &tileB, &B, hit_layMaxShape(layB));
	hit_tileSelect( &tileC, &C,  hit_layShape(layC) );

//	hit_tileSelectNoBoundary( &tileA, &A, maxMaxShape); // al mayor de los dos los shapes
//	hit_tileSelectNoBoundary( &tileB, &B, maxMaxShape);
//	hit_tileSelect( &tileC, &C,  maxMaxShape);
//	hit_tileSelectNoBoundary( &tileC, &C,  maxMaxShape);

	hit_tileAlloc( &tileA ); hit_tileAlloc( &tileB ); hit_tileAlloc( &tileC );
//	hit_tileClone( &copyTileA, &tileA ); hit_tileClone( &copyTileB, &tileB );




	/* 5. INITIALIZE MATRICES */
	initMatrices( m, p, tileA, tileB, tileC, layA, layB );


	/* 6. INITIAL ALIGNMENT PHASE */
	HitCom commRow, commColumn;



	commRow = hit_comShiftDim( layA, 1, -hit_laySelfRanksDim(layA,0), &tileA, HIT_INT );
	commColumn = hit_comShiftDim( layB, 0, -hit_laySelfRanksDim(layB,1), &tileB, HIT_INT );



	hit_comDo( &commRow );
	hit_comDo( &commColumn );
	hit_comFree( commRow );
	hit_comFree( commColumn );

	/* 7. REPEATED COMM PATTERN */
	HitPattern shift = hit_pattern( HIT_PAT_UNORDERED );
	hit_comTagSet( TAG_A, TAG_B );
//	hit_patternAdd( &shift, hit_comShiftDimAsync(layA, 1, 1, &tileA, &copyTileA, HIT_INT, TAG_A));
//	hit_patternAdd( &shift, hit_comShiftDimAsync(layB, 0, 1, &tileB, &copyTileB, HIT_INT, TAG_B));
	hit_patternAdd( &shift, hit_comShiftDim( layA, 1, 1, &tileA, HIT_INT ) );
	hit_patternAdd( &shift, hit_comShiftDim( layB, 0, 1, &tileB, HIT_INT ) );



	/* GPU workload*/
	hit_acelFasterDeviceWorkLoad(FasterDeviceWorkLoad);

	/* To initialize the mutual*/	
	hit_acelMExclusionInit();

	/* 8. DO COMPUTATION */
	int loopIndex;
	int loopLimit = max( layA.numActives[0], layA.numActives[1] );
	for (loopIndex = 0; loopIndex < loopLimit-1; loopIndex++) {


		/* 8.1. START COMMUNICATIONS FROM TILES TO REMOTE TILE BUFFERS */
//		hit_patternStartAsync( shift );
	
		/* Mutual exclusion start*/
		hit_acelExclusionStart();


			/* Device kernel variables*/
//			hit_acelDeclTile(tileA,partitionPA,hit_shapeALLALL,int);
//			hit_acelDeclTile(tileB,partitionPB,hit_shapeALLALL,int);
//			hit_acelDeclTile(tileC,partitionPC,hit_shapeALLALL,int);




			/* The data are sent to GPU*/
//			hit_acelSendData(tileA);
//			hit_acelSendData(tileB);
//			hit_acelSendData(tileC);


				/* Kernel launch*/
					/* The kernel is launched on a GPU only in heterogeneous system only if the mpi threads < number of current GPUs*/

			hit_clockContinue( productClock );
			
				//hit_acelKernelLaunch(matrix_mul, partitionPC, hit_tileIN_OUT(3,IN, IN, INOUT),tileA, tileB, tileC);
//hit_acelKernelLaunch(matrix_mul, partition,hit_partition(plug_partNaive, hit_layShape(layC), 3,hit_shapeTHISTHIS, hit_shapeTHISALL, hit_shapeALLTHIS), hit_tileIN_OUT(3,IN, IN, INOUT),tileA, tileB, tileC);
   HitPartition PART=hit_partition(plug_partNaive, hit_layShape(layC), 3,hit_acelPatTHISTHIS, hit_acelPatTHISALL, hit_acelPatALLTHIS);
hit_acelKernelLaunch(matrix_mul, PART, int, hit_acelTileIN_OUT(3,IN,IN,INOUT),tileA, tileB,tileC);
//hit_acelKernelLaunch(resid, partition, float, hit_acelTileIN_OUT(3,IN,IN,INOUT),m_u, m_v,m_r);
			
			hit_clockStop( productClock );

			/* Receiving data*/
//			hit_acelRecvData(tileC);

			/* Freeing memory*/
//			hit_acelFree(tileA);
//			hit_acelFree(tileB);
//			hit_acelFree(tileC);

			/* Liberating the accelerator threads*/
			hit_acelThreadExit();

		/* Mutual exclusion stop*/
		hit_acelExclusionStop();


		/* 8.2. COMMUNICATE TILES */
		hit_patternDo( shift );
		/* 8.3. END COMMUNICATIONS, MOVE DATA FROM TILE BUFFERS */
//		hit_patternEndAsync( shift );





	}//loop loopIndex


	/* 9. LAST ITERATION MULTIPLICATION: NO COMMUNICATION AFTER */
	//matrixBlockProduct( tileA, tileB, tileC );


	/* Mutual exclusion start*/
	hit_acelExclusionStart();




		/* Device kernel variables*/
//		hit_acelDeclTile(tileA,partitionPA, hit_shapeTHISTHIS,int);
//		hit_acelDeclTile(tileB,partitionPB, hit_shapeTHISTHIS,int);
//		hit_acelDeclTile(tileC,partitionPC, hit_shapeTHISTHIS,int);

		/* The data are sent to GPU*/
//		hit_acelSendData(tileA);				
//		hit_acelSendData(tileB);
//		hit_acelSendData(tileC);

			/* Kernel launch*/

		hit_clockContinue( productClock );

				//hit_acelKernelLaunch(matrix_mul, partitionPC, hit_tileIN_OUT(3,IN, IN, INOUT),tileA, tileB, tileC);

   HitPartition PART=hit_partition(plug_partNaive, hit_layShape(layC), 3,hit_acelPatTHISTHIS, hit_acelPatTHISALL, hit_acelPatALLTHIS);
hit_acelKernelLaunch(matrix_mul, PART, int, hit_acelTileIN_OUT(3,IN,IN,INOUT),tileA, tileB,tileC);


		hit_clockStop( productClock );

	
		/* Receiving data*/
//		hit_acelRecvData(tileC);

		/* Freeing memory*/
//		hit_acelFree(tileA);
//		hit_acelFree(tileB);
//		hit_acelFree(tileC);

		/* Liberating the accelerator threads*/
		hit_acelThreadExit();

	/* Mutual exclusion stop*/
	hit_acelExclusionStop();

//printf("ADIOS   %d\n",HIT_TOPOLOGY_INFO.selfRank);
	/* 10. CLOCK RESULTS */
	hit_clockStop( mainClock );
	hit_clockMaxReduce( topo, mainClock );
	hit_clockMaxReduce( topo, productClock );
	hit_clockPrint( mainClock );
	hit_clockPrint( productClock );








	/* 11. WRITE RESULT MATRIX */
/*
{
int i,j;
for(i=0;i <hit_tileDimCard(tileC,0);i++){
	for(j=0; j<hit_tileDimCard(tileC,1);j++){
		printf("[%d] %02d %02d %014d\n", hit_Rank, i+hit_tileDimBegin(tileC,0), j+hit_tileDimBegin(tileC,1), hit_tileElemAtNoStride2(tileC,i,j));
	}
}
}
*/
			/*Loop 3: dot product iterator*/
#ifdef OUTPUT
	int err;	
	if ( ( err=hit_tileFileWriteText( &tileC, "C.dtxt", HIT_FILE_INT, 14, 4 ) ) ) {
		fprintf(stderr,"CTRL [%d] Error writing file. Error: %d\n", hit_Rank, err);
	}
#endif

	/* 12. FREE RESOURCES */
	hit_patternFree( &shift );
	hit_topFree( topo );
}





/**********************************************************************************************
*
*				KERNEL DEFINITIONS (without CUDA obviously)
*
**********************************************************************************************/

/*KERNEL 2: matrix multiplication (cannon implementation) with CPU CORES*/
kernelDefinitionCORE(CORE,matrix_mul, HitTile_int  A, HitTile_int B, HitTile_int C){

//fprintf(stderr,"ESTAMOS EN CORE   %d;;;;   A%d B%d  C%d***************\n\n",hit_Rank,A.data[0],B.data[0],C.data[0]);
	/*Variables*/
	int i,j,k,sum_aux;

	/*Loop 1: first dimension*/
	for(i=0;i <hit_tileDimCard(C,0);i++){

		/*Loop 2: second dimension*/		
		for(j=0; j<hit_tileDimCard(C,1);j++){

			sum_aux=hit_tileElemAtNoStride2(C,i,j);

			/*Loop 3: dot product iterator*/
			for(k=0; k <hit_tileDimCard(A,1);k++){

				/* Dot product operation*/
				sum_aux+=hit_tileElemAtNoStride2(A,i,k)*hit_tileElemAtNoStride2(B,k,j);

			}//for k
			
		/* Store the dot product result in the C matrix*/
		hit_tileElemAtNoStride2(C,i,j)= sum_aux;
		
		}//for j

	}//for i	
	
}//kernelDefinitionCORE



