#include <hit_tile.h>
#include <hit_layout.h>
#include <hit_utils.h>
#include <hit_acel.h>
#include <hit_partition.h>
#include <hit_sig.h>
#include <cutil.h>

#include <cuda_utils.h>
#include <cublas.h>

/* Declaring the Hitmap globlal variables types*/
hit_tileNewType( int );

/*********************************************************************************
*
*				KERNELS SECTION 
*
*********************************************************************************/

#define HIT_PARAM_KERNEL_LAUNCH_matrix_mul  tileA,tileB,tileC
#define HIT_PARAM_KERNEL_LAUNCH_POINTER  dataA,dataB,dataC

/* dim3 struct*/
#define blockStruct(name, paramA, paramB) dim3 name(paramA, paramB)

/* Kernel variables initialization (threadBlock and grid shape)*/
#define acelKernelVariables() blockStruct(block, hit_sigCard(hit_shapeSig(part.blockShape,0)),hit_sigCard(hit_shapeSig(part.blockShape,1)));\
					blockStruct(grid, hit_partition_totalSape(part,0),hit_partition_totalSape(part,1))


/*KERNEL 1: matrix multiplication with square thread-blocks (cannon implementation) with GPU*/
kernelDefinitionCUDA(GPUFER,matrix_mul, HitTile_int  *tileA, HitTile_int *tileB, HitTile_int *tileC){

	/* Variables definition and initialization*/
	int begin_point_x,begin_point_y;
	int result_point_x, result_point_y, result_aux;
	int fila,columna, iteraciones, extra_thread;

	begin_point_x	=	blockIdx.y*(blockDim.y);  
	begin_point_y	=	blockIdx.x*(blockDim.x);

	result_point_x	=	begin_point_x + threadIdx.y;
	result_point_y	=	begin_point_y + threadIdx.x;

	fila		=	result_point_x;
	columna		=	result_point_y;
	iteraciones	=	(hit_tileDimCard(*tileA,1)<hit_tileDimCard(*tileB,0)) ? hit_tileDimCard(*tileA,1):hit_tileDimCard(*tileB,0);

	extra_thread	=	hit_tileDimCard(*tileC,1);

	/* To test the threads execution zone*/
	threadZoneTest{	

			/* For the A and B matrices; not to exit form the cardinalities*/
			if (columna<iteraciones){
	
			result_aux= hit_tileElemAtNoStride(*tileC, 2, result_point_x, result_point_y);
	
				/* Computational loop*/
				for (int i=0;i<iteraciones;i++){
		
					result_aux= result_aux+  hit_tileElemAtNoStride(*tileA, 2, fila, i) * hit_tileElemAtNoStride(*tileB, 2, i, columna); 
		
				} //for
			}//if
		
			/* Storing the result*/
			hit_tileElemAtNoStride(*tileC, 2, result_point_x, result_point_y)=result_aux;
	}//threadZoneTest
}//__global__






/*********************************************************************************
*
*				INTERMEDIATE FUNCTIONS 
*
*********************************************************************************/


/* intermediate function to launch the kernel*/
//extern "C"
//void INTERNAL_matrix_mul(HitPartition part,HitTile_int *tileA, HitTile_int *tileB, HitTile_int *tileC, void *dataA, int *dataB,int *dataC, int rank)
//{
	/* Kernel variables definition*/
//	acelKernelVariables();

	/* Kernel launch*/
//	GPUFER_matrix_mul<<<grid,block>>>( tileA, tileB, tileC);
//	CUDA_SAFE_CALL(cudaThreadSynchronize());

//}//INTERNAL_matrix_mul










