#include <hit_tile.h>
#include <hit_layout.h>
#include <hit_utils.h>
#include <hit_partition.h>

#include <cutil.h>
#include <cublas.h>

#include "config.h"

#include <hit_acel.h>
#include <cuda_utils.h>

/** Smoother operator (for class S and A) */
#ifdef CLASS_SWA

#define operatorS_Kernel0    (-3.0/8.0)
#define operatorS_Kernel1    (1.0/32.0)
#define operatorS_Kernel2    (-1.0/64.0)
#define operatorS_Kernel3    (0.0)

#else

#define operatorS_Kernel0    (-3.0/17.0)
#define operatorS_Kernel1    (1.0/33.0)
#define operatorS_Kernel2    (-1.0/61.0)
#define operatorS_Kernel3    (0.0)

#endif


#define operatorA_Kernel0    (-8.0/3.0)
#define operatorA_Kernel1    (0.0)
#define operatorA_Kernel2    (1.0/6.0)
#define operatorA_Kernel3    (1.0/12.0)

#define operatorP_Kernel0    (1.0/2.0)
#define operatorP_Kernel1    (1.0/4.0)
#define operatorP_Kernel2    (1.0/8.0)
#define operatorP_Kernel3    (1.0/16.0)

#define operatorQ_Kernel0    (1.0)
#define operatorQ_Kernel1    (1.0/2.0)
#define operatorQ_Kernel2    (1.0/4.0)
#define operatorQ_Kernel3    (1.0/8.0)




/* Declaring the Hitmap globlal variables types*/
hit_tileNewType( float);

/*********************************************************************************
*
*				KERNELS SECTION 
*
*********************************************************************************/

#define HIT_PARAM_KERNEL_LAUNCH_psinv  tileR,tileZ
#define HIT_PARAM_KERNEL_LAUNCH_resid  tileU,tileV,tileR
#define HIT_PARAM_KERNEL_LAUNCH_rprj3  tileRO,tileR
#define HIT_PARAM_KERNEL_LAUNCH_interp  tileZO,tileZ



/* psinv MG kernel*/
kernelDefinitionCUDA(GPUFER,psinv, HitTile_float  *tileR, HitTile_float *tileZ){

	
	/* Macros*/	

	#undef z
	#undef r
	#define z(a,b,c) hit_tileElemAtNoStride(*tileZ, 3, a, b, c)
	#define r(a,b,c) hit_tileElemAtNoStride(*tileR, 3, a, b, c)

	/* Variables*/

		/* Thread identification*/
		int tidx= threadIdx.x;
		int tidy= threadIdx.y;

		/* ThreadBlock identification*/
		int bidx= blockIdx.x;
		int bidy= blockIdx.y;

		/* Thread global positions*/
		int Gtidx= bidx*blockDim.x + tidx;	int Gcolumn	= Gtidx+1;
		int Gtidy= bidy*blockDim.y + tidy;	int Grow	= Gtidy+1;

		int fila= Gtidy;
		int columna= Gtidx;

	
	threadZoneTestDim3{
		
		/* Operation loop*/
		for(int ii=0; ii <  hit_sigCard(hit_shapeSig(dataShape,0)); ii++){
			int i=ii+1;
				/* Computations*/
        	            z(i,Grow,Gcolumn) = z(i,Grow,Gcolumn) + (
        	                + operatorS_Kernel0 * r(i,Grow,Gcolumn)
        	                + operatorS_Kernel1 * ( r(i-1,  Grow,  Gcolumn) + r(i+1,  Grow,  Gcolumn)
        	                                 + r(  i,Grow-1,  Gcolumn) + r(  i,Grow+1,  Gcolumn)
        	                                 + r(  i,  Grow,Gcolumn-1) + r(  i,  Grow,Gcolumn+1))
        	                + operatorS_Kernel2 * ( r(i-1,Grow-1,  Gcolumn) + r(i-1,Grow+1,  Gcolumn)
        	                                 + r(i-1,  Grow,Gcolumn-1) + r(i-1,  Grow,Gcolumn+1)
        	                                 + r(i+1,Grow-1,  Gcolumn) + r(i+1,Grow+1,  Gcolumn)
        	                                 + r(i+1,  Grow,Gcolumn-1) + r(i+1,  Grow,Gcolumn+1)
        	                                 + r(  i,Grow-1,Gcolumn-1) + r(  i,Grow-1,Gcolumn+1)
        	                                 + r(  i,Grow+1,Gcolumn-1) + r(  i,Grow+1,Gcolumn+1))
	
				);
		}// for
		
	}//threadZoneTest
	
}//__global__ psinv



/* resid MG kernel*/
kernelDefinitionCUDA(GPUFER,resid, HitTile_float  *tileU, HitTile_float *tileV, HitTile_float *tileR){
	
	/* Macros*/	

	#undef u
	#undef v
	#undef r
	#define u(a,b,c) hit_tileElemAtNoStride(*tileU,3,a,b,c)
	#define v(a,b,c) hit_tileElemAtNoStride(*tileV,3,a,b,c)
	#define r(a,b,c) hit_tileElemAtNoStride(*tileR,3,a,b,c)

	/* Variables*/

		/* Thread identification*/
		int tidx= threadIdx.x;
		int tidy= threadIdx.y;

		/* ThreadBlock identification*/
		int bidx= blockIdx.x;
		int bidy= blockIdx.y;

		/* Thread global positions*/
		int Gtidx= bidx*blockDim.x + tidx;	int Gcolumn	= Gtidx+1;
		int Gtidy= bidy*blockDim.y + tidy;	int Grow	= Gtidy+1;

		int fila= Gtidy;
		int columna= Gtidx;

	
	threadZoneTestDim3{
		
		/* Operation loop*/
		for(int ii=0; ii <  hit_sigCard(hit_shapeSig(dataShape,0)); ii++){
			int i=ii+1;
				/* Computations*/
        	        r(i,Grow,Gcolumn) = v(i,Grow,Gcolumn) - (
								+ operatorA_Kernel0 * u(i,Grow,Gcolumn)							
								+ operatorA_Kernel2 * ( u(i-1,Grow-1,  Gcolumn) + u(i-1,Grow+1,  Gcolumn)
										+ u(i-1,  Grow,Gcolumn-1) + u(i-1,  Grow,Gcolumn+1)
										+ u(i+1,Grow-1,  Gcolumn) + u(i+1,Grow+1,  Gcolumn)
										+ u(i+1,  Grow,Gcolumn-1) + u(i+1,  Grow,Gcolumn+1)
										+ u(  i,Grow-1,Gcolumn-1) + u(  i,Grow-1,Gcolumn+1)
										+ u(  i,Grow+1,Gcolumn-1) + u(  i,Grow+1,Gcolumn+1))
								+ operatorA_Kernel3 * ( u(i-1,Grow-1,Gcolumn-1) + u(i-1,Grow-1,Gcolumn+1)
										+ u(i-1,Grow+1,Gcolumn-1) + u(i-1,Grow+1,Gcolumn+1)
										+ u(i+1,Grow-1,Gcolumn-1) + u(i+1,Grow-1,Gcolumn+1)
										+ u(i+1,Grow+1,Gcolumn-1) + u(i+1,Grow+1,Gcolumn+1))
                    );
		}// for
		
	}//threadZoneTest	
	
}//__global__ resid



/* rprj3 MG kernel*/
kernelDefinitionCUDA(GPUFER,rprj3, HitTile_float  *tileRO, HitTile_float *tileR){

	/* Macros*/	

	#undef ro
	#undef r
	#define ro(a,b,c)	hit_tileElemAtNoStride(*tileRO,3,a,b,c)
	#define r(a,b,c)	hit_tileElemAtNoStride(*tileR,3,a,b,c)

	/* Variables*/

		/* Thread identification*/
		int tidx= threadIdx.x;
		int tidy= threadIdx.y;

		/* ThreadBlock identification*/
		int bidx= blockIdx.x;
		int bidy= blockIdx.y;

		/* Thread global positions*/
		int Gtidx= bidx*blockDim.x + tidx;	int Gcolumn	= Gtidx+1;
		int Gtidy= bidy*blockDim.y + tidy;	int Grow	= Gtidy+1;

		int fila= Gtidy;
		int columna= Gtidx;
		
		
	threadZoneTestDim3{
		
		/* Operation loop*/
		for(int ii=0; ii <  hit_sigCard(hit_shapeSig(dataShape,0))-2; ii++){
			
			int i=ii+1;
			int i1o = i * 2;
			int i2o = Grow * 2;
			int i3o = Gcolumn * 2;
				
			/* Computations*/
            r(i,Grow,Gcolumn) = (
                + operatorP_Kernel0 * ro(i1o,i2o,i3o)
                + operatorP_Kernel1 *
                    ( ro(i1o-1,  i2o,  i3o) + ro(i1o+1,  i2o,  i3o)
                    + ro(  i1o,i2o-1,  i3o) + ro(  i1o,i2o+1,  i3o)
                    + ro(  i1o,  i2o,i3o-1) + ro(  i1o,  i2o,i3o+1))
                + operatorP_Kernel2 *
                    ( ro(i1o-1,i2o-1,  i3o) + ro(i1o-1,i2o+1,  i3o)
                    + ro(i1o-1,  i2o,i3o-1) + ro(i1o-1,  i2o,i3o+1)
                    + ro(i1o+1,i2o-1,  i3o) + ro(i1o+1,i2o+1,  i3o)
                    + ro(i1o+1,  i2o,i3o-1) + ro(i1o+1,  i2o,i3o+1)
                    + ro(  i1o,i2o-1,i3o-1) + ro(  i1o,i2o-1,i3o+1)
                    + ro(  i1o,i2o+1,i3o-1) + ro(  i1o,i2o+1,i3o+1))
                + operatorP_Kernel3 *
                    ( ro(i1o-1,i2o-1,i3o-1) + ro(i1o-1,i2o-1,i3o+1)
                    + ro(i1o-1,i2o+1,i3o-1) + ro(i1o-1,i2o+1,i3o+1)
                    + ro(i1o+1,i2o-1,i3o-1) + ro(i1o+1,i2o-1,i3o+1)
                    + ro(i1o+1,i2o+1,i3o-1) + ro(i1o+1,i2o+1,i3o+1))
                );
        	            
        	            
        	            
		}// for

	}//threadZoneTest
		
}//__global__ rprj3


/* interp MG kernel*/
kernelDefinitionCUDA(GPUFER,interp, HitTile_float  *tileZO, HitTile_float *tileZ){
	
	
	/* Macros*/	

	#undef zo
	#undef z
	#define zo(a,b,c) hit_tileElemAtNoStride(*tileZO,3,a,b,c)
	#define z(a,b,c) hit_tileElemAtNoStride(*tileZ,3,a,b,c)

	/* Variables*/

		/* Thread identification*/
		int tidx= threadIdx.x;
		int tidy= threadIdx.y;

		/* ThreadBlock identification*/
		int bidx= blockIdx.x;
		int bidy= blockIdx.y;

		/* Thread global positions*/
		int Gtidx= bidx*blockDim.x + tidx;	int Gcolumn	= Gtidx;
		int Gtidy= bidy*blockDim.y + tidy;	int Grow	= Gtidy;

		int fila= Gtidy;
		int columna= Gtidx;


	threadZoneTestDim3{

		/* Operation loop*/
		for(int i1o=0; i1o <  hit_sigCard(hit_shapeSig(dataShape,0)); i1o++){  

			int i1 = i1o * 2;
			int i2 = Grow * 2;
			int i3 = Gcolumn * 2;
			int i2o= Grow;
			int i3o= Gcolumn;
				
			/* Computations*/
 
        	        z(i1,i2,i3) += operatorQ_Kernel0 * zo(  i1o,  i2o,  i3o);
                    z(i1,i2,i3+1) += operatorQ_Kernel1 *
                        (zo(  i1o,  i2o,i3o+1) + zo(  i1o,  i2o,  i3o));
                        
                    //    __syncthreads(); 
                        
                    z(i1,i2+1,i3) += operatorQ_Kernel1 * 
                    	( zo(  i1o,i2o+1,  i3o)   + zo(  i1o,  i2o,  i3o));
                    z(i1,i2+1,i3+1) += operatorQ_Kernel2 *
                        ( zo(  i1o,i2o+1,  i3o)   + zo(  i1o,  i2o,  i3o)
                        + zo(  i1o,i2o+1,  i3o+1) + zo(  i1o,  i2o,  i3o+1) );
        	            
        	      //      __syncthreads(); 

                    z(i1+1,i2,i3) += operatorQ_Kernel1 * 
                    	( zo(i1o+1,  i2o,  i3o)   + zo(  i1o,  i2o,  i3o));
                    z(i1+1,i2,i3+1) += operatorQ_Kernel2 *
                        ( zo(i1o+1,  i2o,  i3o)   + zo(  i1o,  i2o,  i3o)
                        + zo(i1o+1,  i2o,  i3o+1) + zo(  i1o,  i2o,  i3o+1) );

				//		__syncthreads(); 

                    z(i1+1,i2+1,i3) += operatorQ_Kernel2 * 
                    	( zo(i1o+1,i2o+1,  i3o)   + zo(i1o+1,  i2o,  i3o)
                    	+ zo(  i1o,i2o+1,  i3o)   + zo(  i1o,  i2o,  i3o));
                    z(i1+1,i2+1,i3+1) += operatorQ_Kernel3 *
                        ( zo(i1o+1,i2o+1,  i3o)   + zo(i1o+1,  i2o,  i3o)
                        + zo(  i1o,i2o+1,  i3o)   + zo(  i1o,  i2o,  i3o)
						+ zo(i1o+1,i2o+1,  i3o+1) + zo(i1o+1,  i2o,  i3o+1)
						+ zo(  i1o,i2o+1,  i3o+1) + zo(  i1o,  i2o,  i3o+1)) ;                        
        	            
        	            
		}// for

	}//threadZoneTest	
	
}//__global__ interp
