/* 
 * FFT: Sequential example
 * 		Code obtained at (last visit Mar 18th, 2013): 
 * 		http://rosettacode.org/wiki/Fast_Fourier_transform
 *
 * 		Modified to support a parametric size for the array, and dynamic allocation
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <complex.h>
 
double PI;
typedef double complex cplx;
 
void _fft(cplx buf[], cplx out[], int n, int step)
{
	int i;
	if (step < n) {
		_fft(out, buf, n, step * 2);
		_fft(out + step, buf + step, n, step * 2);
 
		for (i = 0; i < n; i += 2 * step) {
			cplx t = cexp(-I * PI * i / n) * out[i + step];
			buf[i / 2]     = out[i] + t;
			buf[(i + n)/2] = out[i] - t;
		}
	}
}
 
void fft(cplx buf[], int n)
{
	int i;
	//cplx out[n];
	cplx *out;
	out = (double complex *)malloc( n * sizeof(double complex) );
	for (i = 0; i < n; i++) out[i] = buf[i];
 
	_fft(buf, out, n, 1);
}
 
int main(int argc, char *argv[])
{
	if ( argc < 2 ) {
		fprintf(stderr, "Usage: %s <numElemsLog2>\n", argv[0] );
		exit(-1);
	}
	int numElemsLog2 = atoi( argv[1] );
	if ( numElemsLog2 < 1 || numElemsLog2 > 32 ) {
		fprintf(stderr, "Usage: %s <numElemsLog2>  ( parameter should be in the range [ 1, 32 ] )\n", argv[0] );
		exit(-1);
	}
	int numElems = 1 << numElemsLog2;

	//cplx buf[] = {1, 1, 1, 1, 0, 0, 0, 0};
	cplx *buf;
	buf = (double complex *)malloc( numElems * sizeof(double complex) );

	int k;
	for ( k=0; k<numElems/2; k++ ) buf[k] = 1;
	for ( k=numElems/2; k<numElems; k++ ) buf[k] = 0;

	PI = atan2(1, 1) * 4;
	int i;
 
	void show(const char * s) {
		printf("%s", s);
		//for (i = 0; i < 8; i++)
		for (i = 0; i < numElems; i++)
			if (!cimag(buf[i]))
				printf("%g ", creal(buf[i]));
			else
				printf("(%g, %g) ", creal(buf[i]), cimag(buf[i]));
		printf("\n");
	}
 
	//show("Data: ");
	//fft(buf, 8);
	fft(buf, numElems);
#ifdef WRITE_RESULT
	show("\nFFT : ");
#endif
 
	return 0;
}

