
/**
 * Program to test the access performance of the different tiles.
 *
 * @file matrix_sig_vs_bit.c
 * @author Javier Fresno
 */

/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2007-2013, Javier Fresno Bausela
 *
 * All Rights Reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <hitmap.h>
#include <hit_com.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <sys/time.h>


// Define the Tile_int
hit_tileNewType(int);

// Layout
HitLayout lay = HIT_LAYOUT_NULL_STATIC;


#define printfR(...) { if( hit_Rank == 0 ) { printf(__VA_ARGS__); fflush(stdout); }}

/**
 * Get time function
 * @return The current time.
 */
double gettime(){

	if(lay.type != 0){
		hit_comBarrier(lay);
	}

	struct timeval tv;
	gettimeofday(&tv, (void *) 0);
	return tv.tv_sec + 1.0e-6 * tv.tv_usec;
}

int main(int argc, char ** vargs){

	int nodes;

	if(argc != 2){
		printf("Usage: %s NUMBER_OF_NODES\n",vargs[0]);
		exit(0);
	}
	sscanf(vargs[1],"%d",&nodes);
	
	hit_comInit(&argc,&vargs);
	
	printfR("nodes: %d\n",nodes);
	
	printfR("Sig Shape\tSig Tile\tSTile NStride\tBitmap Shape\tGBTile iterator\tGBTile direct\n");
	
	int i,j;
	double t;
	
	// A.1 Create Sig Shape
	t = gettime();
	HitSig sig = hit_sig(0,nodes-1,1);
	HitShape shape = hit_shape(2,sig,sig);
	printfR("%f\t",gettime()-t);
	
	// A.2 Allocate and set zeros
	HitTile_int tile;
	hit_tileDomainShapeAlloc(&tile,int,shape);
	
	t = gettime();
	for(i=0;i<nodes;i++)
	for(j=0;j<nodes;j++){
		hit_tileElemAt(tile,2,i,j) = 0;
	}
	printfR("%f\t",gettime()-t);
	
	// A.3 set zeros no stride
	t = gettime();
	for(i=0;i<nodes;i++)
	for(j=0;j<nodes;j++){
		hit_tileElemAtNoStride(tile,2,i,j) = 0;
	}
	printfR("%f\t",gettime()-t);
	
	// A.3 Free
	hit_tileFree(tile);

	
	
	// B.1 Create Bit shape
	t = gettime();
	HitShape bshape = hit_bitmapShape(nodes);
	for(i=0;i<nodes;i++)
	for(j=0;j<nodes;j++){
		//hit_bShapeAddEdge(&bshape,i,j);
		hit_bShapeSet(bshape,i,j);
	}
	printfR("%f\t",gettime()-t);
	
	// B.2 Allocate and set zeros
	HitTile_int btile;
	hit_gbTileDomainShapeAlloc(&btile, int, bshape, HIT_EDGES);
	
	t = gettime();
	int vertex, edge;
	hit_bShapeVertexIterator(vertex,bshape){
		hit_bShapeEdgeIterator(edge,bshape,vertex){
		
			hit_gbTileEdgeIteratorAt(btile,vertex,edge) = 0;
		
		}
	}
	printfR("%f\t",gettime()-t);
	
	
	// B.3 set zeros without iterator
	t = gettime();
	for(i=0;i<nodes;i++)
	for(j=0;j<nodes;j++){
		hit_gbTileEdgeAt(btile,i,j) = 0;
	}
	printfR("%f\t",gettime()-t);
	
	
	// B.4 Free
	hit_tileFree(btile);
	
	printfR("\n");
	
	hit_comFinalize();
	
	return 0;
}














