
/*
 * Test Shape functions
 * @version 1.0
 * @date Feb 2012
 * @author Ana Moreton Fernandez
 */
#ifndef _HitDomain_
#define _HitDomain_

#include <stdarg.h>

#include <hit_error.h>
#include <hit_shape.h>


/* Hit DOMAIN */
/* 1. DOMAIN DATA STRUCTURE */
/**
 * A domain is a group of shapes
 *
 * @implements HitSigDomain
 */
typedef struct {
	int numShapes;// Number of shapes.
	HitShape *shape;// HitDomain objects to define each shape.
}HitSigDomain;


/**
 * Null value for domains.
 * The number of shapes of this value is 0.
 */
extern HitSigDomain HIT_DOMAIN_NULL;

/** Null static value for domain. */
#define HIT_DOMAIN_NULL_STATIC { 0, NULL }


 /* 1. Hit DOMAIN: UNION OPERATOR */
 /**
  * Join two shapes.
  * @warning It only works for HitSigShape, strides in all dimension must be equal 
  * @param sh1 shape.
  * @param sh2 shape.
  * @return HitSigDomain with the juction of sh1 and sh2. (s1 U s2)
  */
HitSigDomain hit_shapeUnion(HitShape sh1, HitShape sh2);


 /* 2. Hit DOMAIN: SUBSTRACT OPERATOR */
 /**
  * Substract two shapes.
  * @warning It only works for HitSigShape, strides in all dimension must be equal 
  * @param sh1 shape.
  * @param sh2 shape.
  * @return  HitSigDomain with the substract of sh1 and sh2.
  */
HitSigDomain hit_shapeSubtract(HitShape sh1, HitShape sh2);


 /* 3. Hit DOMAIN: JOIN TWO HITSIGDOMAIN */
 /**
  * Join two HitSigDomains in the first param.
  * @warning It only works for HitSigShape, strides in all dimension must be equal 
  * @param Domain1.
  * @param Domain2.
  */
void hit_JoinDomains(HitSigDomain *Domain1, HitSigDomain Domain2);


 /* 4. Hit DOMAIN: FREE HITSIGDOMAIN */

//hit_FreeDomain(HitSigDomain *Domain)

/* 6. HIT DOMAIN: SHORT THE SHAPES IN A DOMAIN */

//HitSigDomain hit_NormalizeDomain( HitSigDomain Domain);

#endif


