/**
 * 
 * Hitmap error macros. A set of macro-functions to show error and warnings.
 * 
 * @file hit_error.h
 * @version 1.2
 * @author Javier Fresno Bausela
 * @author Arturo Gonzalez-Escribano
 * @date Mar 2013
 * 
 */

/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2007-2013, Arturo Gonzalez Escribano
 *                          Javier Fresno Bausela
 *                          Yuri Torres de la Sierra
 * All Rights Reserved.
 */

#ifndef _HitError_
#define _HitError_

#include <stdio.h>

/* MACRO TO AVOID WARNINGS WHEN A PARAMETER IS NOT USED */
#define HIT_NOT_USED(x) ((void)(x))

/* ERROR CODES */
#define HIT_OK				0
#define HIT_ERROR			-1
#define HIT_ERR_INTERNAL	-10
#define HIT_ERR_USER		-100





/* INTERNAL ERROR AND WARNING MESSAGES */
#define hit_errInternal( routine, text, extraParam, file, numLine )	\
	{							\
	fprintf(stderr,"Hit RunTime-Error (%s): Internal - %s %s, used in %s[%d]\n", routine, text, extraParam, file, numLine); \
	exit( HIT_ERR_INTERNAL );		\
	}

#define hit_warnInternal( routine, text, extraParam, file, numLine )	\
	{							\
	fprintf(stderr,"Hit RunTime-Warning (%s): Internal - %s %s, used in %s[%d]\n", routine, text, extraParam, file, numLine); \
	}


/* ERROR AND WARNING MESSAGES FOR PROGRAMMERS/USERS */
#define hit_error( name, file, numLine )	\
	{							\
	fprintf(stderr,"Hit Programmer, RunTime-Error: %s, in %s[%d]\n", name, file, numLine); \
	exit( HIT_ERR_USER );		\
	}

#define hit_warning( name, file, numLine )	\
	{							\
	fprintf(stderr,"Hit Programmer, RunTime-Warning: %s, in %s[%d]\n", name, file, numLine); \
	}

#define hit_error_here(name)	hit_error(name, __FILE__, __LINE__)
#define hit_warning_here(name)	hit_warning(name, __FILE__, __LINE__)


#endif

/* END OF HEADER FILE _HitError_ */
