

#ifndef _HitTask_
#define _HitTask_

#include <stdlib.h>


/**
 * A task to be executed by a worker.
 */
typedef struct {
	int		n_elems;	/**< Number of elements in the task. */
	size_t	baseExtend;	/**< Size of each single elements. */
	double	weight;		/**< Weight of the task, used by the scheduler */
	void	*data;		/**< The pointer to the data */
} HitTask;


/**
 * Null variable for the HitTask
 */
extern HitTask HIT_TASK_NULL;
#define HIT_TASK_NULL_STATIC {0,0,0,NULL}



/**
 * Macro to create new task types.
 */
#define hit_taskNewType(newType)			\
typedef struct HitTask_##newType {		\
	int					n_elems;			\
	size_t				baseExtend;			\
	double				weight;				\
	newType				*data;				\
} HitTask_##newType, *HitPTask_##newType

/**
 * Allocate memory for the task.
 * @param newTaskP Task pointer.
 * @param n_elems Number of elements in the task.
 * @param baseExtent Size of the task data.
 */
void hit_taskAllocInternal(void * newTaskP, int n_elems, size_t baseExtent);

/**
 * Allocate memory for a task with a single element.
 * @param newTaskP Task pointer.
 * @param baseType Task type.
 */
#define hit_taskAlloc(newTaskP, baseType) hit_taskAllocInternal((newTaskP), 1, sizeof(baseType))

/**
 * Allocate memory for an array task.
 * @param newTaskP Task pointer.
 * @param n_elems Number of elements in the task.
 * @param baseType Task type.
 */
#define hit_taskArrayAlloc(newTaskP, n_elems, baseType) hit_taskAllocInternal((newTaskP), (n_elems), sizeof(baseType))


/**
 * Frees the task memory.
 * @param newTaskP Task pointer.
 */
void hit_taskFree(void * newTaskP);

/**
 * Macro to get the task data.
 */
#define hit_taskData(task) (hit_taskArrayData(task,0))

/**
 * Macro to get the task data.
 */
#define hit_taskArrayData(task,elem) ((task).data[(elem)])

/**
 * Macro to get the task weight.
 */
#define hit_taskWeight(task) ((task).weight)

/**
 * Macro to get the task weight.
 */
#define hit_taskNElements(task) ((task).n_elems)

/**
 * Macro to test if the task is null.
 */
#define hit_taskIsNull(task) ((task).n_elems==0)




#endif

