/**
 * HitMap library main include file.
 * It includes all the header files needed to use all the functionalities of HitMap.
 *
 * @file hitmap.h
 * @version 1.1.1
 * @author Arturo Gonzalez-Escribano
 * @author Javier Fresno Bausela
 * @author Carlos de Blas Carton
 * @date Feb 2013
 *
 */

/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2007-2013, Arturo Gonzalez Escribano
 *                          Javier Fresno Bausela
 *                          Carlos de Blas Carton
 *                          Yuri Torres de la Sierra
 * All Rights Reserved.
 */



/**
 * @mainpage
 * Hitmap is a highly-efficient library for hierarchical tiling and mapping of
 * arrays. It is designed to be used with an SPMD (Single Process, Multiple Data)
 * parallel programming language and it aims to simplify parallel programming,
 * providing functionalities to create, manipulate, distribute, and communicate
 * tiles and hierarchies of tiles.
 * 
 * Hitmap library supports functionalities to:
 * - Generate a virtual topology structure arranging the available physical
 *   processors.
 * - mapping the data to the different processor with chosen partitioning and
 *   load-balancing techniques.
 * - automatically determine inactive processors at any stage of the computation,
 *   preventing the active processors may be affected.
 * - identify the neighbor processors to use in point-to-point or collective
 *   communications.
 * - build communication patterns to be reused across algorithm iterations.
 * 
 * The library functionalities can be classified in three sets:
 * - Tiling functions that contains the definition and manipulation of tiles.
 * - Mapping functions for data distribution and layout functions to automatically
 *   part domains, depending on the virtual topology selected.
 * - Communication functions for create reusable communication patterns for
 *   distributed hierarchical tiles. The following sections contain more detailed
 *   information about the Hitmap library, they explain its architecture and
 *   functionalities.
 * 
 * @section Hitmap Architecture
 * Hitmap is designed with an object-oriented approach, although it is
 * implemented in C language.
 */



#include "hit_funcop.h"

#include "hit_topology.h"
#include "hit_layout.h"
#include "hit_topContext.h"
#include "hit_pattern.h"

#include "hit_shape.h"
#include "hit_cshape.h"
#include "hit_bshape.h"

#include "hit_dump.h"

#include "hit_tile.h"
#include "hit_gctile.h"
#include "hit_gbtile.h"
#include "hit_mctile.h"
#include "hit_mbtile.h"


#include "hit_utils.h"
#include "hit_file.h"

#include "hit_blockTile.h"
