
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <cutil.h>
#include "acelNVIDIA.h"

/* A. To allocate memory on device (NVIDIA)*/
extern"C"
void acelDeclDataNVIDIA (HitTile *tile, HitTile **acelTile, void **acelData, size_t typeSize){

	/* Temporal variable*/
	//HitTile tmp= *tile;
	
	/* cudaMalloc for temporal variable*/
	//CUDA_SAFE_CALL( cudaMalloc((void**)&(tmp.data), (tile)->baseExtent*hit_tileCard(*tile)) );
	CUDA_SAFE_CALL( cudaMalloc((void**)&(tile->data), (tile)->baseExtent*hit_tileCard(*tile)) );
		
	/* Pointer to temporal variable*/
	//*acelData = tmp.data;
	*acelData = tile->data;
	
	/* cudaMalloc fro acelerator tile*/
	CUDA_SAFE_CALL( cudaMalloc(acelTile, sizeof(HitTile)) );
	
	/* Copying data to device memory (acelerator tile)*/
	//CUDA_SAFE_CALL( cudaMemcpy(*acelTile, &tmp, sizeof(HitTile),cudaMemcpyHostToDevice) );
	CUDA_SAFE_CALL( cudaMemcpy(*acelTile, tile, sizeof(HitTile),cudaMemcpyHostToDevice) );
					
}//acelDeclDataNVIDIA


/* B. To send data from Host to device (NVIDIA)*/		
extern"C" 
void acelSendDataNVIDIA (HitTile *tile, void *acelData){

	/* Variables*/
	int i,j;
	HitTile ancestor= *(HitTile *)(tile->ref);

	/* Two loops to copy the last two dimensions from Host to GPU device*/
	for (i=0; i<hit_tileDimCard(ancestor,0); i++){

		for (j=0; j<hit_tileDimCard(ancestor,1); j++){

			CUDA_SAFE_CALL( cudaMemcpy(
					(char *)acelData+(size_t)tile->baseExtent*(i*hit_tileDimCard(*tile,1)+j)*hit_tileDimCard(*tile,2), 
					(char *)ancestor.data+(size_t)ancestor.baseExtent*(i*hit_tileDimCard(ancestor,1)+j)*hit_tileDimCard(ancestor,2), 
					(tile)->baseExtent*hit_tileDimCard(ancestor,2),cudaMemcpyHostToDevice) );

		}//for j

	}//for i


	/* One loop to copy the last dimension from Host to GPU device*/
	//for (i=0; i<hit_tileDimCard(ancestor,0); i++){
	//	CUDA_SAFE_CALL( cudaMemcpy(
	//		(char *)acelData+(size_t)i*tile->baseExtent*hit_tileDimCard(*tile,1), 
	//			(char *)ancestor.data+(size_t)i*ancestor.baseExtent*hit_tileDimCard(ancestor,1), 
	//			(tile)->baseExtent*hit_tileDimCard(ancestor,1),cudaMemcpyHostToDevice) );
	//}//for 
	

	/* Copying data from the host tile variable to acelerator tile variable*/
	//CUDA_SAFE_CALL( cudaMemcpy(acelData, tile->data, (tile)->baseExtent*hit_tileCard(*tile),cudaMemcpyHostToDevice) );


}//acelSendDataNVIDIA


/* C. To receive data from device to Host (NVIDIA)*/
extern"C"
void acelRecvDataNVIDIA (HitTile *tile, void *acelData){

	/* Variables*/
	int i,j;
	HitTile ancestor= *(HitTile *)(tile->ref);


	/* Two loops to copy the last two dimensions from Host to GPU device*/
	for (i=0; i<hit_tileDimCard(ancestor,0); i++){

		for (j=0; j<hit_tileDimCard(ancestor,1); j++){

				CUDA_SAFE_CALL( cudaMemcpy(
					(char *)ancestor.data+(size_t)ancestor.baseExtent*(i*hit_tileDimCard(ancestor,1)+j)*hit_tileDimCard(ancestor,2),
					(char *)acelData+(size_t)tile->baseExtent*(i*hit_tileDimCard(*tile,1)+j)*hit_tileDimCard(*tile,2), 
					(tile)->baseExtent*hit_tileDimCard(ancestor,2),cudaMemcpyDeviceToHost) );
		}//for j

	}//for i


	/* One loop to copy the last dimension from GPU to Host device*/
	//	for (i=0; i<hit_tileDimCard(ancestor,0); i++){
	//		CUDA_SAFE_CALL( cudaMemcpy(
	//				(char *)ancestor.data+(size_t)i*ancestor.baseExtent*hit_tileDimCard(ancestor,1),
	//				(char *)acelData+(size_t)i*tile->baseExtent*hit_tileDimCard(*tile,1), 
	//				(tile)->baseExtent*hit_tileDimCard(ancestor,1),cudaMemcpyDeviceToHost) );		
	//
	//	}//for 
	

	/* Copying data from the host tile variable to acelerator tile variable*/
	//CUDA_SAFE_CALL( cudaMemcpy(tile->data, acelData, tile->baseExtent*hit_tileCard(*tile),cudaMemcpyDeviceToHost) );
	
}//acelRecvDataNVIDIA


/* 
 * 
 * D. To free the hitTile variables on the acelerator (NVIDIA)
 * 
 * */
extern"C"
void acelFreeNVIDIA (HitTile *tile){

	/* Freeing memory*/		
	CUDA_SAFE_CALL( cudaFree(tile) );

}//acelFreeNVIDIA



/* 
 * 
 * E. To copy the data from/to Host and Gpu devices
 * 
 * */
extern"C" 
void acelCopyDataFromToNVIDIA ( void *host, void *gpu, size_t size, int target ){

	/* The target device is the GPU*/
	if ( target == HIT_UPDATE_FROM ) {
		
		CUDA_SAFE_CALL(cudaMemcpy( gpu, host, size, cudaMemcpyHostToDevice ) );
		
	} //if
	
	/* The target device is the Host*/
	else {
		
		CUDA_SAFE_CALL( cudaMemcpy( host, gpu, size, cudaMemcpyDeviceToHost ) );
		
	} //else

} //acelCopyDataFromToNVIDIA


