

#include <hit_tile.h>
#include <hit_acel.h>
 
;

/*  
 * 
 * GPUs properties are stored in the HIT_TOPOLOGY_INFO.GPUInf fields 
 * 
 * */
extern "C"
void get_GPU_pro(void) {
	
	/* Local variables*/
	int i,j;

	cudaDeviceProp deviceProp;

	/* Getting the number of GPUs in the system*/
	cudaGetDeviceCount(&HIT_GPUs_TOPOLOGY_INFO.numGPUs);

	/* For each GPU device in the system... storing: numSMs, numCores, maxThreadsPerBlock*/
	for (i=0; i < HIT_GPUs_TOPOLOGY_INFO.numGPUs ;i++){

		/* Reading the ith property*/
		cudaGetDeviceProperties(&deviceProp, i);

		/* Storing numSMs, maxThreadsPerBlock and GPU internal name*/
		HIT_GPUs_TOPOLOGY_INFO.GPUInf[i].numSMs=deviceProp.multiProcessorCount;
		HIT_GPUs_TOPOLOGY_INFO.GPUInf[i].maxThreadsPerBlock=deviceProp.maxThreadsPerBlock;
		HIT_GPUs_TOPOLOGY_INFO.GPUInf[i].internalName=i;	//redundant line because de arrays position indicates the GPU name (is the same number)

		HIT_GPUs_TOPOLOGY_INFO.GPUInf[i].major=deviceProp.major;		//Release number
		HIT_GPUs_TOPOLOGY_INFO.GPUInf[i].minor=deviceProp.minor;		//Release Number

		/* Storing the maximun number of block per Grid dimension*/
		for (j=0; j < MAX_NUM_GRID_DIMENSIONS; j++)
			HIT_GPUs_TOPOLOGY_INFO.GPUInf[i].maxBlocksPerDimension[j]= deviceProp.maxGridSize[j];

		/* Storing the maximun number of threads per ThreadBlock dimension*/
		for (j=0; j < MAX_NUM_BLOCK_DIMENSIONS; j++)
			HIT_GPUs_TOPOLOGY_INFO.GPUInf[i].maxThreadsPerDimension[j]= deviceProp.maxThreadsDim[j];
	
		/* Depending of the underlying architecture (Fermi-1024 threads per block
		*  or pre-Fermi with 512 threads per block )*/			
		if (deviceProp.maxThreadsPerBlock==MAX_THREAD_PER_BLOCK_FERMI){ 
			HIT_GPUs_TOPOLOGY_INFO.GPUInf[i].numCores=deviceProp.multiProcessorCount*
							NUMBER_OF_CORES_PER_SM_FERMI;			//32 is the number of cores per SMs
			HIT_GPUs_TOPOLOGY_INFO.GPUInf[i].maxThreadsPerSM=MAX_THREADS_PER_SM_FERMI;		//maximum number of threads per SM	 
		}
		else {
			HIT_GPUs_TOPOLOGY_INFO.GPUInf[i].numCores=deviceProp.multiProcessorCount*
							8;		//8 is the number of cores per SMs
			HIT_GPUs_TOPOLOGY_INFO.GPUInf[i].maxThreadsPerSM=MAX_THREADS_PER_SM_PRE_FERMI;	//maximum number of threads per SM	  
		}

	}//for, for each GPU device

}//get_GPU_pro


/* 
 * 
 * To select a specific GPU device
 * 
 * */
extern "C"
void useDevice(int gpuIdentifier) {

	/* cudaSetDevice primitive*/	
	cudaSetDevice(gpuIdentifier);

}//useDevice



/* 
 * 
 * To finalize the threads
 * 
 * */
extern "C"
void freeTileCuda(HitTile *tile, void *acelData, HitTile *tilePadding) {

	/* Releasing the memory*/	
	cudaFree(tile);
	cudaFree(acelData);  
	cudaFree(tilePadding);

}//threadExit



/* 
 * 
 * To finalize the threads
 * 
 * */
extern "C"
void threadExitCuda(void) {

	/* cudaSetDevice primitive*/	
	cudaThreadExit();

}//threadExit

