/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2011-2012, Yuri Torres de la Sierra
 *			    Arturo Gonzalez Escribano
 *                          
 * All Rights Reserved.
 */


#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>

#include <hit_partition.h>


/* 
 * 
 * 
 * Initializing the HitPartition struct with values by default
 * 
 * 
 * */
#define	HIT_PARTITION_NULL_STATIC	{ HIT_SHAPE_NULL_STATIC, HIT_SHAPE_NULL_STATIC, HIT_SHAPE_NULL_STATIC, HIT_SHAPE_NULL_STATIC, HIT_SHAPE_NULL_STATIC, HIT_SHAPE_NULL_STATIC }


/*
 * 
 *  A. Naive partition
 * 
 * */
HitPartition hit_partition_plug_partNaive (HitShape dataShape, int n, ...){

	/* Variables*/

		int i;		
	
		/* To read the arguments*/
		va_list ap;					

		/* Result variable*/		
		HitPartition res= HIT_PARTITION_NULL_STATIC;

		/* Block shape variables*/
		HitSig blockSig_0;
		HitSig blockSig_1;
		HitSig blockSig_2;



	/* Reading the threads access pattern*/
	va_start(ap, n);
	for (i=0; i<n; i++){

		res.accessPattern[i]=va_arg(ap, HitShape);

	}//for i
	va_end(ap);


	/* Creating the threadBlock signatures (initial threadBlock shape) per dimension */
	
	if ( hit_sigCard(hit_shapeSig(dataShape,hit_shapeSigDims(dataShape)-1)) <=MAX_THREAD_PER_BLOCK_PRE_FERMI){
		
		/*Dimension x (rows); number of data elements in the dataShape last dimension */
		blockSig_0= hit_sigStd(hit_sigCard(hit_shapeSig(dataShape,hit_shapeSigDims(dataShape)-1)));
	
	}
	else{
		
		blockSig_0= hit_sigStd(MAX_THREAD_PER_BLOCK_PRE_FERMI);	//Dimension x (rows)	//512threads
	
	}
	
	blockSig_1= hit_sigStd(1);				//Dimension y	//1 thread
	blockSig_2= hit_sigStd(1);				//Dimension z	//1 thread


		/* Creating the threadBlock shape*/
		res.blockShape= hit_shape3(blockSig_0, blockSig_1, blockSig_2);

		/* The Grid shape initialization*/

		/* Second dimension (rows y); number fo grid rows== number of data rows*/
		//hit_shapeSig(res.gridShape, GPU_DIM_ROWS)= hit_sigStd(hit_sigCard(hit_shapeSig(dataShape,0)));

		/* Firs dimenson (colums x); ceil use to include the padding*/
		//hit_shapeSig(res.gridShape, GPU_DIM_COLUMS)= hit_sigStd((int)  ceil( (double)hit_sigCard(hit_shapeSig(dataShape,1))/(hit_sigCard(blockSig_0))));

		/* Second dimension (rows y); number fo grid rows== number of data rows*/
		hit_shapeSig(res.gridShape, GPU_DIM_ROWS)= hit_sigStd(hit_sigCard(hit_shapeSig(dataShape,(hit_shapeSigDims(dataShape)-2))));

		/* Firs dimenson (colums x); ceil use to include the padding*/
		hit_shapeSig(res.gridShape, GPU_DIM_COLUMS)= hit_sigStd((int)  ceil( (double)hit_sigCard(hit_shapeSig(dataShape,(hit_shapeSigDims(dataShape)-1)))/(hit_sigCard(blockSig_0))));

		/* Third dimension (depth z); a single unit*/
		hit_shapeSig(res.gridShape, GPU_DIM_DEPTH)= hit_sigStd(1);



	/* Canculating and initializing the data cardinality (data + padding)*/
	res.paddingDataShape= hit_shape3( hit_sigStd(hit_sigCard(hit_shapeSig(dataShape,0))),
			hit_sigStd(hit_sigCard(hit_shapeSig(res.gridShape,GPU_DIM_ROWS))),
					hit_sigStd (hit_sigCard(hit_shapeSig(res.gridShape,GPU_DIM_COLUMS) )*hit_sigCard(blockSig_0)));


	/* Copying the original data shape*/
	res.origDataShape= dataShape;

	/* Copying the data portion with memory access pattern (Padding data + accessPatter)*/
	res.origDataPatternShape=hit_partitionDataPaddingPatern(dataShape, res, res.accessPattern[0]);

	/* Final instruction*/
	return res;

}//hit_partition_plug_partNaive



/* 
 * 
 * Function to calculate the data padding depending on the global memory access pattern
 * 
 * */
HitShape hit_partitionDataPaddingPaternInternal (HitShape dataShape, HitPartition partition, HitShape accessPattern){

	/* Variables*/
	HitShape res= dataShape;						//result
	int i,j;
	int numDims= hit_shapeSigDims(accessPattern);	//Number of dims in the accessPattern shape

	/* Loop to calculate per accessPatter shape dimension  the data padding*/
	for(i=0; i < numDims; i++) {

		/* For each global memory access pattern dimension its stride*/		
		int stride= hit_shapeSig(accessPattern,i).stride;

		/* The different cases of strides*/
		switch(stride){
			case 1:		//if ==1 then both signatures are added
				hit_shapeSig(res,i)= hit_sigStd(hit_sigCard(hit_sigExpand(hit_shapeSig(accessPattern,i), hit_shapeSig(dataShape,i))));
			break;
			
			case -1:	//if ==-1 we take the original signature
				hit_shapeSig(res,i)= hit_shapeSig(dataShape,i);
			break;

			default:
				fprintf(stderr,"Error in data shape padding\n");
				exit(-1);
			break;

		}//switch

	}//for

	/* Final instruction*/
	return res;

}//hit_partitionDataPaddingPaternInternal



/* 
 * 
 * Function to calculate a global shape including the computation and data shapes
 * 
 * */
HitShape hit_partitionExtendedDataPaddingPaternInternal(HitTile *tile, HitPartition partition){

	/* Variables*/
	HitShape res= HIT_SHAPE_NULL;
	int i;
	
	/* For the number of the tile shape dimensions*/
	for (i=0; i < hit_shapeSigDims(tile->shape); i++){

		/* We extend the signature*/
		hit_shapeSig(res,i).begin	= (hit_shapeSig(tile->shape,i).begin <= hit_shapeSig(partition.paddingDataShape,i).begin)? hit_shapeSig(tile->shape,i).begin : hit_shapeSig(partition.paddingDataShape,i).begin;
		hit_shapeSig(res,i).end		=  (hit_shapeSig(tile->shape,i).end >= hit_shapeSig(partition.paddingDataShape,i).end)? hit_shapeSig(tile->shape,i).end : hit_shapeSig(partition.paddingDataShape,i).end-1;
		hit_shapeSig(res,i).stride	= hit_shapeSig(tile->shape,i).stride;

	}//for

	/* The number of data dimensions*/
	hit_shapeSigDims(res)=hit_shapeSigDims(tile->shape);
	
	return res;
}//hit_partitionExtendedDataPaddingPaternInternal





////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////								////////////////////////////////////////////////
/////////////////////////////		 OLD CODE ABOUT PARTITION PLUG-INS		////////////////////////////////////////////////
//////////////////////////////								////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#ifdef nunca
/* A. PLUGIN TO MANUALLY INTRODUCE THE THREADBLOCK SHAPE (also chechs the the values)*/
HitPartition hit_partition_plug_partBlocksFixedSize ( HitShape dataShape, HitShape blockShape){

	/* Variables*/
		HitPartition res= HIT_PARTITION_NULL_STATIC;	//Initializating the result variable
		
//		/* Creating new pointer from the tile parametres*/				
//		HitTile *newVarA = (HitTile *)A;
//		HitTile *newVarB = (HitTile *)B;
//		HitTile *newVarC = (HitTile *)C;

		/* The signatures of each block dimension*/
		HitSig blockSig_0;
		HitSig blockSig_1;
		HitSig blockSig_2;


	/* Creating the threadBlock signatures (initial threadBlock shape) per dimension */
	blockSig_0= hit_sig( 0, hit_sigCard(hit_shapeSig(blockShape,0))-1, 1);		//Dimension x
	blockSig_1= hit_sig( 0, hit_sigCard(hit_shapeSig(blockShape,1))-1, 1);		//Dimension y
	blockSig_2= hit_sig( 0, hit_sigCard(hit_shapeSig(blockShape,2))-1, 1);		//Dimension z
	
	/* Testing and putting the threadBlock and gid size and shape*/
	hit_partition_set_GPU_shapes(&res, dataShape, blockSig_0, blockSig_1, blockSig_2);


	/* Check point*/
//	printf("B0 %d  B1 %d  B2 %d\n\n",hit_sigCard(hit_shapeSig(res.blockShape,0)),
//					hit_sigCard(hit_shapeSig(res.blockShape,1)),hit_sigCard(hit_shapeSig(res.blockShape,2)));
//	printf("G0 %d  G1 %d\n\n",hit_sigCard(hit_shapeSig(res.gridShape,0)),hit_sigCard(hit_shapeSig(res.gridShape,1)));
//	printf("P0 %d  P1 %d\n\n",hit_sigCard(hit_shapeSig(res.gridPadding,0)),hit_sigCard(hit_shapeSig(res.gridPadding,1)));



	/* Final instruction*/
	return res;

}//hit_partition_plug_BlocksFixedSize




/* B. PLUGIN TO CREATE THE THREADBLOCK SHAPE FROM THREAD ACCESS PATTERN*/
HitPartition hit_partition_plug_partBlocksPattern (HitShape dataShape, int num_threadPatterns, ...){

	/*Variables*/
		/* Normal*/
		va_list ap;					//To read the arguments
		int i, j, maxthreads;
		//int numThreads[num_threadPatterns];
		HitShape patternAux[num_threadPatterns];	//To store the intermediate threadBlock shape
		HitSig  sig_neutral =	HIT_SIG_ZERO;		//To create a neutral signature for the partition function
	
		/*HitPartition*/
		HitPartition res= HIT_PARTITION_NULL_STATIC;	//Initializating the result variable

	/* Check point*/
//	fprintf(stderr,"There are %d different threadBlock access patterns\n\n", num_threadPatterns);

	/* Start the ... parameters*/
	va_start(ap, num_threadPatterns);

		/* Loop: for each threadBlock access pattern*/			//Currently 3 different access patterns, one for each matrix
		for (i=0; i<num_threadPatterns; i++){
			
			/* Reading each ... parameter*/
			patternAux[i]= va_arg(ap, HitShape);
			
			//numThreads[i]=0;					//0-initialize of the number of threads 

			/* Loop: for each threadBlock dimension (in this case for 2 dimension; two-dimensional arrays)*/
			for (j=0; j <MAX_NUM_BLOCK_DIMENSIONS -1; j++){
				
				/* If the thread access pattern in this dimension is HIT_SIG_ONE_POS_IDENTIFIER*/
				if (hit_sig_stride(hit_shapeSig(patternAux[i],j))==HIT_SIG_ONE_POS_IDENTIFIER){
					
					/* 1:1*/
					maxthreads=1;
					hit_shapeSig(patternAux[i],j)= hit_sigStd(maxthreads);

				}//else
				else {
					/* 1:n horizontal shape*/
					if (HIT_TOPOLOGY_INFO.selfRank==0)
						maxthreads= min( hit_sigCard(hit_shapeSig(dataShape,j)), 512);  //(max_threads_per_block(FERMI_GPU)
					else 
						maxthreads= min( hit_sigCard(hit_shapeSig(dataShape,j)), 512);

					if(maxthreads<=0){maxthreads=1;} //if there is not threads then we put one (not to divide by 0)
					hit_shapeSig(patternAux[i],j)= hit_sigStd(512);  //hit_sigStd(maxthreads);

				
				}//else
				
				/* To store the number of threads per access pattern and after select one of them*/
				/* to have a criterion to select the global access pattern*/	
				//numThreads[i]= numThreads[i]*maxthreads;
			}//j		

			/* Putting and checking the threadBlock and Grid size and shape*/ //a camiar en su momento para que no se acumulen los valores		
			//hit_partition_set_GPU_shapes ( &res, dataShape, hit_shapeSig(patternAux[i],0), hit_shapeSig(patternAux[i],1), sig_neutral);
		
		}//for i		

	/* Finish the ... parameters */
	va_end(ap);
			

	/* Extends the threadBlock to reach as far as possible the maximun size*/			
	patternAux[0] =hit_partition_extend_block(dataShape, patternAux[0]);	



	/* Calculating the final partition (we select the first access pattern to calculate the threadBlock shape)*/
	hit_partition_set_GPU_shapes ( &res, dataShape, hit_shapeSig(patternAux[0],0), hit_shapeSig(patternAux[0],1), hit_sig(0, 0, 1));

	/* Check point*/
//fprintf(stderr,"partitionb %d %d %d\n", hit_sigCard(hit_shapeSig(res.blockShape,0)), hit_sigCard(hit_shapeSig(res.blockShape,1)),hit_sigCard(hit_shapeSig(res.blockShape,2)));
//fprintf(stderr,"partitiong %d %d %d\n", hit_sigCard(hit_shapeSig(res.gridTotal,0)), hit_sigCard(hit_shapeSig(res.gridTotal,1)),hit_sigCard(hit_shapeSig(res.gridTotal,2)));

/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////

		HitSig blockSig_0;
		HitSig blockSig_1;
		HitSig blockSig_2;


	/* Creating the threadBlock signatures (initial threadBlock shape) per dimension */
	blockSig_0= hit_sig( 0, 511, 1);		//Dimension x
	blockSig_1= hit_sig( 0, 0, 1);			//Dimension y
	blockSig_2= hit_sig( 0, 0, 1);			//Dimension z
		
	res.blockShape= hit_shape3(blockSig_0, blockSig_1, blockSig_2);

	hit_shapeSig(res.gridTotal, 0)= hit_sig(0,((int)  ceil( (double)hit_sigCard(hit_shapeSig(dataShape,1))/512) )-1,1);
	hit_shapeSig(res.gridTotal, 1)= hit_sig(0,hit_sigCard(hit_shapeSig(dataShape,0))-1,1); 
	hit_shapeSig(res.gridTotal, 2)= hit_sig(0,0,1);

	hit_shapeSig(res.currentDataShape, 1)=hit_sigStd (hit_sigCard(   hit_shapeSig(res.gridTotal,0) )*512); 
	hit_shapeSig(res.currentDataShape, 0)= hit_sigStd(hit_sigCard(   hit_shapeSig(res.gridTotal,1) ));

	res.currentDataShape.info.sig.numDims=2;

/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////
/////////////////////////////////////////

	/* Final instruction*/
	return res;

}//hit_partition_plug_BlocksPattern



/* Extends the threadBlock to reach as far as possible the maximun occupancy*/			
HitShape hit_partition_extend_block(HitShape dataShape, HitShape threadBlockShape){

	/* Variables*/
	int i,threads;					//iterator
	int maxThreads= 1;			//The number of threads in the threadBlockShape
	HitShape res= threadBlockShape;		//Result variable


	/* Loop: Calculating the number of thread in the shape*/
	for (i=0; i<MAX_NUM_BLOCK_DIMENSIONS-1; i++){
		
		/* Multiplying the number of thread per dimension*/
		maxThreads=  maxThreads *hit_sigCard(hit_shapeSig(threadBlockShape,i));

	}//for i

	/* Loop: for each dimension we expand the threadBlock shape to reach a big block*/
	for (i=0; i<MAX_NUM_BLOCK_DIMENSIONS-1; i++){

		/* If there is a single thread in this dimension then.....*/
		if(  (hit_sigCard(hit_shapeSig(threadBlockShape,i))) == 1){
			/* The maximum number of thread for this dimension*/
			if (HIT_TOPOLOGY_INFO.selfRank==0)
				threads= min( (512/maxThreads) , (hit_sigCard(hit_shapeSig(dataShape,i)))  ); //(max_threads_per_block(FERMI_GPU)
			else
				threads= min( (512/maxThreads) , (hit_sigCard(hit_shapeSig(dataShape,i)))  ); 

			hit_shapeSig(res,i)=hit_sig(0, (threads-1),1);		//storing the new number of threads for the current dimension	
			maxThreads=maxThreads*threads;				//Recalculating the curren number of threads
		}//if

	}//for i

	/* Check point*/
//	fprintf(stderr,"Numero de hilos acutales en el bloque. %d %d %d\n", (hit_sigCard(hit_shapeSig(res,0))), 
//						(hit_sigCard(hit_shapeSig(res,1))),1);

	/* Final instruction*/
	return res;

}//hit_partition_extend_block



/* SET GRID AND THREADBLOCK SHAPE MANUALLY*/					//yuri
//int hit_partition_set_GPU_shapes (HitLayout *lay, HitShape shape, HitSig bx, HitSig by, HitSig bz){
int hit_partition_set_GPU_shapes (HitPartition *part, HitShape shape, HitSig bx, HitSig by, HitSig bz){

	/* Variables*/
		int num_gpu;
		if (HIT_TOPOLOGY_INFO.selfRank==1)
			num_gpu=1;
		else
			num_gpu=0;
						
		/* Number of blocks per dimension in the grid*/
		HitSig gx;

		if (hit_sigCard2(shape.info.sig.sig[1])< hit_sigCard(bx))
			gx = hit_sig(0,0,1);
		else
			gx = hit_sig(0,(hit_sigCard2(shape.info.sig.sig[1])/hit_sigCard(bx))-1,1); 		//Grid x-dimension signature
		HitSig gy = hit_sig(0,hit_sigCard2(shape.info.sig.sig[0])/hit_sigCard(by)-1,1);		//Grid y-dimension signature
//		HitSig gy = hit_sig(0,(hit_sigCard2(shape.info.sig.sig[1])/hit_sigCard(by))-1,1);





		/* Number of extra thread needed*/ //*****************************************************************
		HitSig byFringe = hit_sig(0,(hit_sigCard(shape.info.sig.sig[0])%hit_sigCard(bx)-1),1); //Fringe block x-dimension signature //columnas de verdad
		HitSig bxFringe = hit_sig(0,(hit_sigCard(shape.info.sig.sig[1])%hit_sigCard(by)-1),1);		//Fringe block y-dimension signature


//if (HIT_TOPOLOGY_INFO.selfRank==2){
//
//printf("blockes %d %d %d\n",bx.begin,bx.end,bx.stride);
//printf("blockes %d %d %d\n\n",bx.begin,bx.end,bx.stride);
//
//printf("HIlos extra %d %d \n",hit_sigCard(shape.info.sig.sig[0]),hit_sigCard(shape.info.sig.sig[1]));
//printf("HIlos extra %d %d \n",hit_sigCard(byFringe),hit_sigCard(bxFringe));


//}

				//		printf("signature cards %d %d\n",hit_sigCard2(shape.info.sig.sig[0]),hit_sigCard2(shape.info.sig.sig[1]));
//						printf("signature  gx %d %d\n",hit_sigCard(shape.info.sig.sig[0]),hit_sigCard(shape.info.sig.sig[1]));
				//printf("signature bx  %d %d\n",hit_sigCard2(bxFringe ),hit_sigCard2(byFringe ));

	/* Layout gpushape with null initialization*/
	//lay->gpushape = HIT_GPU_NULL;

/* BLOCK SHAPE*/	
	
	/* Reading HIT_TOPOLOGY_INFO infor about the maximum thread per block simension*/
	int max_block_threads_x = HIT_TOPOLOGY_INFO.GPUInf[num_gpu].maxThreadsPerDimension[0];
	int max_block_threads_y = HIT_TOPOLOGY_INFO.GPUInf[num_gpu].maxThreadsPerDimension[1];
	int max_block_threads_z = HIT_TOPOLOGY_INFO.GPUInf[num_gpu].maxThreadsPerDimension[2];

	/* IF any dimension has a wrong value...*/
	if ( (hit_sigCard2(bx) > max_block_threads_x) || (hit_sigCard2(bx) <  1) ||		//ThreadBlock first dimension test
		(hit_sigCard2(by) > max_block_threads_y) || (hit_sigCard2(by) <  0) ||		//ThreadBlock second dimension test
		(hit_sigCard2(bz) > max_block_threads_z) || (hit_sigCard2(bz) <  0)		//ThreadBlock third dimension test
	)
		{

		fprintf(stderr,"ThreadBlock shape with a wrong value %d\n", hit_sigCard2(bz));
		return -1;

	}//if 

	/* Reading HIT_TOPOLOGY_INFO infor about the maximum thread per SM*/	
	int max_threads_SM = HIT_TOPOLOGY_INFO.GPUInf[num_gpu].maxThreadsPerBlock;

	/* If the total number of threadBlock threads is greater than the supported*/
	if ((hit_sigCard2(bx)*hit_sigCard2(by)/*+hit_sigCard2(bz)*/) > max_threads_SM ){
	
		fprintf(stderr,"The number of threadBlock thread is greater than the supported %d %d\n",hit_sigCard2(bx),hit_sigCard2(by));
		return -1;

	}//if 

	else{
		/* Copiying the threadBlock shape*/	
		part->blockShape= hit_shape3(bx, by, bz);
		//part->blockShape.sig[0]= bx;
		//part->blockShape.sig[1]= by;
		//part->blockShape.sig[2]= bz;
	}//else


/* FRINGE BLOCK SHAPE*/
	/* Copying the thread-Block extra edge*/
	//part->blockPadding= hit_shape2(byFringe, bxFringe);   ///////YURIIII


//	lay->gpushape.blockShapeFringe.sig[2]= gzFringe;	//So far, only two-dimensional arrays

		/* Calculating the layoutSig shape; the shape of the orig tile with the fringe cardinalities********/	
//		lay->info.layoutSig.shape= hit_shape2(hit_sigExpand(lay->origShape.sig[0], bxFringe), hit_sigExpand(lay->origShape.sig[1], byFringe));

		/* Check point*/
//		printf("bxFringe.begin... %d,bxFringe.end... %d,bxFringe.stride... %d\n\n",bxFringe.begin,bxFringe.end,bxFringe.stride);
//		printf("byFringe.begin... %d,byFringe.end... %d,byFringe.stride... %d\n\n",byFringe.begin,byFringe.end,byFringe.stride);



/* GRID SHAPE*/

	/* Reading HIT_TOPOLOGY_INFO infor about the maximum blocks per Grid simension*/
	int max_block_Grid_x = HIT_TOPOLOGY_INFO.GPUInf[num_gpu].maxBlocksPerDimension[0];
	int max_block_Grid_y = HIT_TOPOLOGY_INFO.GPUInf[num_gpu].maxBlocksPerDimension[1];

	/* IF any dimension has a wrong value...*/
	if ( (hit_sigCard2(gx) > max_block_Grid_x) || (hit_sigCard2(gx) <  1) ||		//ThreadBlock first dimension test
		(hit_sigCard2(gy) > max_block_Grid_y) || (hit_sigCard2(gy) <  1) 		//ThreadBlock second dimension test
	)
		{

		fprintf(stderr,"Grid shape with a wrong value\n");
		return -1;

	}//if 
	
	else{
		/* Copiying the Grid shape*/	
		part->gridShape= hit_shape2(gx, gy);
//			printf("signature gx %d %d\n",hit_sigCard2(shape.info.sig.sig[0]),hit_sigCard2(shape.info.sig.sig[1]));
		
		//part->gpushape.gridShape.sig[0]= gx;
		//part->gpushape.gridShape.sig[1]= gy;

	
	}//else


/* FRING GRID SHAPE*/

	/* If we have extra blocks threads in the blocks then, extra block*/
	if ((hit_sigCard2(bxFringe)-1) > 0){					//First dimension of the GridS

		//hit_shapeSig(part->gridPadding, 1)= hit_sig(0, 0, 1);  //FILAS de la matriz orig
	}//if

	if ((hit_sigCard2(byFringe)-1) > 0){					//Second dimnesion of the Grid

		hit_shapeSig(part->gridPadding, 0)= hit_sig(0, 0, 1); 

	}//if

//if (HIT_TOPOLOGY_INFO.selfRank==2){

//printf("grid padding %d  \n",hit_sigCard(hit_shapeSig(part->gridPadding, 1)));
//printf("grid padding %d  \n\n",hit_sigCard(hit_shapeSig(part->gridPadding, 0)));
//}

	/* Calculating always the grid shape (orig plus padding shape)*/
	hit_shapeSig(part->gridTotal, 0)=  hit_sigStd(( hit_sigCard(hit_shapeSig(part->gridShape, 0)) +hit_sigCard(hit_shapeSig(part->gridPadding, 0)))); 	
	hit_shapeSig(part->gridTotal, 1)=  hit_sigStd(( hit_sigCard(hit_shapeSig(part->gridShape, 1))));

	/* Calculating always the current Data shape (orig plus padding shape)*/

//if (HIT_TOPOLOGY_INFO.selfRank==2){

//printf("grid total %d  \n",hit_sigCard(hit_shapeSig(part->gridTotal, 0)));
//printf("grid total %d  \n",hit_sigCard(hit_shapeSig(part->gridTotal, 1)));
//}

	hit_shapeSig(part->currentDataShape, 1)=  hit_sigStd(hit_partitionTotalThreadShapeP(part, 0));			
	hit_shapeSig(part->currentDataShape, 0)=  hit_sigStd(hit_partitionTotalThreadShapeP(part, 1));				
	hit_shapeSig(part->currentDataShape, 2)=  hit_sig(-1, -1, 1);	//the last dimension is 1	
	part->currentDataShape.info.sig.numDims=2;

	HitShape f = hit_shapeIntersect( hit_shape(1,hit_sig(0,20,2)), hit_shape(1,hit_sig(4,22,2)));
fprintf(stderr,"CTRL %d\n", hit_shapeSig( f, 0 ).begin );

	/* Final instruction*/
	return 0;

//printf("end %d\n",bx.end);
}//hit_layout_set_GPU_shapes


#endif //nunca



