/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2011-2012,	Yuri Torres de la Sierra
 *			   				Arturo Gonzalez Escribano
 *                          
 * All Rights Reserved.
 */


#ifndef _HitPartition_
#define _HitPartition_

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


//#include "hitmap.h"
#include "hit_tile.h"
#include "hit_topology.h"
#include "hit_funcop.h"
#include "hit_shape.h"
#include "hit_acel.h"




/* Struct in order to store all information about blocks partitions*/
typedef struct HitPartition{

	/* The thread access pattern shape (one for each variable)*/
	HitShape	accessPattern[4];

	/* The threadBlock shape*/
	HitShape	blockShape;
	
	/* The grid shape*/
	HitShape	gridShape;		
 
	/* Original data shape (without padding)*/
	HitShape	origDataShape;		

	/* Padding data shape (data+padding by the GPU blocks)*/
	HitShape	paddingDataShape;

	/* Padding data + accessPatter*/
	HitShape	origDataPatternShape;			

} HitPartition, *HitPPartition;



/* PARTITION CONSTRUCTORS AND PREDEFINED PARTITION FUNCTIONS (pluging functions)*/

#define	hit_partition( name, dataShape, n, ...)	hit_partition_##name( dataShape, n, __VA_ARGS__ )

/* A. Naive partition*/
HitPartition hit_partition_plug_partNaive (HitShape dataShape, int n, ... );


/* Macro for calculate the data padding depending on the global memory access pattern*/
#define hit_partitionDataPaddingPatern(dataShape, partition, accessPattern)  hit_partitionDataPaddingPaternInternal(dataShape, partition, accessPattern)


#define hit_partitionExtendedDataPaddingPatern(tile,partition)	hit_partitionExtendedDataPaddingPaternInternal(tile,partition)

/* Function to calculate the data padding depending on the global memory access pattern*/
HitShape hit_partitionDataPaddingPaternInternal (HitShape dataShape, HitPartition partition, HitShape accessPattern);

/* Function to calculate a global shape including the computation and data shapes*/
HitShape hit_partitionExtendedDataPaddingPaternInternal(HitTile *tile, HitPartition partition);









////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////								////////////////////////////////////////////////
/////////////////////////////		 OLD CODE ABOUT PARTITION PLUG-INS		////////////////////////////////////////////////
//////////////////////////////								////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


/* Calculates the number of total blocks (including padding) in the grig for the selected dimension*/
#define hit_partition_totalSape(part, dim) ( hit_sigCard(hit_shapeSig(part.gridShape, dim)) ) //+ hit_sigCard(hit_shapeSig(part.gridPadding, dim))   )

/* Calculates the number of threads in a specific dimension*/
#define hit_partitionTotalThreadShapeP(part, dim) ( hit_sigCard(hit_shapeSig((part)->gridShape, dim))*hit_sigCard(hit_shapeSig((part)->blockShape, dim))   )

/* Calculates the number of data in a specific dimension*/
#define hit_partitionTotalCard(part, dim) hit_sigCard(hit_shapeSig((part).paddingDataShape, dim))

/* Number of threads in a specific dimension*/
#define hit_partition_threads(part, dim) (  hit_sigCard(hit_shapeSig(part.blockShape,dim))  )

/* The max shape of the tile, shape + padding*/
#define hit_partitionMaxShape(part)  (part).paddingDataShape




//#define	hit_partition( name, dataShape, ...)	hit_partition_##name( dataShape, __VA_ARGS__ )

/* A. PLUGIN TO MANUALLY INTRODUCE THE THREADBLOCK SHAPE*/
HitPartition hit_partition_plug_partBlocksFixedSize (HitShape dataShape, HitShape blockShape );

/* B. PLUGIN TO CREATE THE THREADBLOCK SHAPE FROM THREAD ACCESS PATTERN*/
HitPartition hit_partition_plug_partBlocksPattern (HitShape dataShape, int num_threadPatterns, ...);



	/* Set the Grid and threadBlock shapes in the layout from manual values*/							
		/* This function helps the previous one checking the shape parameters*/
	int hit_partition_set_GPU_shapes (HitPartition *part, HitShape shape, HitSig bx, HitSig by, HitSig bz);

	/* Extends the threadBlock to reach as far as possible the maximun size*/			
	HitShape hit_partition_extend_block(HitShape dataShape, HitShape threadBlockShape);

/* END OF HEADER FILE _HitPartition_ */
#endif






