/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2011-2012,	Yuri Torres de la Sierra
 *			   				Arturo Gonzalez Escribano
 *                          
 * All Rights Reserved.
 */



#include <hit_tile.h>
#include <acelNVIDIA.h>

/* TO SOLVE THE PROBLEMS OF THE UPDATE FUNCTIONS*/
/* Hit UPDATE from to HOST and GPU memories */

void acelSendRecvDataNVIDIA( void *hostTile, void *gpuTile,int fromTo, char *name){

	HitPTile shadowOrig = (HitPTile)hostTile;
	HitPTile shadow = (HitPTile)gpuTile;
	
//	hit_tileDumpTileFile(*shadowOrig,"dumpTileTestFile.txt");
//	hit_tileDumpTileFile(*shadow,"dumpTileTestFile.txt");

	HitTile reducedDomainShadow;
	int	ind[ hit_tileDims( *shadow ) ];
	int i,j,dim;
	//size_t k;
	size_t chunkSize;

	/* 0.1. SKIP UPDATING NULL */
	if ( shadow->memStatus == HIT_MS_NULL ) return;

	/* 1. COMPUTE INTERSECTION SHAPE TO DETECT SELECTIONS WHICH EXPAND THE ORIGINAL SHAPE.
	 * 		TILES CREATED WITH THE NO-BOUNDARY SELECTION FUNCTION.
	 * 		FOR NORMAL SELECTIONS THE INTERSECTION IS THE SHAME SHAPE THAN THE SHAPE IN shadow */
	/* 1.1. COMPUTE INTERSECTION SHAPE */
	HitShape intersection = hit_shapeIntersect( hit_tileShape(*shadow),hit_tileShape(*shadowOrig) );
	/* 1.2. IF NEEDED, CHANGE shadow BY A NEW SELECTION TILE WITH THE INTERSECTION SHAPE 
	 *		NO NEED TO UPDATE ref FIELD, WE WILL ONLY USE THE CARDINALITIES AND DATA POINTER */
	if ( ! hit_shapeCmp( intersection, hit_tileShape( *shadow ) ) ) {
		hit_tileSelectArrayCoords( &reducedDomainShadow, shadow, intersection );
		shadow = &reducedDomainShadow;
	}

	/* 2. LOCATE POSSIBLE CONTIGUOUS DIMENSIONS (dim in [-1,numDims-1]) */
	/* NEW v1.1: THE shadow VARIABLE IS HIT_MS_OWNER AND THE WHOLE DATA IS CONTIGUOUS,
	 * 				BUT WE NEED TO CHECK THE shadowOrig VARIABLE */
	/* NEW v1.1: RECORD THE SIZE OF THE CONTIGUOS CHUNKS TO COPY */
	chunkSize = (size_t)1;
	for ( dim = hit_tileDims( *shadow )-1
		;
		dim >= 0 
		&& shadow->card[dim] == shadowOrig->card[dim]
		&& shadow->origAcumCard[dim+1] == shadowOrig->origAcumCard[dim+1]
		//&& shadow->origAcumCard[dim+1] == shadowOrig->origAcumCard[dim]		//yuri
		//&& shadow->origAcumCard[dim+1] == shadow->card[dim]					//yuri
		&& shadowOrig->qstride[dim] == 1
		;
		dim-- ) 
		chunkSize = chunkSize * (size_t)shadow->card[ dim ];

//printf("card %d\n",dim); //exit(-1);
	/* THE CHUNK SIZE IS MULTIPLIED BY THE NUMBER OF ELEMENTS IN THIS DIMENSION WHICH
	 * ARE CONTIGUOS IN THE shadow (HIT_NS_OWNER), PROVIDED THE ORIGINAL HAS NO qstride */
	// IF shadow SELECTIONS WERE POSSIBLE, THEY SHOULD ALSO CHECK HERE THEIR qstride 
	if ( shadowOrig->qstride[dim] == 1 ) {
		chunkSize = chunkSize * (size_t)shadow->card[ dim ];
		dim--;
	}

	/* 3. IF dim=-1, COPY THE WHOLE DATA OF THE ORIGINAL VARIABLE IN ONE STEP, AND RETURN */
	if ( dim == -1 ) {
		// NATIVE TYPE TILES
		if ( shadowOrig->hierDepth == HIT_NONHIERARCHICAL ) {
			
			acelCopyDataFromToNVIDIA((char*)(shadowOrig->data), (char*)(shadow->data), shadow->baseExtent*shadow->acumCard, fromTo );
				hit_sigShapeAccess(shadow->shape).sig[0].begin++;

		}
		// PARANOID CHECK: IMPOSIBLE VALUE IN FIELD
		else hit_errInternal("tileUpdateFromAncestor", "Wrong value in hierarchy-level field of variable:", name,__FILE__,__LINE__); 

		return;
	}
 
	/* 6. INIT INDICES */
	for(i=0; i< dim+1; i++) ind[i] = 0;

	/* 7. COPY DATA (AS MANY CHUNKS AS NEEDED) */
	for(i=0; i< (int)((size_t)shadow->acumCard/chunkSize); i++) {
		int pos = 0;
		/* 7.1. COMPUTE NEXT POSITION IN ORIGINAL SELECTION */
		for(j=0; j< dim+1; j++) {
			pos = pos + ind[j] * shadowOrig->qstride[dim] * shadowOrig->origAcumCard[j+1];
		}

		/* 7.2. COMPUTE NEXT POSITION IN SHADOW (IF IT IS A SELECTION) */
		// shadow DOES NOT HAVE qstride BECAUSE CURRENTLY IT IS HIT_MS_OWNER
		int posShadow = 0;
		for(j=0; j< dim+1; j++) {
			posShadow = posShadow + ind[j] * shadow->origAcumCard[j+1];
		}

		/* 7.3. COPY AT THE NEXT POSITION IN NEW DATA */
		/* BUG CORRECTED: IF shadow IS A SELECTION THE LOCAL DATA MAY NOT BE CONTIGUOUS */
		if ( shadowOrig->hierDepth == HIT_NONHIERARCHICAL ) {
			acelCopyDataFromToNVIDIA( 
					(char*)(shadowOrig->data)+(size_t)pos*shadow->baseExtent,
					(char*)(shadow->data)+(size_t)posShadow*shadow->baseExtent,
					shadow->baseExtent*chunkSize,
					fromTo );	
		}
		// PARANOID CHECK: IMPOSIBLE VALUE IN FIELD
		else hit_errInternal("tileUpdateFromAncestor", "Wrong value in hierarchy-level field of variable:", name, __FILE__,__LINE__); 

		/* 7.3. ADVANCE INDECES */
		for(j=dim; j>=0; j--) {
			ind[j]++;
			if (ind[j]==shadow->card[j]) ind[j]=0;
			else break;
		}
	}

	/* 8. RETURN */
	return;
}



