/**
* Definition of the operator for HitSigDomain.
* Constructors and functions to manipulate them.
*
* @file hit_domain.c
* @version 1.0
* @author Arturo González-Escribano
* @author Ana Moretón Fernández
* @date Mar 2012
*/


#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <limits.h>
#include <hit_domain.h>
#include <hit_allocP.h>
#include <hit_shape.h>
#include <hit_sig.h>
#include <unistd.h>
#include <hit_funcop.h>


/* 1. Hit SHAPE WHOLE CONSTANT */
HitSigDomain HIT_DOMAIN_NULL = HIT_DOMAIN_NULL_STATIC;


/* 2.  Hit DOMAIN: SUBSTRACT OPERATOR */

/**
* Analysis by dimension.
*/
void drawshapes(HitSigDomain *result ,HitShape sh1, HitShape sh3, int numDim, int fijo){

int pos=numDim-fijo;
int posn=-1;
int priority[HIT_MAXDIMS]= {0,1,2,3};
int t,g;


if(fijo>0){
		
	/* First excepcion: the begin of sh1 is less than the begin of sh2 */

	if(hit_shapeSig(sh1,priority[pos]).begin<hit_shapeSig(sh3,priority[pos]).begin){
		HitSig s[numDim];
		for( t=0;t<numDim;t++){
			if(t==priority[pos]){s[t]=hit_sig(hit_shapeSig(sh1,t).begin,hit_shapeSig(sh3,t).begin-1,hit_shapeSig(sh1,t).stride);}
								
			else{
				/* Search the position of t in priority */
				for( g=0;g<numDim;g++){if(t==priority[g]){posn=g;}}

				/* if t is before priority[pos], we use sh3, else sh1 */
				if(posn<pos){s[t]=hit_shapeSig(sh3,t);}
				if(posn>pos){s[t]=hit_shapeSig(sh1,t);}
			   }	

			}
		
		/* Build the shapes */
		HitShape a=HIT_SHAPE_NULL;
		if(numDim==1){ a=hit_shape1(s[0]);}

		if(numDim==2){ a=hit_shape2(s[0],s[1]);}
														
		if(numDim==3){ a=hit_shape3(s[0],s[1],s[2]);}
													
		if(numDim==4){a=hit_shape4(s[0],s[1],s[2],s[3]);}
		
		result->numShapes++;							
		result->shape=(HitShape *)realloc(result->shape,(size_t)(result->numShapes)*sizeof(HitShape));
		(result->shape[result->numShapes-1])=a;									
		

	}

	/* Second excepcion: the begin of sh2 is less than the begin of sh1 */
	if(hit_shapeSig(sh3,priority[pos]).end<hit_shapeSig(sh1,priority[pos]).end){
		HitSig s[numDim];
		for( t=0;t<numDim;t++){
			if(t==priority[pos]){s[t]=hit_sig(hit_shapeSig(sh3,t).end+1,hit_shapeSig(sh1,t).end,hit_shapeSig(sh1,t).stride);}

			else{
				for( g=0;g<numDim;g++){if(t==priority[g]){posn=g;}}
	
				if(posn<pos){s[t]=hit_shapeSig(sh3,t);}
				if(posn>pos){s[t]=hit_shapeSig(sh1,t);}
			   }					
		}	

		HitShape b=HIT_SHAPE_NULL;
		if(numDim==1){ b=hit_shape1(s[0]);}

		if(numDim==2){ b=hit_shape2(s[0],s[1]);}
													
		if(numDim==3){ b=hit_shape3(s[0],s[1],s[2]);}
													
		if(numDim==4){ b=hit_shape4(s[0],s[1],s[2],s[3]);}
	
	
		result->numShapes++;							
		result->shape=(HitShape *)realloc(result->shape,(size_t)(result->numShapes)*sizeof(HitShape));
		(result->shape[(result->numShapes)-1])=b;
	}
		
	/* Third excepcion: Recursive function. Analize the following dimension */

	drawshapes(result,sh1,sh3,numDim,fijo-1);	
}

}
        
/**
* HitShape's Substract.
*/

HitSigDomain hit_shapeSubtract(HitShape sh1, HitShape sh2){
	
	HitSigDomain result=HIT_DOMAIN_NULL;
	HitShape sh3=HIT_SHAPE_NULL;
	
	sh3=hit_shapeIntersect(sh1,sh2);
	
	/* 1. If the shapes not intersect */
	if(hit_shapeCmp(sh3,HIT_SHAPE_NULL)==1){

		/* numshapes=1 and shape=sh1 */	
		result.numShapes++;							
		result.shape=(HitShape *)realloc(result.shape,(size_t)(result.numShapes)*sizeof(HitShape));
		(result.shape[result.numShapes-1])=sh1;
		}
	
	/* 2. If the intersection between shapes exist */
	else{	int r=0,strides=0;
		int numDim=hit_shapeDims(sh1);
		for(r=0;r<numDim;r++){
			if(hit_shapeSig(sh1,r).stride!=hit_shapeSig(sh2,r).stride){strides=1;}		
			}

		/* If the strides are the same */
		if(strides==0){
		drawshapes(&result,sh1,sh3,numDim,numDim);
		}
		else printf("\n Warning: diferent strides in HitShapes");
	}

	return result;
}


/* 3.  Hit DOMAIN: UNION OPERATOR */

/**
* Analysis by dimension.
*/
void join(HitSigDomain *result,HitShape sh1, HitShape sh3, HitShape sh2, int numDim, int fijo,int near){


int pos=numDim-(fijo+1);
int posn=-1;
HitSig sig1[numDim],sig2[numDim];
int priority[HIT_MAXDIMS]= {0,1,2,3};
int t,g,cond,condsh1,condsh2;


	/* 1. Check the dimensions */
	condsh1=0;
	condsh2=0;
	for(g=0;g<numDim;g++){

	cond=0;


	for( t=0;t<numDim;t++){if(g==priority[t]){posn=t;}}

	if(posn<pos){sig1[g]=hit_shapeSig(sh3,g); sig2[g]=hit_shapeSig(sh3,g); }
	if(posn>=pos){sig1[g]=hit_shapeSig(sh1,g); sig2[g]=hit_shapeSig(sh2,g);}

	    /* All the dimensions except the dimension of the analysis */
	    if(priority[pos]!=g){

		/* if the dimensions except the dimension of the analysis are equal */
		if(hit_sigCmp(sig1[g],sig2[g])==0){	
			cond=1;		
			}
		/* If the shape's dimension except the dimension of the analysis 
		is in the dimension of the other shape */		
		if(sig1[g].begin<=sig2[g].begin && sig1[g].end>=sig2[g].end){
			condsh1++;
			}
		if(sig2[g].begin<=sig1[g].begin && sig2[g].end>=sig1[g].end){
			condsh2++;
			}
	     }	

	if(priority[pos]==g){

		if(sig1[g].begin<=sig2[g].begin && sig1[g].end>=sig2[g].end){
			condsh1--;
			}
		if(sig2[g].begin<=sig1[g].begin && sig2[g].end>=sig1[g].end){
			condsh2--;
			}
		}	

	}


	/* First exception: If a shape is in the other, built the shapes*/ 
	if(((condsh1==(numDim-1) && condsh2==0)
	|| (condsh2==(numDim-1)) && condsh1==0)
	&& cond==1 && fijo>0){


		HitSig s[numDim],s1[numDim];
		
		if(condsh1==(numDim-1)){
			for(t=0;t<numDim;t++){

				/* The dimension to be analyzed */
				if(t==priority[pos]){
					
					if(near==0){
printf("a");
						s[t]=hit_shapeSig(sh1,t);

						if(hit_shapeSig(sh1,t).begin<hit_shapeSig(sh2,t).begin){
							s1[t]=hit_sig(hit_shapeSig(sh3,t).end+1,hit_shapeSig(sh2,t).end,hit_shapeSig(sh2,t).stride);
							}
						else{s1[t]=hit_sig(hit_shapeSig(sh2,t).begin,hit_shapeSig(sh3,t).begin-1,hit_shapeSig(sh2,t).stride);}
						}
					/* shapes don't intersect */
					if(near==1){
					printf("b");	
						s[t]=hit_shapeSig(sh1,t);

						if(hit_shapeSig(sh1,t).begin<hit_shapeSig(sh2,t).begin){
							s1[t]=hit_sig(hit_shapeSig(sh3,t).begin,hit_shapeSig(sh2,t).end,hit_shapeSig(sh2,t).stride);
							}
						else{s1[t]=hit_sig(hit_shapeSig(sh2,t).begin,hit_shapeSig(sh3,t).end,hit_shapeSig(sh2,t).stride);}

						}

					
					
					}
				else{ s[t]=sig1[t]; s1[t]=sig2[t];}
				}
			}

		if(condsh2==(numDim-1)){

			for(t=0;t<numDim;t++){

				/* The dimension to be analyzed */
				if(t==priority[pos]){
					
					if(near==0){
						s[t]=hit_shapeSig(sh2,t);
						
						if(hit_shapeSig(sh2,t).begin<hit_shapeSig(sh1,t).begin){

							s1[t]=hit_sig(hit_shapeSig(sh3,t).end+1,hit_shapeSig(sh1,t).end,hit_shapeSig(sh1,t).stride);
							}
						else{s1[t]=hit_sig(hit_shapeSig(sh1,t).begin,hit_shapeSig(sh3,t).begin-1,hit_shapeSig(sh1,t).stride);}
						}
					/* shapes don't intersect */
					if(near==1){  
						
						s[t]=hit_shapeSig(sh2,t);
						
						if(hit_shapeSig(sh2,t).begin<hit_shapeSig(sh1,t).begin){
							s1[t]=hit_sig(hit_shapeSig(sh3,t).begin,hit_shapeSig(sh1,t).end,hit_shapeSig(sh1,t).stride);
							}
						else{s1[t]=hit_sig(hit_shapeSig(sh1,t).begin,hit_shapeSig(sh3,t).end,hit_shapeSig(sh1,t).stride);}
						}

					
					
					}
				else{ s[t]=sig2[t]; s1[t]=sig1[t];}


				}
			}

		HitShape a=HIT_SHAPE_NULL,b=HIT_SHAPE_NULL;
		
		if(numDim==1){ a=hit_shape1(s[0]); b=hit_shape1(s1[0]);}

		if(numDim==2){ a=hit_shape2(s[0],s[1]); b=hit_shape2(s1[0],s1[1]);}
														
		if(numDim==3){ a=hit_shape3(s[0],s[1],s[2]); b=hit_shape3(s1[0],s1[1],s1[2]);}
														
		if(numDim==4){a=hit_shape4(s[0],s[1],s[2],s[3]); b=hit_shape4(s1[0],s1[1],s1[2],s1[3]);}
		
		
		result->numShapes++;							
		result->shape=(HitShape *)realloc(result->shape,(size_t)(result->numShapes)*sizeof(HitShape));
		(result->shape[result->numShapes-1])=a;
		result->numShapes++;							
		result->shape=(HitShape *)realloc(result->shape,(size_t)(result->numShapes)*sizeof(HitShape));
		(result->shape[result->numShapes-1])=b;
		}

   else{

	/* Second exception: If the dimensions except the analized are equal or is the last dimension*/
	if(cond==0 || fijo==0){

	     HitSig s[numDim];
	     int begin=0,end=0;
	     for( t=0;t<numDim;t++){
		if(t==priority[pos]){
				if(hit_shapeSig(sh1,t).begin<hit_shapeSig(sh2,t).begin){begin=hit_shapeSig(sh1,t).begin;}
				else {begin=hit_shapeSig(sh2,t).begin;}
				
				if(hit_shapeSig(sh1,t).end>hit_shapeSig(sh2,t).end){end=hit_shapeSig(sh1,t).end;}
				else {end=hit_shapeSig(sh2,t).end;}
				
				s[t]=hit_sig(begin,end,hit_shapeSig(sh2,t).stride);
	
				}										
		else{	s[t]=sig1[t];}	
		}
		
		HitShape a=HIT_SHAPE_NULL;
		if(numDim==1){ a=hit_shape1(s[0]);}

		if(numDim==2){ a=hit_shape2(s[0],s[1]);}
														
		if(numDim==3){ a=hit_shape3(s[0],s[1],s[2]);}
														
		if(numDim==4){a=hit_shape4(s[0],s[1],s[2],s[3]);}
			
		result->numShapes++;							
		result->shape=(HitShape *)realloc(result->shape,(size_t)(result->numShapes)*sizeof(HitShape));
		(result->shape[result->numShapes-1])=a;				
		
       }


	/* third exception: If the general analysis*/
	if(fijo>0 && cond!=0){
	
		/* 1.1 The begin of sh1 is less tan the begin of sh2 */
		if(hit_shapeSig(sh1,priority[pos]).begin<hit_shapeSig(sh3,priority[pos]).begin){
			HitSig s[numDim];
			for( t=0;t<numDim;t++){
				if(t==priority[pos]){s[t]=hit_sig(hit_shapeSig(sh1,t).begin,hit_shapeSig(sh3,t).begin-1,hit_shapeSig(sh1,t).stride);}
								
				else{   s[t]=sig1[t];}	
				}
			
			HitShape a=HIT_SHAPE_NULL;
			if(numDim==1){ a=hit_shape1(s[0]);}

			if(numDim==2){ a=hit_shape2(s[0],s[1]);}
														
			if(numDim==3){ a=hit_shape3(s[0],s[1],s[2]);}
														
			if(numDim==4){a=hit_shape4(s[0],s[1],s[2],s[3]);}
			
			result->numShapes++;							
			result->shape=(HitShape *)realloc(result->shape,(size_t)(result->numShapes)*sizeof(HitShape));
			(result->shape[result->numShapes-1])=a;

		}
	
	
		/* 1.2 The begin of sh2 is less tan the begin of sh1 */
		if(hit_shapeSig(sh2,priority[pos]).begin<hit_shapeSig(sh3,priority[pos]).begin){
			HitSig s[numDim];
			for( t=0;t<numDim;t++){
				if(t==priority[pos]){s[t]=hit_sig(hit_shapeSig(sh2,t).begin,hit_shapeSig(sh3,t).begin-1,hit_shapeSig(sh2,t).stride);}
									
				else{	s[t]=sig2[t]; }	
				}
			
			HitShape a=HIT_SHAPE_NULL;
			if(numDim==1){ a=hit_shape1(s[0]);}
	
			if(numDim==2){ a=hit_shape2(s[0],s[1]);}
															
			if(numDim==3){ a=hit_shape3(s[0],s[1],s[2]);}
														
			if(numDim==4){a=hit_shape4(s[0],s[1],s[2],s[3]);}
			
			result->numShapes++;							
			result->shape=(HitShape *)realloc(result->shape,(size_t)(result->numShapes)*sizeof(HitShape));
			(result->shape[result->numShapes-1])=a;									
		}
	
	
		/* 2.1 The end of sh1 is bigger tan the end of sh2 */
		if(hit_shapeSig(sh3,priority[pos]).end<hit_shapeSig(sh1,priority[pos]).end){
			HitSig s[numDim];
			for( t=0;t<numDim;t++){
				if(t==priority[pos]){s[t]=hit_sig(hit_shapeSig(sh3,t).end+1,hit_shapeSig(sh1,t).end,hit_shapeSig(sh1,t).stride);}
	
				else{	s[t]=sig1[t];}					
			}	
			
			HitShape b=HIT_SHAPE_NULL;
			if(numDim==1){ b=hit_shape1(s[0]);}
	
			if(numDim==2){ b=hit_shape2(s[0],s[1]);}
														
			if(numDim==3){ b=hit_shape3(s[0],s[1],s[2]);}
														
			if(numDim==4){ b=hit_shape4(s[0],s[1],s[2],s[3]);}
			
			result->numShapes++;							
			result->shape=(HitShape *)realloc(result->shape,(size_t)(result->numShapes)*sizeof(HitShape));
			(result->shape[result->numShapes-1])=b;											
			
		}
			
		/* 2.2 The end of sh2 is bigger tan the end of sh1 */
		if(hit_shapeSig(sh3,priority[pos]).end<hit_shapeSig(sh2,priority[pos]).end){
			HitSig s[numDim];
			for( t=0;t<numDim;t++){
				if(t==priority[pos]){s[t]=hit_sig(hit_shapeSig(sh3,t).end+1,hit_shapeSig(sh2,t).end,hit_shapeSig(sh2,t).stride);}
	
				else{	s[t]=sig2[t];}					
			}	
			
			HitShape b=HIT_SHAPE_NULL;
			if(numDim==1){ b=hit_shape1(s[0]);}
	
			if(numDim==2){ b=hit_shape2(s[0],s[1]);}
														
			if(numDim==3){ b=hit_shape3(s[0],s[1],s[2]);}
														
			if(numDim==4){ b=hit_shape4(s[0],s[1],s[2],s[3]);}
																			
			result->numShapes++;							
			result->shape=(HitShape *)realloc(result->shape,(size_t)(result->numShapes)*sizeof(HitShape));
			(result->shape[result->numShapes-1])=b;
		}
	
		/* 3. Recursive function. Analize the following dimension */
		join(result,sh1,sh3,sh2,numDim,fijo-1,near);			
	}
   }	
}

/**
* HitShape's Union.
*/

HitSigDomain hit_shapeUnion(HitShape sh1, HitShape sh2){

	HitSigDomain result=HIT_DOMAIN_NULL;
	HitShape sh3=HIT_SHAPE_NULL;
	int near=0;
	sh3=hit_shapeIntersect(sh1,sh2);
	
	/* 1. If the shapes not intersect */
	if(hit_shapeCmp(sh3,HIT_SHAPE_NULL)==1){
		int numDim1=hit_shapeDims(sh1);
		int numDim2=hit_shapeDims(sh2);
		int cont1=0,cont2=0,dim=-1;
		int i;
		/* if the dimension are equal count the intersect between sigs*/
		if(numDim1==numDim2 ){
			for(i=0;i<numDim1;i++){
			    if(hit_shapeSig(sh1,i).stride==hit_shapeSig(sh2,i).stride){
				HitSig a=hit_sigIntersect(hit_shapeSig(sh1,i),hit_shapeSig(sh2,i));
				if(hit_sigCmp(a,HIT_SIG_NULL)==1){
					if((hit_shapeSig(sh1,i).begin -hit_shapeSig(sh2,i).end)==hit_shapeSig(sh2,i).stride ||
						(hit_shapeSig(sh2,i).begin -hit_shapeSig(sh1,i).end)==hit_shapeSig(sh2,i).stride ) 
						{cont1++;dim=i;}
					}
				else {
					if((hit_shapeSig(sh1,i).begin==hit_shapeSig(sh2,i).begin) && (hit_shapeSig(sh1,i).end==hit_shapeSig(sh2,i).end)){
					cont2++;}
				}

			   }
			}
		}

//TODO
		/* if there is a dimension to one stride, new sh3 */
		if( cont1==1 && cont2==(numDim2-1)){
			near=1;
			// make sh3
			HitSig s[numDim1];
			for( i=0;i<numDim1;i++){
				if(i==dim){
					if(hit_shapeSig(sh1,i).begin>hit_shapeSig(sh2,i).end){
						s[i]=hit_sig(hit_shapeSig(sh2,i).begin,hit_shapeSig(sh1,i).end,hit_shapeSig(sh2,i).stride);
						}
					if(hit_shapeSig(sh2,i).begin>hit_shapeSig(sh1,i).end){
						s[i]=hit_sig(hit_shapeSig(sh1,i).begin,hit_shapeSig(sh2,i).end,hit_shapeSig(sh2,i).stride);
						}
					}
				else{	s[i]=hit_sigIntersect(hit_shapeSig(sh1,i),hit_shapeSig(sh2,i));
					}					
			}	
	
			
			if(numDim1==1){ sh3=hit_shape1(s[0]);}
	
			if(numDim1==2){ sh3=hit_shape2(s[0],s[1]);}
														
			if(numDim1==3){ sh3=hit_shape3(s[0],s[1],s[2]);}
														
			if(numDim1==4){ sh3=hit_shape4(s[0],s[1],s[2],s[3]);}
			
			join(&result,sh1,sh3,sh2,numDim1,numDim1-1,near);
			return result;
			
			}

		else{
		/* numshapes=2 and shape=sh1,sh2 */	
		
		result.numShapes++;							
		result.shape=(HitShape *)realloc(result.shape,(size_t)(result.numShapes)*sizeof(HitShape));
		(result.shape[result.numShapes-1])=sh1;
		
		result.numShapes++;							
		(result.shape)=(HitShape *)realloc(result.shape,(size_t)(result.numShapes)*sizeof(HitShape));
		(result.shape[result.numShapes-1])=sh2;		

			}
		}

    if(hit_shapeCmp(sh3,HIT_SHAPE_NULL)==0 ){


	/* 2. If a shape is in the other shape*/
	if(hit_shapeCmp(sh3,sh2)==1 || hit_shapeCmp(sh3,sh1)==1){
		int r=0,strides=0;
		int numDim=hit_shapeDims(sh1);
		for(r=0;r<numDim;r++){
			if(hit_shapeSig(sh1,r).stride!=hit_shapeSig(sh2,r).stride){strides=1;}	
			}
		/* If the strides are the same */
		if(strides==0){
		
		if(hit_shapeCmp(sh3,sh1)==1){sh1=sh2;}
		result.numShapes++;							
		(result.shape)=(HitShape *)realloc(result.shape,(size_t)(result.numShapes)*sizeof(HitShape));
		(result.shape[(result.numShapes)-1])=sh1;
			}
		else printf("\n Warning: diferent strides in HitShapes");
		}

	/* 3. If the intersection between shapes exist */
	else{
		int r=0,strides=0;
		int numDim=hit_shapeDims(sh1);
		for(r=0;r<numDim;r++){
			if(hit_shapeSig(sh1,r).stride!=hit_shapeSig(sh2,r).stride){strides=1;}			
			}
		/* If the strides are the same */
		if(strides==0){
		  join(&result,sh1,sh3,sh2,numDim,numDim-1,0);
		}
		else printf("\n Warning: diferent strides in HitShapes");
	}
    }

	return result;

}
