/**
 * Hierarchical tiling of multi-dimensional dense array.
 * Definitions and functions to manipulate HitTile types. 
 * This type defines a handler for dense arrays and sparse structures
 * with meta-information to create and manipulate hierarchical tiles. 
 *
 * @file hit_dump.c
 * @version 1.1
 * @author Arturo Gonzalez-Escribano
 * @date Sep 2012
 */


/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2007-2012, Arturo Gonzalez Escribano
 * All Rights Reserved.
 */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <hit_dump.h>

//#include <hit_tile.h>
//#include <hit_cshape.h>
//#include <hit_bshape.h>


/* DEBUG: DUMP VARIABLE STRUCTURE INFORMATION */
void hit_dumpTileInternal(const void *var, const char* name, FILE *file) {
	int i;
	const HitTile *v = (const HitTile *)var;

	fprintf(file,"Dump tile %s, address %p\n",name,var);
	if (var == NULL) return;

	fprintf(file,"\tmemStatus: %d\n", v->memStatus);
	fprintf(file,"\tnumDims:   %d\n", hit_shapeDims(v->shape));
	fprintf(file,"\tbaseExtent:   %d\n", (unsigned int) v->baseExtent);
	for (i=0; i<hit_shapeDims(v->shape); i++) fprintf(file,"\tcard[%d]: %d\n", i, v->card[i]);
	fprintf(file,"\tacumCard:     %d\n", v->acumCard);
	for (i=0; i<hit_shapeDims(v->shape)+1; i++) fprintf(file,"\torigAcumCard[%d]: %d\n", i, v->origAcumCard[i]);
	for (i=0; i<hit_shapeDims(v->shape); i++) {
		fprintf(file,"\tqstride[%d]: %d\n", i, v->qstride[i]);
	}
	for (i=0; i<hit_shapeDims(v->shape); i++) {
		fprintf(file,"\tSig[%d]: %d, %d, %d\n", i, hit_shapeSig(v->shape,i).begin, hit_shapeSig(v->shape,i).end, hit_shapeSig(v->shape,i).stride);
	}
	fprintf(file,"\thierDepth:   %d\n", v->hierDepth);
	fprintf(file,"\tchildBegin: { ");
	for (i=0; i<HIT_MAXDIMS; i++) fprintf(file,"%d ", v->childBegin[i]);
	fprintf(file,"}\n");
	fprintf(file,"\tchildSize: { ");
	for (i=0; i<HIT_MAXDIMS; i++) fprintf(file,"%d ", v->childSize[i]);
	fprintf(file,"}\n");
	fprintf(file,"\tdata: %p\n", v->data);
	fprintf(file,"\tmemPtr: %p\n", v->memPtr);
	fprintf(file,"\tref: %p\n", (void*) v->ref);
	fprintf(file,"\n");
}

