

#include <hit_gctile.h>
#include <hit_cshape.h>
#include <hit_allocP.h>



void hit_gcTileDomainShapeInternal(void * newVarP, size_t baseExtent, HitShape shape, int allocOpts) {

	HitPTile newVar = (HitPTile)newVarP;
	*newVar = HIT_TILE_NULL;

	// 1. Num of dimensions, Mark no memory status, and Null pointers.
	newVar->shape = shape;
	newVar->baseExtent = baseExtent;
	newVar->memStatus = HIT_MS_NOMEM;
	newVar->hierDepth = (char) HIT_NONHIERARCHICAL;
	newVar->ref = NULL;
	newVar->data = NULL;
	newVar->memPtr = NULL;
	newVar->type = HIT_GC_TILE;
	

	// 2. Process Vertices and Edges cardinalities.
	if(allocOpts & HIT_VERTICES){
		newVar->card[0] = hit_cShapeNvertices(shape);
	}
		
	if(allocOpts & HIT_EDGES){
		newVar->card[1] = hit_cShapeNedges(shape);
	}

}


void hit_gcTileAllocInternal(void *newVarP, const char *name, const char *file, int numLine) {
	
	HitPTile newVar = (HitPTile)newVarP;

	// 0. Skip alloc when Null.
	if ( newVar->memStatus == HIT_MS_NULL) return;
	
	// 1. Check variable type, cannot have already its own memory.
	if ( newVar->memStatus == HIT_MS_OWNER ){
		hit_errInternal(__FUNCTION__,"Trying to reallocate a Tile: ",name,file,numLine);
	}

	// 3. Allocate memory.
	if(newVar->card[0] != 0){
		hit_malloc(newVar->memPtrVertices,(size_t)newVar->card[0] * newVar->baseExtent,void*);
		newVar->dataVertices = newVar->memPtrVertices;
	}
	
	if(newVar->card[1] != 0){
		hit_malloc(newVar->memPtr,(size_t)newVar->card[1] * newVar->baseExtent,void*);
		newVar->data = newVar->memPtr;
	}
	
	// 4. Change memory status.
	newVar->memStatus = HIT_MS_OWNER;
}




int hit_gcTileEdgeAtIndex(void * varP, int local1, int local2){
	
	/* 1. Get the shape of the tile */
	HitTile *var = (HitTile *)varP;
	HitShape shape = hit_tileShape(*var);

	int edge;
	hit_cShapeEdgeIterator(edge,shape,local1){
		int dst = hit_cShapeEdgeTarget(shape,edge);
		if(dst == local2) return edge;
	}

	return -1;
}


int hit_gcTileGraphEdgeAtIndex(void * varP, int pos1, int pos2){

	/* 1. Get the shape of the tile */
	HitTile *var = (HitTile *)varP;
	HitShape shape = hit_tileShape(*var);

	int local1 = hit_cShapeVertexToLocal(shape,pos1);
	int local2 = hit_cShapeVertexToLocal(shape,pos2);

	int edge;
	hit_cShapeEdgeIterator(edge,shape,local1){
		int dst = hit_cShapeEdgeTarget(shape,edge);
		if(dst == local2) return edge;
	}
	
	return -1;
}


inline int hit_gcTileGraphVertexAtIndex(void * varP, int vertex){

	/* 1. Get the shape of the tile */
	HitTile *var = (HitTile *)varP;
	HitShape shape = hit_tileShape(*var);

	return hit_cShapeVertexToLocal(shape,vertex);

}




void hit_gcTileClearVertices(void * varP){

	/* 1. Get the shape of the tile */
	HitTile *var = (HitTile *)varP;
	HitShape shape = hit_tileShape(*var);
	int nvertices = hit_cShapeNvertices(shape);
	
	bzero(var->dataVertices, var->baseExtent  * (size_t) nvertices);
}




void hit_gcTileCopyVertices(void * destP, void * srcP){

	/* 1. Get the shape of the tile */
	HitTile *dest = (HitTile *)destP;
	HitTile *src = (HitTile *)srcP;
	
	HitShape shape = hit_tileShape(*src);
	
	int nvertices = hit_cShapeNvertices(shape);
	
	memcpy(dest->dataVertices,src->dataVertices,src->baseExtent  * (size_t) nvertices);

}
