
#include <hit_mbtile.h>
#include <hit_bshape.h>
#include <hit_allocP.h>


void hit_mbTileDomainShapeInternal(void * newVarP, size_t baseExtent, HitShape shape) {

	HitPTile newVar = (HitPTile)newVarP;
	*newVar = HIT_TILE_NULL;

	// 1. Num of dimensions, Mark no memory status, and Null pointers.
	newVar->shape = shape;
	newVar->baseExtent = baseExtent;
	newVar->memStatus = HIT_MS_NOMEM;
	newVar->hierDepth = (char) HIT_NONHIERARCHICAL;
	newVar->ref = NULL;
	newVar->data = NULL;
	newVar->memPtr = NULL;
	newVar->type = HIT_MB_TILE;
	
	// 2. Process Vertices and Edges cardinalities.
	newVar->card[1] = hit_bShapeCard(shape,0);
	newVar->card[2] = hit_bShapeCard(shape,1);


}


void hit_mbTileAllocInternal(void *newVarP, const char *name, const char *file, int numLine) {
	
	HitPTile newVar = (HitPTile)newVarP;

	// 0. Skip alloc when Null.
	if ( newVar->memStatus == HIT_MS_NULL) return;
	
	// 1. Check variable type, cannot have already its own memory.
	if ( newVar->memStatus == HIT_MS_OWNER ){
		hit_errInternal(__FUNCTION__,"Trying to reallocate a Tile: ",name,file,numLine);
	}

	// 3. Allocate memory.
	if(newVar->card[1] != 0 && newVar->card[2] != 0){
		hit_malloc(newVar->memPtr,(size_t)(newVar->card[1]*newVar->card[2]) * newVar->baseExtent,void*);
		newVar->data = newVar->memPtr;
	}
	
	// 4. Change memory status.
	newVar->memStatus = HIT_MS_OWNER;
}


int hit_mcTileElemAtIndex(void * varP, int row, int column){
	
	/* 1. Get the shape of the tile */
	HitTile *var = (HitTile *)varP;
	HitShape shape = hit_tileShape(*var);

	//int card1 = hit_bShapeCard(shape,0);
	int card2 = hit_bShapeCard(shape,1);

	return row * card2 + column;
}


int hit_mcTileGraphElemAtIndex(void * varP, int pos1, int pos2){

	/* 1. Get the shape of the tile */
	HitTile *var = (HitTile *)varP;
	HitShape shape = hit_tileShape(*var);

	int row = hit_bShapeCoordToLocal(shape,0,pos1);
	int column = hit_bShapeCoordToLocal(shape,1,pos2);

	//int card1 = hit_bShapeCard(shape,0);
	int card2 = hit_bShapeCard(shape,1);
	
	return row * card2 + column;
}



/* @arturo to @javier: TODO: Check this function, potential bugs */
void hit_mbTileClear(void * tileP){

	/* 1. Get the shape of the tile */
	HitTile *tile = (HitTile *)tileP;
	//HitShape shape = hit_tileShape(*tile);

	int card1 = tile->card[1];
	int card2 = tile->card[1];

	bzero(tile->data, tile->baseExtent  * (size_t) (card1 * card2) );
}





