


#include <hit_taskcontainers.h>
#include <string.h>
#include <hit_com.h>



/**
 * Null constant for the task list.
 */
HitTaskList HIT_TASK_LIST_NULL = HIT_TASK_LIST_NULL_STATIC;

/**
 * HitBufferInfo mpi type variable
 */
MPI_Datatype HitBufferInfoType = MPI_DATATYPE_NULL;



/**
 * Creates a new Task list with n channels
 * @param list Pointer to the struct
 * @param n_channels Number of channels
 */
void hit_taskListCreate(HitTaskList * list, int n_channels){

	*list = HIT_TASK_LIST_NULL;
	list->n_channels = n_channels;

}

/**
 * Remove the list and deallocate the resources.
 * @param list A task list pointer.
 */
void hit_taskListFree(HitTaskList * list){
	*list = HIT_TASK_LIST_NULL;
}


#ifdef debugddfasfaf

void dump_taskList(HitTaskList * list, int input, HitTask * t){

	if (input){
		printf("\n>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> %d\n",list->n_channels);
		if(t != NULL){
			printf("* Input tasks: ");
			int i;
			for(i=0;i<list->n_channels;i++){
				if(t[i].weight > 0){
					if(i > 0) printf("               ");
					printf("%d, size_t: %d, weight: %f, data: %p[%d]\n",
							t[i].n_elems,t[i].baseExtend,t[i].weight,t[i].data,*((int*)t[i].data));
				}else{
					printf("NULL\n");
				}
			}
		}
		printf("==============================\n");
	}else{
		printf("\n<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< %d\n",list->n_channels);
		if(t != NULL){
			printf("* Output tasks: ");
					int i;
					for(i=0;i<list->n_channels;i++){
						if(t[i].weight > 0){
							if(i > 0) printf("                ");
							printf("%d, size_t: %d, weight: %f, data: %p[%d]\n",
									t[i].n_elems,t[i].baseExtend,t[i].weight,t[i].data,*((int*)t[i].data));
						}else{
							printf("NULL\n");
						}
					}
		}
		printf("==============================\n");
	}

	printf("Number of tasks: %d\n",list->size);
	printf("First->: %p\n",list->first);
	printf("Last ->: %p\n",list->last);


	HitTaskListElem * elem = list->first;

	while(elem != NULL){

		HitTask * task = &elem->task[0];
		printf("Task[%p->%p->%p]: %d, weight: %f, data: %p[%d],\n", elem->prev, elem, elem->next,
				task->n_elems,task->weight,task->data,*((int*)task->data));

		elem = elem->next;
	}



	if (input){
		printf(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>\n");
	}else{
		printf("<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<\n");
	}
	fflush(stdout);

}

#endif


void hit_taskListAddLast(HitTaskList * list, void * taskPArray){

	pthread_mutex_lock(&(list->mutex));


	// Alloc a new element
	HitTaskListElem * elem;
	hit_malloc(elem,sizeof(HitTaskListElem),HitTaskListElem*);
	elem->next = NULL;
	elem->prev = NULL;

	if(list->size == 0){
		list->first = elem;
	} else {
		list->last->next = elem;
		elem->prev = list->last;
	}

	list->last = elem;
	// Now we have to copy all the tasks int the pointer (elem->task = task)
	HitTask * array = (HitTask*) taskPArray;

	int i;
	for(i=0; i<list->n_channels; i++){
		elem->task[i] = array[i];
		list->weight += array[i].weight;
	}
	list->size++;

	pthread_cond_broadcast(&(list->cond));
	pthread_mutex_unlock(&(list->mutex));



}


void hit_taskListAddFirst(HitTaskList * list, void * taskPArray){

	pthread_mutex_lock(&(list->mutex));

	// Alloc a new element
	HitTaskListElem * elem;
	hit_malloc(elem,sizeof(HitTaskListElem),HitTaskListElem*);


	elem->next = NULL;
	elem->prev = NULL;

	if(list->size == 0){
		list->last = elem;
	} else {
		list->first->prev = elem;
		elem->next = list->first;
	}

	list->first = elem;
	// Now we have to copy all the tasks int the pointer (elem->task = task)
	HitTask * array = (HitTask*) taskPArray;
	int i;
	for(i=0; i<list->n_channels; i++){
		elem->task[i] = array[i];
		list->weight += array[i].weight;
	}
	list->size++;


	pthread_cond_broadcast(&(list->cond));
	pthread_mutex_unlock(&(list->mutex));
}


int hit_taskListRemoveLast(HitTaskList * list, void * taskPArray){

	HitTask * array = (HitTask*) taskPArray;
	int i;
	for(i=0; i<list->n_channels; i++){
		array[i] = HIT_TASK_NULL;
	}
	int valueReturn = 0;

	pthread_mutex_lock(&(list->mutex));

	// Remove last do not wait it returns null because is the function used
	// by the work stealing, if we there is no tasks we can't send them.


	if(list->size != 0){

		for(i=0; i<list->n_channels; i++){
			array[i] = list->last->task[i];
			list->weight -= array[i].weight;
		}

		valueReturn = 1;

		if(list->size == 1){
			free(list->last);
			list->first = list->last = NULL;
		} else {
			HitTaskListElem * elem = list->last->prev;
			elem->next = NULL;
			free(list->last);
			list->last = elem;
		}

		list->size--;
	}


	pthread_mutex_unlock(&(list->mutex));

	return valueReturn;
}


int hit_taskListRemoveFirst(HitTaskList * list, void * taskPArray){

	HitTask * array = (HitTask*) taskPArray;
	int i;
	for(i=0; i<list->n_channels; i++){
		array[i] = HIT_TASK_NULL;
	}
	int valueReturn = 0;

	pthread_mutex_lock(&(list->mutex));

	// @todo Quitar, ws solo empieza una vez que hemos procesado una tarea nuestra,
	// no se hace ws cuando aún no hemos empezado a procesar.
	//pthread_cond_broadcast(&(list->cond_ws));

	while(list->size == 0 && (!list->exhausted_input || !list->exhausted_ws)){
		list->wakeup_ws = 1; pthread_cond_broadcast(&(list->cond_ws)); // @javfres @new Despertamos el hilo ws por si tiene que pedir. 
		pthread_cond_wait(&(list->cond), &(list->mutex));
	}

	if(list->size != 0){

		for(i=0; i<list->n_channels; i++){

			array[i] = list->first->task[i];

			list->weight -= array[i].weight;
		}
		valueReturn = 1;

		if(list->size == 1){
			free(list->first);
			list->first = list->last = NULL;
		} else {
			HitTaskListElem * elem = list->first->next;
			elem->prev = NULL;
			free(list->first);
			list->first = elem;
		}

		list->size--;
	}

	list->wakeup_ws = 1; pthread_cond_broadcast(&(list->cond_ws));
	pthread_mutex_unlock(&(list->mutex));

	return valueReturn;

}


/**
 * Return the size of the list.
 * @param list The pointer to the list.
 * @return Number of tasks in the list.
 */
int hit_taskListSize(HitTaskList * list){

	int size;

	pthread_mutex_lock(&(list->mutex));

	size = list->size;

	pthread_mutex_unlock(&(list->mutex));

	return size;

}


/**
 * Return the weight of the list.
 * @param list The pointer to the list.
 * @return Total weight of the tasks in the list.
 */
double hit_taskListWeight(HitTaskList * list){

	double w;

	pthread_mutex_lock(&(list->mutex));

	w = list->weight;

	pthread_mutex_unlock(&(list->mutex));

	return w;

}



/**
 * Set the input stream task as exhausted.
 * @param list The pointer to the list.
 */
void hit_taskListSetExhausted_input(HitTaskList * list){

	pthread_mutex_lock(&(list->mutex));

	list->exhausted_input = 1;

	pthread_cond_broadcast(&(list->cond));

	
	list->wakeup_ws = 1; pthread_cond_broadcast(&(list->cond_ws));

	pthread_mutex_unlock(&(list->mutex));

}

/**
 * Set the work stealing stream task as exhausted.
 * @param list The pointer to the list.
 */
void hit_taskListSetExhausted_ws(HitTaskList * list){

	pthread_mutex_lock(&(list->mutex));

	list->exhausted_ws = 1;

	pthread_cond_broadcast(&(list->cond));
	list->wakeup_ws = 1; pthread_cond_broadcast(&(list->cond_ws));

	pthread_mutex_unlock(&(list->mutex));

}


/**
 * This function waits until a tasks was been removed from the list.
 * This is used by the ws thread to ask for more tasks.
 * @param list The list pointer.
 * @return It returns 1 if if it could be called again. It return 0 if
 * there is no need to wait.
 */
int hit_taskListWaitRemove(HitTaskList * list){

	

	pthread_mutex_lock(&(list->mutex));

	if(list->size == 0 && list->exhausted_input && list->exhausted_ws){
		pthread_mutex_unlock(&(list->mutex));
		return 0;
	}

	// @javfres @new There was a pending wakeup
	if(!list->wakeup_ws){
		pthread_cond_wait(&(list->cond_ws), &(list->mutex));
	}
	list->wakeup_ws = 0;

	// pthread_cond_wait(&(list->cond_ws), &(list->mutex));
	

	int result = 1;

	if(list->size == 0 && list->exhausted_input && list->exhausted_ws){
		result = 0;
	}

	pthread_mutex_unlock(&(list->mutex));

	return result;

}






/**
 * Wakes up the work stealing thread waiting for a task removal.
 * @param list The list pointer.
 */
void hit_taskListWakeUp(HitTaskList * list){

	pthread_mutex_lock(&(list->mutex)); // @javfres @new Maybe not necesarry
	list->wakeup_ws = 1; // @javfres @new
	pthread_cond_broadcast(&(list->cond_ws));
	pthread_mutex_unlock(&(list->mutex));

}




// ----------------------------------------------------------- //



int printdatatype( MPI_Datatype datatype ) {
    int *array_of_ints;
    MPI_Aint *array_of_adds;
    MPI_Datatype *array_of_dtypes;
    int num_ints, num_adds, num_dtypes, combiner;
    int i;

    MPI_Type_get_envelope( datatype,
                           &num_ints, &num_adds, &num_dtypes, &combiner );
    switch (combiner) {
    case MPI_COMBINER_NAMED:
        printf( "Datatype is named:" );
        /* To print the specific type, we can match against the
           predefined forms. We can NOT use a switch statement here
           We could also use MPI_TYPE_GET_NAME if we prefered to use
           names that the user may have changed.
         */
        if      (datatype == MPI_INT)    printf( "MPI_INT\n" );
        else if (datatype == MPI_DOUBLE) printf( "MPI_DOUBLE\n" );
        //... else test for other types ...
        return 0;
        break;
    case MPI_COMBINER_STRUCT:
    case MPI_COMBINER_STRUCT_INTEGER:
        printf( "Datatype is struct containing" );
        array_of_ints   = (int *)malloc( (size_t) num_ints * sizeof(int) );
        array_of_adds   =
                   (MPI_Aint *) malloc(  (size_t)  num_adds * sizeof(MPI_Aint) );
        array_of_dtypes = (MPI_Datatype *)
            malloc(  (size_t)  num_dtypes * sizeof(MPI_Datatype) );
        MPI_Type_get_contents( datatype, num_ints, num_adds, num_dtypes,
                         array_of_ints, array_of_adds, array_of_dtypes );
        printf( " %d datatypes:\n", array_of_ints[0] );
        for (i=0; i<array_of_ints[0]; i++) {
            printf( "blocklength %d, displacement %ld, type:\n",
                    array_of_ints[i+1], array_of_adds[i] );
            if (printdatatype( array_of_dtypes[i] )) {
                /* Note that we free the type ONLY if it
                   is not predefined */
                MPI_Type_free( &array_of_dtypes[i] );
            }
        }
        free( array_of_ints );
        free( array_of_adds );
        free( array_of_dtypes );
        break;
        //... other combiner values ...
    default:
        printf( "Unrecognized combiner type\n" );
        break;
    }
    return 1;
}











/**
 * Task buffer null variable.
 */
HitTaskBuffer HIT_TASK_BUFFER_NULL = HIT_TASK_BUFFER_NULL_STATIC;


/**
 * Task buffer constructor
 * @param bufferP Pointer to the buffer
 * @param num_elem Number of elements in the buffer.
 * @param comm Communicator
 * @param tag Input tag
 * @param n_types Number of task types
 * @param ... List of hit data types.
 */
void hit_taskBufferCreate(HitTaskBuffer * bufferP, int num_elem, MPI_Comm comm, int tag, int n_types, ...){

	*bufferP = HIT_TASK_BUFFER_NULL;

	bufferP->num_elem = num_elem;

	bufferP->comm = comm;
	bufferP->tag = tag;

	bufferP->n_types = n_types;


	/**
	 * Create the type of the buffer info struct
	 */
	if(HitBufferInfoType == MPI_DATATYPE_NULL){
		hit_comTypeStruct(&HitBufferInfoType,HitBufferInfo,2,n_elems,1,MPI_INT,weight,1,MPI_DOUBLE);
	}


	// Calculate the size needed for the buffer
	int sizeof_element;
	int ok = MPI_Pack_size(1, HitBufferInfoType, comm, &sizeof_element);
	if(ok != MPI_SUCCESS) hit_error_here("Can not create the buffer, error in MPI_Pack_size(HitBufferInfoType)");

	va_list arglist;
	int i;
	va_start(arglist, n_types);
	for(i=0; i<n_types; i++){
		MPI_Datatype type = va_arg(arglist, MPI_Datatype);
		bufferP->type[i] = type;
		int size;
		int ok2 = MPI_Pack_size(1, type, comm, &size);
		if(ok2 != MPI_SUCCESS) hit_error_here("Can not create the buffer, error in MPI_Pack_size");
		sizeof_element += size;
		ok2 = MPI_Type_size(type, &bufferP->size[i]);
		if(ok2 != MPI_SUCCESS) hit_error_here("Can not create the buffer, error in MPI_Type_size");

	}
	va_end(arglist);

	// @todo @javfres: Alloc the maximun size of the task (Just in case)

	// Buffer alloc. +1 for the End of Buffer mark
	bufferP->buffer_size = (num_elem+1) * sizeof_element;
	hit_malloc((bufferP->buffer),(size_t) num_elem * (size_t) sizeof_element, void*);
	bufferP->position = 0;

}


/**
 * Frees the buffer memory.
 * @param buffer Pointer to the buffer.
 */
void hit_taskBufferFree(HitTaskBuffer * bufferP){

	hit_free(bufferP->buffer);
	*bufferP = HIT_TASK_BUFFER_NULL;
}


/**
 * Adds a task (or several) to the buffer.
 * @param bufferP Pointer buffer.
 * @param taskArray Task array.
 */
void hit_taskBufferAdd(HitTaskBuffer * bufferP, void * taskArray){

	HitTask * tasks = (HitTask*) taskArray;


	int i;
	for(i=0; i<bufferP->n_types; i++){

		int ok;
		HitBufferInfo info;

		info.n_elems = tasks[i].n_elems;
		info.weight  = tasks[i].weight;

		// Pack the information for the next task
		ok = MPI_Pack(&info, 1, HitBufferInfoType, bufferP->buffer, bufferP->buffer_size, &bufferP->position, bufferP->comm );
		if(ok != MPI_SUCCESS) hit_error_here("error packing HitBufferInfo with MPI_Pack");

		if(info.n_elems > 0){

			//printf("Data %p\n",tasks[i].data);
			//printf("Nelems %d\n",info.n_elems);
			//printf("Type %d\n",bufferP->type[i]);
			//printf("Buffer %p\n",bufferP->buffer);
			//printf("Buffer size %d\n",bufferP->buffer_size);
			//printf("Position %d\n",bufferP->position);
			//printf("Comm %d\n",bufferP->comm);

			// Pack the task
			ok = MPI_Pack(tasks[i].data, info.n_elems, bufferP->type[i], bufferP->buffer, bufferP->buffer_size, &bufferP->position, bufferP->comm );
			if(ok != MPI_SUCCESS) hit_error_here("error packing tasks with MPI_Pack");

			// @todo @javfres This can be dangerous for multibufferes (WS) which use multilengh tasks.
			bufferP->num_used += info.n_elems;
			bufferP->weight += info.weight;

			//printf("Adding task %d elems \n",info.n_elems);

		} else {
			bufferP->num_used++;
		}
	} // for loop

}


/**
 * Adds a special task to the buffer (EOF, REQUEST).
 * @param bufferP Pointer buffer.
 * @param flag (HIT_BUFFER_EOF/HIT_BUFFER_REQUEST)
 */
void hit_taskBufferAddSpecial(HitTaskBuffer * bufferP, int flag){

	int ok;
	HitBufferInfo info;

	info.n_elems = flag;
	info.weight = 0;

	// Pack the information for the next task
	ok = MPI_Pack(&info, 1, HitBufferInfoType, bufferP->buffer, bufferP->buffer_size, &bufferP->position, bufferP->comm );
	if(ok != MPI_SUCCESS) hit_error_here("error packing HitBufferInfo with MPI_Pack");

	bufferP->num_elem++;

}




/**
 * Send the content of the buffer.
 * @param bufferP Pointer to the buffer.
 * @param dest Rank of destination
 * @param tag Message tag
 */
void hit_taskBufferSend(HitTaskBuffer * bufferP, int dest, int tag){

	//hit_taskBufferAddSpecial(bufferP,&HIT_TASK_EOB);

	int ok = MPI_Send(bufferP->buffer, bufferP->position, MPI_PACKED, dest, tag, bufferP->comm);

	if(ok != MPI_SUCCESS) hit_error_here("MPI_Send error\n");

	bufferP->num_used = 0;
	bufferP->weight = 0;
	bufferP->position = 0;
}



/**
 * Checks the next task in the buffer and updates the flags
 * @param bufferP Buffer pointer
 */
void hit_taskBufferCheckNext(HitTaskBuffer * bufferP){

	// Reset buffer status
	bufferP->empty = 0;
	bufferP->eof = 0;
	bufferP->request = 0;

	// End of the buffer
	if(bufferP->position == bufferP->buffer_received){
		bufferP->empty = 1;
		return;
	}

	// Get the next element
	int position = bufferP->position;
	HitBufferInfo info;
	MPI_Unpack(bufferP->buffer, bufferP->buffer_size, &position, &info, 1, HitBufferInfoType, bufferP->comm);

	// EOF
	if( info.n_elems == HIT_BUFFER_EOF ){
		bufferP->empty = 1;
		bufferP->eof = 1;
		return;
	}

	// Request
	if( info.n_elems == HIT_BUFFER_REQUEST ){
		bufferP->empty = 1;
		bufferP->request = 1;
		return;
	}

}



/**
 * Recv tasks and stores them into the buffer.
 * @param bufferP Pointer to the buffer.
 * @param source Rank of source
 */
void hit_taskBufferRecv(HitTaskBuffer * bufferP, int * source){

	if(bufferP->null_buffer) return;

	//printf("Buffer %p\n",bufferP->buffer);
	//printf("Buffer size %d\n",bufferP->buffer_size);
	//printf("Tag %d\n",bufferP->tag);
	//printf("Comm %d\n",bufferP->comm);

	MPI_Status status;
	int ok = MPI_Recv(bufferP->buffer, bufferP->buffer_size, MPI_PACKED, MPI_ANY_SOURCE, bufferP->tag, bufferP->comm, &status);
	if(ok != MPI_SUCCESS) hit_error_here("MPI_Recv error\n");
	bufferP->buffer_received = status.count;

	// Return the source
	*source = status.MPI_SOURCE;

	// Reset position
	bufferP->position = 0;

	hit_taskBufferCheckNext(bufferP);
}


/**
 * Recv tasks and stores them into the buffer (non-blocking version).
 * @param bufferP Pointer to the buffer.
 * @param source Rank of source
 */
void hit_taskBufferIRecv(HitTaskBuffer * bufferP, MPI_Request * request){

	if(bufferP->null_buffer){
		*request = 	MPI_REQUEST_NULL;
		return;
	}


	int ok = MPI_Irecv(bufferP->buffer, bufferP->buffer_size, MPI_PACKED, MPI_ANY_SOURCE, bufferP->tag, bufferP->comm, request);
	if(ok != MPI_SUCCESS) hit_error_here("MPI_Irecv error\n");
	bufferP->pending_recv = 1;

}

/**
 * Finalizes a IRecv operation
 * @param bufferP Pointer to the buffer.
 * @param source Rank of source
 */
void hit_taskBufferIRecvEnd(HitTaskBuffer * bufferP, int * source, MPI_Status status){

	if(bufferP->null_buffer) return;

	if(status.MPI_ERROR != MPI_SUCCESS) hit_error_here("MPI_WaitAny error\n");
	bufferP->buffer_received = status.count;

	// Return the source
	*source = status.MPI_SOURCE;

	// Reset position
	bufferP->position = 0;

	hit_taskBufferCheckNext(bufferP);

	bufferP->pending_recv = 0;
}



/**
 * Creates a new task using the i buffer element.
 * @param bufferP Pointer to the buffer.
 * @param taskArray Pointer to the task.
 */
void hit_taskBufferCreateTask(HitTaskBuffer * bufferP, void * taskArray){

	if(bufferP->null_buffer){
		HitTask * tasks = (HitTask*) taskArray;

		int i;
		for(i=0; i<bufferP->n_types; i++){
			tasks[i] = HIT_TASK_NULL;
		}

		return;
	}


	if(bufferP->empty == 1){
		hit_error_here("Trying to retrieve tasks from a empty buffer");
	}

	HitTask * tasks = (HitTask*) taskArray;

	int i;
	for(i=0; i<bufferP->n_types; i++){

		HitBufferInfo info;
		MPI_Unpack(bufferP->buffer, bufferP->buffer_size, &bufferP->position, &info, 1, HitBufferInfoType, bufferP->comm);

		//printf("Retreving task %d elems, size %d \n",info.n_elems,bufferP->size[i]);

		hit_taskAllocInternal(&tasks[i],info.n_elems, (size_t) bufferP->size[i]);
		tasks[i].weight = info.weight;

		/*
		printf("Buffer %p\n",bufferP->buffer);
		printf("Buffer size %d\n",bufferP->buffer_size);
		printf("Position %d\n",bufferP->position);
		printf("Data %p\n",tasks[i].data);
		printf("Nelems %d\n",info.n_elems);
		printf("Type %d\n",bufferP->type[i]);
		printf("Comm %d\n",bufferP->comm);
*/

		MPI_Unpack(bufferP->buffer, bufferP->buffer_size, &bufferP->position, tasks[i].data, info.n_elems, bufferP->type[i], bufferP->comm);

	}

	hit_taskBufferCheckNext(bufferP);

}














