/**
 * Physical topology contexts
 * Management of the stack of physical topologies 
 * 
 * @file hit_topContext.c
 * @version 1.0
 * @author Arturo Gonzalez-Escribano
 * @date Feb 2013
 *
 */

/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2007-2013, Arturo Gonzalez Escribano
 * All Rights Reserved.
 */

#include<hit_allocP.h>
#include<hit_topContext.h>

/* 1. POP THE LAST ACTIVE CONTEXT */
void hit_topContextPop( ) {
	/* 0. INTERNAL ERROR */
	if ( HIT_TOPOLOGY_INFO == NULL )
		hit_errInternal( __FUNCTION__, "CRITICAL: Root topology info is lost", "", __FILE__, __LINE__);

	/* 1. CHECK IF THE STACK IS EMPTY */
	if ( HIT_TOPOLOGY_INFO->next == NULL )
		hit_errInternal( __FUNCTION__, "Stack of topology contexts is empty", "", __FILE__, __LINE__);

	/* 2. POP CONTEXT */
	/*
	HitPTopology *toPop = HIT_TOPOLOGY_INFO.next;
	HIT_TOPOLOGY_INFO = *( HIT_TOPOLOGY_INFO.next );
	hit_free( toPop );
	*/
	HitPTopology *toPop = HIT_TOPOLOGY_INFO->next;
	HIT_TOPOLOGY_INFO = HIT_TOPOLOGY_INFO->next;
	hit_free( toPop );
}


/* 2. CREATE AND PUSH A NEW CONTEXT WITH THE PROCESSORS IN THE SAME GROUP OF A LAYOUT */
void hit_topContextPushGroup( HitLayout lay ) {
	/* 1. PUSH CURRENT CONTEXT IN THE STACK */
	HitPTopology *old;
	hit_malloc( old, sizeof(HitPTopology), HitPTopology* );
	//*old = HIT_TOPOLOGY_INFO;
	old = HIT_TOPOLOGY_INFO;

	/* 2. GENERATE NEW CONTEXT */
	HIT_TOPOLOGY_INFO = hit_ptopSplit( HIT_TOPOLOGY_INFO, lay.group );
	HIT_TOPOLOGY_INFO->next = old;
}


