#include <stdio.h>
#include <stdlib.h>
#include <hitmap.h>


hit_tileNewType( double );

void cannonsMM( int n, int m );

int main(){

	cannonsMM(5,5);
	return 0;

}

void matrixBlockProduct( HitTile_double tileA, HitTile_double tileB, HitTile_double tileC );


void cannonsMM( int n, int m )  {
   /* 1. DECLARE FULL MATRICES  WITHOUT MEMORY */
   HitTile_double A, B, C;
   hit_tileDomainShape( &A, HIT_NONHIERARCHICAL, hit_shape(  2, hit_sig(0,n,1), hit_sig(0,m,1) ));
   hit_tileDomainShape( &B, HIT_NONHIERARCHICAL, hit_shape(  2, hit_sig(0,m,1), hit_sig(0,n,1) ));
   hit_tileDomainShape( &C, HIT_NONHIERARCHICAL, hit_shape(  2, hit_sig(0,n,1), hit_sig(0,n,1) ));
    /* 2. CREATE VIRTUAL TOPOLOGY */
    HitTopology topo = hit_topology( plug_topSquare );

    /* 3. COMPUTE PARTITIONS */
    HitLayout layA = hit_layout(plug_layBlocks, topo, hit_tileShape( A ) );
    HitLayout layB = hit_layout(plug_layBlocks, topo, hit_tileShape( B ) );
    HitLayout layC = hit_layout(plug_layBlocks, topo, hit_tileShape( C ) );

    /* 4. CREATE AND ALLOCATE TILES */
    HitTile_double tileA, tileB, tileC, copyTileA, copyTileB;
    hit_tileSelectNoBoundary( &tileA, &A, hit_layMaxShape(layA));
    hit_tileSelectNoBoundary( &tileB, &B, hit_layMaxShape(layB) );
    hit_tileSelect( &tileC, &C,  hit_layShape(layC) );
    hit_tileAlloc( &tileA ); hit_tileAlloc( &tileB ); hit_tileAlloc( &tileC );
    hit_tileClone( &copyTileA, &tileA ); hit_tileClone( &copyTileB, &tileB );

    /* 5. INITIALIZE MATRICES */
    hit_tileFileRead( &tileA, "matrixA.dat" );
    hit_tileFileRead( &tileB, "matrixB.dat" );
    int aux=0; hit_tileFill( &tileC, &aux );

    /* 6. INITIAL ALIGNMENT PHASE */
    HitCom commRow, commColumn;
    // Que tag?
    commRow = hit_comShiftDim( layA, 1, hit_layRank(layA,0), &tileA, &copyTileA, HIT_DOUBLE , 9999 );
    commColumn = hit_comShiftDim( layB, 0, hit_layRank(layB,1), &tileB, &copyTileB, HIT_DOUBLE,  9999 );
    hit_comDo( commRow );
    hit_comDo( commColumn );
    hit_comFree( commRow );
    hit_comFree( commColumn );

    hit_tileUpdateToAncestor( &copyTileA );
    hit_tileUpdateToAncestor( &copyTileB );

    /* 7. REPEATED COMM PATTERN */
    HitPattern shift = hit_pattern( HIT_PAT_UNORDERED );
    hit_patternAdd( &shift, hit_comShiftDim( layA, 1, 1, &tileA, &copyTileA, HIT_DOUBLE, 9999));
    hit_patternAdd( &shift, hit_comShiftDim( layA, 0, 1, &tileB, &copyTileB, HIT_DOUBLE, 9999));

    /* 8. DO COMPUTATION */
    int loopIndex;
    int loopLimit = max( layA.numActives[0], layB.numActives[1] );
    for (loopIndex = 0; loopIndex < loopLimit-1; loopIndex++) {
       matrixBlockProduct( tileA, tileB, tileC );
       hit_patternDo( shift );
       hit_tileUpdateToAncestor( &copyTileA );
       hit_tileUpdateToAncestor( &copyTileB );
    }
    matrixBlockProduct( tileA, tileB, tileC );

    /* 9. WRITE RESULT */
    hit_tileFileWrite( &tileC, "matrixC.dat" );

    /* 10. FREE RESOURCES */
    hit_patternFree( &shift );
    hit_topFree( topo );
 }
