/*
 * Debug functions for test
 *
 * @version 1.0
 * @date Jun 2010
 * @author Javier Fresno Bausela
 */

/*
 *
 * hit v1.0
 * (c) 2010, Javier Fresno Bausela
 */

#include "debug.h"


// Dump a integer array
void dumpArray(int * array,int dims, const char * format, ...){

	char name[1024];

	va_list args;
	va_start( args, format );
	vsprintf( name, format, args );
	va_end( args );


	if(dims == 0){
		printf("%s [X]\n",name);
		return;
	}

	int i;
	printf("%s [%d",name,array[0]);
	for(i=1;i<dims;i++) printf(",%d",array[i]);
	printf("]");

}

// Dump a HitTopology
void dumpTopoCommon( HitTopology topo, char * name ) {

	printf("\n Topology: %s\n",name);
	printf("---------------------------\n");
	printf("NumDims: %2d, ", topo.numDims );

	dumpArray(topo.card,topo.numDims,"Card");
	printf("\n");

}

void dumpTopoSelf( HitTopology topo) {

	printf("Lineal Rank: %2d", topo.linearRank );

	dumpArray(topo.self.rank,topo.numDims,", Self");


	printf(", Active: %d\n", topo.active );
}


//#ifdef WORKING_WITH_GROUPS
void dumpListLayout(HitLayout lay, char * name, double * weights){

	int group, i;

	printf("\n ListLayouty: %s\n",name);
	printf("---------------------------\n");
	printf("Processors: %d, Elements: %d -> Groups: %d\n",
			lay.topo.pTopology.numProcs,
			lay.info.layoutList.numElementsTotal,
			lay.info.layoutList.numGroups);
	printf("---------------------------\n");

	for(group=0;group<lay.info.layoutList.numGroups;group++){

		int leader = lay.info.layoutList.groups[group].leader;
		int nprocs = lay.info.layoutList.groups[group].numProcs;

		printf("Group %d\n   - Num Processors: %d",group,nprocs);

		printf("\n   - Processors: ");
		for(i=0;i<nprocs;i++){
			printf("%d",leader+i);
			if(i!=nprocs-1) printf(", ");
		}

		double weight = 0;
		printf("\n   - Elements: ");
		for(i=0;i<lay.info.layoutList.numElementsTotal;i++){
			if(lay.info.layoutList.assignedGroups[i] == group){
				printf("%d",i);
				printf(", ");
				weight += weights[i];
			}

		}
		printf(" sum weights: %.1f\n",weight);


	}


}
//#endif

void dumpShape(HitShape sh){



	int i;
	printf("[%d:%d:%d",hit_shapeSig(sh,0).begin,hit_shapeSig(sh,0).end,hit_shapeSig(sh,0).stride);
	for(i=1;i<hit_shapeDims(sh);i++) printf(",%d:%d:%d",hit_shapeSig(sh,i).begin,hit_shapeSig(sh,i).end,hit_shapeSig(sh,i).stride);
	printf("] \t cards: [%d",hit_sigCard(hit_shapeSig(sh,0)));
	for(i=1;i<hit_shapeDims(sh);i++) printf(",%d",hit_sigCard(hit_shapeSig(sh,i)));
	printf("]\n");
}


void dumpSignatureLayoutCommon( HitLayout lay , char * name ) {

	printf("\n LayoutSig: %s\n",name);
	printf("---------------------------\n");
	dumpArray(lay.numActives,hit_shapeDims( hit_layShape( lay ) ),"Num Actives: ");
	printf(", Original shape: "); dumpShape(lay.origShape);
}


void dumpSignatureLayoutSelf( HitLayout lay ) {

	printf(", Active: %2d, ", lay.active );
	dumpArray(lay.topo.self.rank,lay.topo.numDims,"Topo ranks");
	printf(", Group: %2d, Leader: %2d",lay.group,lay.leader);
	printf(", My shape: "); dumpShape(hit_layShape(lay));

}




// All prints
void printfAllSec( const char* format, ... ) {

	int i = 0;

	for(i=0;i<HIT_TOPOLOGY_INFO.numProcs;i++){

		if(HIT_TOPOLOGY_INFO.selfRank == i){

			fprintf( stdout, "%2d: ",i);
			va_list args;
			va_start( args, format );
			vfprintf( stdout, format, args );
			va_end( args );

		}
		MPI_Barrier(*( (MPI_Comm*) HIT_TOPOLOGY_INFO.lowLevel));
		fflush(stdout);
		struct timespec ts = {0,10000};
		nanosleep(&ts,&ts);
	}
 }

// All dumps a integer array
void dumpArrayAllSec(int * array,int dims, char * format, ...){

	char name[1024];

	va_list args;
	va_start( args, format );
	vsprintf( name, format, args );
	va_end( args );

	char string[1024];
	char * s = string;

	if(dims == 0){
		s += sprintf(s,"%s [X]\n",name);
		return;
	}

	int i;
	s += sprintf(s,"%s [%d",name,array[0]);
	for(i=1;i<dims;i++) s += sprintf(s,",%d",array[i]);
	s += sprintf(s,"]\n");

	printfAllSec("%s",string);

}


void dumpSparseShape(HitShape shape){

	HitCShape s = shape.info.csr;

	int i,j;

	printf("vertices: %d.   ",s.nvertices);

	for(i=0;i<s.nvertices;i++){

		int begin = s.xadj[i];
		int end = s.xadj[i+1];

		for(j=begin;j<end;j++){
			printf(" %d ", s.adjncy[j]);
		}
		printf("|");

	}

	printf("   Names: ");
	for(i=0;i<s.nvertices;i++)
		printf(" %d ", s.globalNames[i]);


	printf("\n");
}

