
#include "debug.h"
#include <stdio.h>
#include <stdlib.h>
#include <hitmap.h>
#include <hit_com.h>
#include <unistd.h>
#include <time.h>


hit_tileNewType(int);

// Main function
int main(int argc, char ** argv) {

	hit_comInit(&argc,&argv);

	HitShape shape_global = HIT_CSR_SHAPE_NULL;
	HitShape shape, ext_shape;
	HitTile_int global;
	HitTile_int local;

	// Only root has the global values
	if(hit_Rank == 0){

		hit_cShapeAddEdge(&shape_global,9,5);
		hit_cShapeAddEdge(&shape_global,3,5);
		hit_cShapeAddEdge(&shape_global,2,5);
		hit_cShapeAddEdge(&shape_global,7,9);
		hit_cShapeAddEdge(&shape_global,2,4);

		dumpSparseShape(shape_global);

		hit_gcTileDomainShape(&global, sizeof(int), shape_global, 0);
		hit_tileAlloc(&global);

		hit_tileSparseGraphVertexAt(global,9) = 9;
		hit_tileSparseGraphVertexAt(global,3) = 3;
		hit_tileSparseGraphVertexAt(global,5) = 5;
		hit_tileSparseGraphVertexAt(global,2) = 2;
		hit_tileSparseGraphVertexAt(global,7) = 7;
		hit_tileSparseGraphVertexAt(global,4) = 4;


	}

	// Create the topologoy object.
	HitTopology topo = hit_topology(plug_topPlain);

	// Distribute the graph among the processors.
	HitLayout lay = hit_layout(plug_layMetis,topo,&shape_global);

	// Get the shapes for the local and local extended graphs.
	ext_shape = hit_layExtendedShape(lay);
	shape = hit_layShape(lay);

	// Allocate local shape
	hit_gcTileDomainShape(&local, sizeof(int), ext_shape, 0);
	hit_tileAlloc(&local);

	sleep(hit_Rank);
	printf("Proc %d ",hit_Rank); dumpSparseShape(shape);


	HitCom comA = hit_comSparseScatter(lay, &global, &local, HIT_INT);
	hit_comDo(&comA);

	HitCom comB = hit_comSparseUpdate(lay, &local, HIT_INT);
	hit_comDo(&comB);

	sleep(hit_Rank);

	// Iterate trough all the vertices.
	int vertex;
	hit_cShapeVertexIterator(vertex,ext_shape) {

		int gvertex = hit_cShapeVertexToGlobal(ext_shape,vertex);

		printf("Proc: %d, item %d, value %d\n", hit_Rank  ,gvertex,	hit_tileSparseGraphVertexAt(local,gvertex));

	}

	hit_comFree(comA);
	hit_comFree(comB);

	hit_tileFree(local);
	if(hit_Rank == 0) hit_tileFree(global);

	hit_layoutFree(lay);

	hit_shapeFree(shape_global);

	hit_comFinalize();

	return 0;
}
