/*
* Test of hit_com
*
* Hit v1.0
* (c) 2010, Arturo González Escribano
*           Javier Fresno Bausela
*           Carlos de Blas Cartón
*/
#include <stdio.h>
#include <stdlib.h>
#include <hitmap.h>
#include <unistd.h>
#include <time.h>

//#define printfd(...) printf(__VA_ARGS__)
#define printfd(...) { if( HIT_TOPOLOGY_INFO.selfRank == 0 ) { printf(__VA_ARGS__); fflush(stdout); }}


// All prints
void printfAllSec( const char* format, ... ) {

	int i = 0;

	for(i=0;i<HIT_TOPOLOGY_INFO.numProcs;i++){

		if(HIT_TOPOLOGY_INFO.selfRank == i){

			fprintf( stdout, "%2d: ",i);
			va_list args;
			va_start( args, format );
			vfprintf( stdout, format, args );
			va_end( args );

		}
		MPI_Barrier(*( (MPI_Comm*) HIT_TOPOLOGY_INFO.lowLevel));
		fflush(stdout);
		struct timespec ts = {0,10000};
		nanosleep(&ts,&ts);
	}
 }

// Dump a integer array
void dumpArray(int * array,int dims, const char * format, ...){

	char name[1024];

	va_list args;
	va_start( args, format );
	vsprintf( name, format, args );
	va_end( args );


	if(dims == 0){
		printfd("%s [X]\n",name);
		return;
	}

	int i;
	printfd("%s [%d",name,array[0]);
	for(i=1;i<dims;i++) printfd(",%d",array[i]);
	printfd("]\n");

}

// All dumps a integer array
void dumpArrayAllSec(int * array,int dims, char * format, ...){

	char name[1024];

	va_list args;
	va_start( args, format );
	vsprintf( name, format, args );
	va_end( args );

	char string[1024];
	char * s = string;

	if(dims == 0){
		s += sprintf(s,"%s [X]\n",name);
		return;
	}

	int i;
	s += sprintf(s,"%s [%d",name,array[0]);
	for(i=1;i<dims;i++) s += sprintf(s,",%d",array[i]);
	s += sprintf(s,"]\n");

	printfAllSec("%s",string);

}

// Dump a HitTopology
void dumpTopology( HitTopology topo, char * name ) {

	printfd("\n Topology: %s\n",name);
	printfd("---------------------------\n");
	printfd("NumDims:\t%d\n", topo.numDims );

	dumpArray(topo.card,topo.numDims,"Card");
	dumpArray(topo.self.rank,topo.numDims,"Self");

	printfd("Lineal Rank:\t%d\n", topo.linearRank );
	printfd("Active:\t%d\n", topo.active );
}



void dumpShape(HitShape sh){

	int i;
	printfd("[%d:%d:%d",sh.sig[0].begin,sh.sig[0].end,sh.sig[0].stride);
	for(i=1;i<sh.numDims;i++) printfd(",%d:%d:%d",sh.sig[i].begin,sh.sig[i].end,sh.sig[i].stride);
	printfd("] \t cards: [%d",hit_sigCard(sh.sig[0]));
	for(i=1;i<sh.numDims;i++) printfd(",%d",hit_sigCard(sh.sig[i]));
	printfd("]\n");
}

void dumpShapeAllSec(HitShape sh){

	char string[1024];
	char * s = string;

	int i;
	s += sprintf(s,"[%d:%d:%d",sh.sig[0].begin,sh.sig[0].end,sh.sig[0].stride);
	for(i=1;i<sh.numDims;i++) s += sprintf(s,",%d:%d:%d",sh.sig[i].begin,sh.sig[i].end,sh.sig[i].stride);
	s += sprintf(s,"] \t cards: [%d",hit_sigCard(sh.sig[0]));
	for(i=1;i<sh.numDims;i++) s += sprintf(s,",%d",hit_sigCard(sh.sig[i]));
	s += sprintf(s,"]\n");

	printfAllSec("%s",string);

}


void dumpLayoutSig( HitLayout lay , char * name ) {

	printfd("\n LayoutSig: %s\n",name);
	printfd("---------------------------\n");
	printfd("Active:\t%d, ", lay.active );
	printfd("NumDims:\t%d, ", lay.info.layoutSig.shape.numDims );
	dumpArray(lay.numActives,lay.info.layoutSig.shape.numDims,"numActives");
	dumpArray(lay.topo.self.rank,lay.topo.numDims,"topo self");
	dumpArray(lay.group.rank,lay.info.layoutSig.shape.numDims,"group (also know as self Active)");
	printfd("Restrict:\t%d\n", lay.info.layoutSig.restrictToDim );
	printfd("My shape: "); dumpShape(lay.info.layoutSig.shape);
}


int arrayCmp(int* a, int* b, int dims){

	//printfd("** Compare ** \n");
	//dumpArray(a,dims,"a");
	//dumpArray(b,dims,"b");

	int i;
	for(i=0;i<dims;i++){
		if(a[i] != b[i]) return 0;
	}
	return 1;
}





int test1(int argc, char *argv[]);
int test2(int argc, char *argv[]);

int main(int argc, char *argv[]){

	test1(argc,argv);
	//test2(argc,argv);

	return 0;
}

hit_tileNewType(int);

int test1(int argc, char *argv[]){

	// Init Hitmap Comm
	hit_comInit(&argc,&argv);

	// Make a shape
	HitShape sh = hit_shape(2, hit_sig(0,1,1),hit_sig(0,10,1));
	printfd("Original Shape");
	dumpShape(sh);

	// Topology
	HitTopology topo = hit_topology(plug_topArray2D);

	dumpTopology(topo,"Array 2D");
	printfd("\n");

	printfAllSec("Topology active: %d\n",topo.active);
	printfd("\n");
	dumpArrayAllSec(topo.self.rank,2,"my coords");
	printfd("\n");

	// Layout
	printfd("layout start\n");
	HitLayout lay = hit_layout(plug_layBlocksF,topo,sh);
	hit_layWrapNeighbors(&lay);
	printfd("layout end\n");

	dumpLayoutSig(lay,"BlocksF");
	printfd("\n");


	printfAllSec("Layout active: %d\n",lay.active);
	printfd("\n");

	HitRanks activeRank = hit_layRanksToActiveRanksInternal(lay,topo.self);
	dumpArrayAllSec(activeRank.rank,2,"my coords layout");
	printfd("\n");

	HitShape mysh = hit_layShape(lay);

	dumpShapeAllSec(mysh);
	printfd("\n");

	printfAllSec("Topology rank: %d, Layout rank: %d\n",topo.pTopology.selfRank,lay.pTopology[0].selfRank);
	printfd("\n");

	if(lay.active){
		printfAllSec("Creating Tile\n");
	} else {
		printfAllSec("I don't do anything\n");
	}
	printfd("\n");

	// Creating Tile
	HitTile_int tile;
	hit_tileDomainShape(&tile,sizeof(int),HIT_NONHIERARCHICAL,mysh);
	hit_tileAlloc(&tile);


	int i,j;
	for (i=hit_tileDimBegin(tile,0); i<=hit_tileDimEnd(tile,0); i++)
			for (j=hit_tileDimBegin(tile,1); j<=hit_tileDimEnd(tile,1); j++){
				int value = topo.linearRank;
				hit_tileElemAtArrayCoords2(tile,i,j) = value;
			}


	if(lay.active){
		printfAllSec("Creating Aux Tile\n");
	} else {
		printfAllSec("I don't do anything\n");
	}
	printfd("\n");


	char message[1024];
	char * m = message;
	for (i=hit_tileDimBegin(tile,0); i<=hit_tileDimEnd(tile,0); i++){
			for (j=hit_tileDimBegin(tile,1); j<=hit_tileDimEnd(tile,1); j++){
				int value = hit_tileElemAtArrayCoords2(tile,i,j);

				m += sprintf(m,"%d ",value);
			}
			m += sprintf(m,"\n              ");
	}

	if(lay.active){
		printfAllSec("I have:   %s\n",message);
	} else {
		printfAllSec("I don't do anything\n");
	}
	printfd("\n");


	HitShape aux_sh = hit_layMaxShape(lay);
	printfd("Max Shape\n");
	dumpShapeAllSec(aux_sh);
	printfd("\n");

	HitTile_int aux;
	
	hit_tileDomainShape(&aux,sizeof(int),HIT_NONHIERARCHICAL,aux_sh);
	hit_tileAlloc(&aux);


	if(lay.active){
		HitRanks neig;
		neig = hit_layNeighbor(lay,0,+1);
		dumpArrayAllSec(neig.rank,2,"Neighbor dim 0, shift +1: %d:",hit_topRankInternal(topo,neig));
		neig = hit_layNeighbor(lay,0,-1);
		dumpArrayAllSec(neig.rank,2,"Neighbor dim 0, shift -1: %d:",hit_topRankInternal(topo,neig));
	} else {
		printfAllSec("I don't do anything\n");
		printfAllSec("I don't do anything\n");
	}
	printfd("\n");
	


	//debuggeado: falla la funcion hit_comType
	HitCom sr = hit_comSendRecv(lay, hit_layNeighbor(lay,0,1), &tile, hit_layNeighbor(lay,0,-1), &aux, HIT_INT); //<-- error aqui
	hit_comDo(&sr);

	char received[1024];
	char * r = received;
	for (i=hit_tileDimBegin(aux,0); i<=hit_tileDimEnd(aux,0); i++){
			for (j=hit_tileDimBegin(aux,1); j<=hit_tileDimEnd(aux,1); j++){
				int value = hit_tileElemAtArrayCoords2(aux,i,j);

				r += sprintf(r,"%d ",value);
			}
			r += sprintf(r,"\n              ");
	}

	if(lay.active){
		printfAllSec("Received: %s\n",received);
	} else {
		printfAllSec("I don't do anything\n");
	}
	printfd("\n");


	hit_comFinalize(); //<-- error aqui

	return 0;
}

