/*
* test of supertiles communications
*
* Hit v1.0
* (c) 2010, Arturo González Escribano
*           Javier Fresno Bausela
*           Carlos de Blas Cartón
*/


#include <stdio.h>
#include <hitmap.h>
#include <time.h>
#include <unistd.h>


//#define printfd(...) printf(__VA_ARGS__)
#define printfd(...) { if( HIT_TOPOLOGY_INFO.selfRank == 0 ) { printf(__VA_ARGS__); fflush(stdout); }}

hit_tileNewType(int);

// All prints
void printfAllSec( const char* format, ... ) {

	int i = 0;

	for(i=0;i<HIT_TOPOLOGY_INFO.numProcs;i++){

		if(HIT_TOPOLOGY_INFO.selfRank == i){

			fprintf( stdout, "%2d: ",i);
			va_list args;
			va_start( args, format );
			vfprintf( stdout, format, args );
			va_end( args );

		}
		MPI_Barrier(*( (MPI_Comm*) HIT_TOPOLOGY_INFO.lowLevel));
		fflush(stdout);
		struct timespec ts = {0,10000};
		nanosleep(&ts,&ts);
	}
 }

// Dump a integer array
void dumpArray(int * array,int dims, const char * format, ...){

	char name[1024];

	va_list args;
	va_start( args, format );
	vsprintf( name, format, args );
	va_end( args );


	if(dims == 0){
		printfd("%s [X]\n",name);
		return;
	}

	int i;
	printfd("%s [%d",name,array[0]);
	for(i=1;i<dims;i++) printfd(",%d",array[i]);
	printfd("]\n");

}

// All dumps a integer array
void dumpArrayAllSec(int * array,int dims, const char * format, ...){

	char name[1024];

	va_list args;
	va_start( args, format );
	vsprintf( name, format, args );
	va_end( args );

	char string[1024];
	char * s = string;

	if(dims == 0){
		s += sprintf(s,"%s [X]\n",name);
		return;
	}

	int i;
	s += sprintf(s,"%s [%d",name,array[0]);
	for(i=1;i<dims;i++) s += sprintf(s,",%d",array[i]);
	s += sprintf(s,"]\n");

	printfAllSec("%s",string);

}

// Dump a HitTopology
void dumpTopology( HitTopology topo, const char * name ) {

	printfd("\n Topology: %s\n",name);
	printfd("---------------------------\n");
	printfd("NumDims:\t%d\n", topo.numDims );

	dumpArray(topo.card,topo.numDims,"Card");
	dumpArray(topo.self.rank,topo.numDims,"Self");

	printfd("Lineal Rank:\t%d\n", topo.linearRank );
	printfd("Active:\t%d\n", topo.active );
}



void dumpShape(HitShape sh){

	int i;
	printfd("[%d:%d:%d",sh.sig[0].begin,sh.sig[0].end,sh.sig[0].stride);
	for(i=1;i<sh.numDims;i++) printfd(",%d:%d:%d",sh.sig[i].begin,sh.sig[i].end,sh.sig[i].stride);
	printfd("] \t cards: [%d",hit_sigCard(sh.sig[0]));
	for(i=1;i<sh.numDims;i++) printfd(",%d",hit_sigCard(sh.sig[i]));
	printfd("]\n");
}

void dumpShapeAllSec(HitShape sh){

	char string[1024];
	char * s = string;

	int i;
	s += sprintf(s,"[%d:%d:%d",sh.sig[0].begin,sh.sig[0].end,sh.sig[0].stride);
	for(i=1;i<sh.numDims;i++) s += sprintf(s,",%d:%d:%d",sh.sig[i].begin,sh.sig[i].end,sh.sig[i].stride);
	s += sprintf(s,"] \t cards: [%d",hit_sigCard(sh.sig[0]));
	for(i=1;i<sh.numDims;i++) s += sprintf(s,",%d",hit_sigCard(sh.sig[i]));
	s += sprintf(s,"]\n");

	printfAllSec("%s",string);

}


void dumpLayoutSig( HitLayout lay , const char * name ) {

	printfd("\n LayoutSig: %s\n",name);
	printfd("---------------------------\n");
	printfd("Active:\t%d, ", lay.active );
	printfd("NumDims:\t%d, ", lay.info.layoutSig.shape.numDims );
	dumpArray(lay.numActives,lay.info.layoutSig.shape.numDims,"numActives");
	dumpArray(lay.topo.self.rank,lay.topo.numDims,"topo self");
	dumpArray(lay.group.rank,lay.info.layoutSig.shape.numDims,"group (also know as self Active)");
	printfd("Restrict:\t%d\n", lay.info.layoutSig.restrictToDim );
	printfd("My shape: "); dumpShape(lay.info.layoutSig.shape);
}


int arrayCmp(int* a, int* b, int dims){

	//printfd("** Compare ** \n");
	//dumpArray(a,dims,"a");
	//dumpArray(b,dims,"b");

	int i;
	for(i=0;i<dims;i++){
		if(a[i] != b[i]) return 0;
	}
	return 1;
}


int main(int argc, char **argv) {
       // Init Hitmap Comm
	hit_comInit(&argc,&argv);

	// Make a shape
	HitShape sh = hit_shape(1, hit_sig(0,9,1));
	printfd("Original Shape");
	dumpShape(sh);

	// Topology
	HitTopology topo = hit_topology(plug_topPlain);

	dumpTopology(topo,"plain");
	printfd("\n");

	printfAllSec("Topology active: %d\n",topo.active);
	printfd("\n");
	dumpArrayAllSec(topo.self.rank,1,"my coords");
	printfd("\n");

	if(hit_NProcs!=2) {
                if( HIT_TOPOLOGY_INFO.selfRank == 0 ) printf("needed 2 processes\n");
                hit_comFinalize();
                return 0;
        }


	// Layout
	printfd("layout start\n");
	HitLayout lay = hit_layout(plug_layBlocksF,topo,sh);
	hit_layWrapNeighbors(&lay);
	printfd("layout end\n");

	dumpLayoutSig(lay,"BlocksF");
	printfd("\n");


	printfAllSec("Layout active: %d\n",lay.active);
	printfd("\n");

	HitRanks activeRank = hit_layRanksToActiveRanksInternal(lay,topo.self);
	dumpArrayAllSec(activeRank.rank,1,"my coords layout");
	printfd("\n");

	HitShape mysh = hit_layShape(lay);

	dumpShapeAllSec(mysh);
	printfd("\n");

	printfAllSec("Topology rank: %d, Layout rank: %d\n",topo.pTopology.selfRank,lay.pTopology[0].selfRank);
	printfd("\n");

	if(lay.active){
		printfAllSec("Creating Tile\n");
	} else {
		printfAllSec("I don't do anything\n");
	}
	printfd("\n");

	// Creating Tile
	HitTile_int pat;
	hit_tileDomainAlloc(&pat,sizeof(int),HIT_NONHIERARCHICAL,2,3,3);

	HitTile_HitTile tile;
	hit_tileDomainShapeAlloc(&tile,sizeof(HitTile),HIT_HIERARCHICAL,mysh);
	
	int i=0;
	hit_tileFill(&pat,&i);
	hit_tileFill(&tile,&pat);

	HitCom hiercom;
        HitTile_int ti;	
        int j,k,contador;
	if (hit_Rank == 0) {
	  hiercom=hit_comSendSelect(lay, hit_ranks(1,1), &tile, hit_shape(1,hit_sig(2,3,1)), HIT_COM_TILECOORDS, HIT_INT);
	  contador=0;
	  
	  for (j=hit_tileDimBegin(tile,0); j<=hit_tileDimEnd(tile,0); j++) {
		  
	      ti=*(HitTile_int*)&hit_tileGet(tile,1,j);
	      for (k=0; k<hit_tileCard(ti); k++) {
			hit_tileSet(ti,1,contador++,k);
		  }
	  }
		
	  hit_comDo(&hiercom);
	  
	} else {
	  hiercom=hit_comRecvSelect(lay, hit_ranks(1,0), &tile, hit_shape(1,hit_sig(1,2,1)), HIT_COM_TILECOORDS, HIT_INT);
		
		hit_comDo(&hiercom);
		
		for (j=hit_tileDimBegin(tile,0); j<=hit_tileDimEnd(tile,0); j++) {
			
			ti=*(HitTile_int*)&hit_tileGetArrayCoords(tile,1,j);
			printf("tile %d = ",j);
			for (k=0; k<hit_tileCard(ti); k++) {
				contador=hit_tileGet(ti,1,k);
				printf("%d ",contador);
			}
			printf("\n");
		}
	}





	hit_comFinalize(); 
	return 0;
}

