/*
* Test of dimensional collective broadcast
*
* Hit v1.0
* (c) 2010, Arturo González Escribano
*           Javier Fresno Bausela
*           Carlos de Blas Cartón
*/
#include <stdio.h>
#include <stdlib.h>
#include <hitmap.h>
#include <unistd.h>
#include <time.h>

//#define printfd(...) printf(__VA_ARGS__)
#define printfd(...) { if( HIT_TOPOLOGY_INFO.selfRank == 0 ) { printf(__VA_ARGS__); fflush(stdout); }}


// All prints
void printfAllSec( const char* format, ... ) {

	int i = 0;

	for(i=0;i<HIT_TOPOLOGY_INFO.numProcs;i++){

		if(HIT_TOPOLOGY_INFO.selfRank == i){

			fprintf( stdout, "%2d: ",i);
			va_list args;
			va_start( args, format );
			vfprintf( stdout, format, args );
			va_end( args );

		}
		MPI_Barrier(*( (MPI_Comm*) HIT_TOPOLOGY_INFO.lowLevel));
		fflush(stdout);
		struct timespec ts = {0,10000};
		nanosleep(&ts,&ts);
	}
 }

// All dumps a integer array
void dumpArrayAllSec(int * array,int dims, char * format, ...){

	char name[1024];

	va_list args;
	va_start( args, format );
	vsprintf( name, format, args );
	va_end( args );

	char string[1024];
	char * s = string;

	if(dims == 0){
		s += sprintf(s,"%s [X]\n",name);
		return;
	}

	int i;
	s += sprintf(s,"%s [%d",name,array[0]);
	for(i=1;i<dims;i++) s += sprintf(s,",%d",array[i]);
	s += sprintf(s,"]\n");

	printfAllSec("%s",string);

}



hit_tileNewType(int);


int main(int argc, char *argv[]){


	// Init Hitmap Comm
	hit_comInit(&argc,&argv);

	// Make a shape
	HitShape sh = hit_shape(3, hit_sig(0,9,1),hit_sig(0,9,1),hit_sig(0,9,1));

	// Topology
	HitTopology topo = hit_topology(plug_topArray3D);
	printfAllSec("lineal rank topo: %d/%d\n",topo.pTopology.selfRank,topo.pTopology.numProcs);

	dumpArrayAllSec(topo.self.rank,3,"my coords");
	printfd("\n");

	// Layout
	printfd("layout start\n");

	HitLayout lay = hit_layout(plug_layBlocks,topo,sh);
	hit_layWrapNeighbors(&lay);

	printfAllSec("lineal rank: %d/%d\n",lay.pTopology[0].selfRank,lay.pTopology[0].numProcs);

	hit_comAllowDims(lay);
	printfAllSec("dim %d, lineal rank dim: %d/%d\n",1,lay.pTopology[2].selfRank,lay.pTopology[2].numProcs);

	HitTile_int tile;
	hit_tileDomainAlloc(&tile,sizeof(int),HIT_NONHIERARCHICAL,1,1);

	hit_tileElemAt(tile,1,0) = HIT_TOPOLOGY_INFO.selfRank;
	printfAllSec("I have: %d\n",hit_tileElemAt(tile,1,0));


	HitRanks root = {{0,0,0,-1}};
	HitCom com = hit_comBroadcastDim(lay, 0, root, &tile, HIT_INT);


	hit_comDoBroadcast(&com);


	printfAllSec("Now, I have: %d\n",hit_tileElemAt(tile,1,0));

	hit_layoutFree(lay);


	hit_comFinalize(); 

	return 0;
}

