/*
* Test of hit_layout
*
* Hit v1.0
* (c) 2010, Arturo González Escribano
*           Javier Fresno Bausela
*           Carlos de Blas Cartón
*/
#include <stdio.h>
#include <stdlib.h>
#include <hitmap.h>

// Dump a integer array
void dumpArray(int * array,int dims, char * name){

	if(dims == 0){
		printf("%s [X]\n",name);
		return;
	}

	int i;
	printf("%s [%d",name,array[0]);
	for(i=1;i<dims;i++) printf(",%d",array[i]);
	printf("]\n");

}

// Dump a HitTopology
void dumpTopology( HitTopology topo, char * name ) {

	printf("\n Topology: %s\n",name);
	printf("---------------------------\n");
	printf("NumDims:\t%d\n", topo.numDims );

	dumpArray(topo.card,topo.numDims,"Card");
	dumpArray(topo.self.rank,topo.numDims,"Self");

	printf("Lineal Rank:\t%d\n", topo.linearRank );
	printf("Active:\t%d\n", topo.active );
}



void dumpShape(HitShape sh){

	int i;
	printf("[%d:%d:%d",sh.sig[0].begin,sh.sig[0].end,sh.sig[0].stride);
	for(i=1;i<sh.numDims;i++) printf(",%d:%d:%d",sh.sig[i].begin,sh.sig[i].end,sh.sig[i].stride);
	printf("] \t cards: [%d",hit_sigCard(sh.sig[0]));
	for(i=1;i<sh.numDims;i++) printf(",%d",hit_sigCard(sh.sig[i]));
	printf("]\n");
}


void dumpLayoutSig( HitLayout lay , char * name ) {

	printf("\n LayoutSig: %s\n",name);
	printf("---------------------------\n");
	printf("Active:\t%d, ", lay.active );
	printf("NumDims:\t%d, ", lay.info.layoutSig.shape.numDims );
	dumpArray(lay.numActives,lay.info.layoutSig.shape.numDims,"numActives");
	dumpArray(lay.topo.self.rank,lay.topo.numDims,"topo self");
	dumpArray(lay.group.rank,lay.info.layoutSig.shape.numDims,"group (also know as self Active)");
	printf("Restrict:\t%d\n", lay.info.layoutSig.restrictToDim );
	printf("My shape: "); dumpShape(lay.info.layoutSig.shape);
}


int arrayCmp(int* a, int* b, int dims){

	//printf("** Compare ** \n");
	//dumpArray(a,dims,"a");
	//dumpArray(b,dims,"b");


	int i;
	for(i=0;i<dims;i++){
		if(a[i] != b[i]) return 0;
	}
	return 1;

}


int test1();
int test2();

int main(){

	test1();
	//test2();

	return 0;
}


#define PROCS_SETS 3
int procs_sets[PROCS_SETS] = {4,9,17};

int test1() {



	int  set, nprocs, self;

	// Make a shape
	HitShape sh = hit_shape(2, hit_sig(-5,10,2),hit_sig(1,139,1));

	printf("\n Original shape: ");
	dumpShape(sh);
	printf("==================================\n");



	/* Select the set of processors */
	for(set=0;set<PROCS_SETS;set++){

		nprocs = procs_sets[set];

		self = nprocs / 2;

		printf("\n\n Processors Number: %d, Self: %d\n",nprocs,self);
		printf("==================================\n");

		// Set the Physical topology info
		HIT_TOPOLOGY_INFO.numProcs = nprocs;
		HIT_TOPOLOGY_INFO.selfRank = self;

		// Create the virtual topologies
		HitTopology topo1, topo2;

		topo1 = hit_topology(plug_topPlain);
		topo2 = hit_topology(plug_topArray2D);

		dumpTopology(topo1,"topPlain");
		dumpTopology(topo2,"topArray2D");


		// Create the layouts
		HitLayout lay, lay1a, lay1b, lay2a, lay2b;

		lay1a = hit_layout(plug_layBlocks, topo1, sh);
		lay1b = hit_layout(plug_layBlocksF, topo1, sh);

		lay2a = hit_layout(plug_layBlocksL, topo2, sh);
		lay2b = hit_layout(plug_layCyclic, topo2, sh, 1);

		HitRanks neigh;

		/*************************************************/
//		lay = lay1a;
//		dumpLayoutSig(lay,"topPlain -> layBlocks");
//
//		printf("\nNeighbors\n");
//		neigh = hit_layNeighbor(lay,0,-3);
//		dumpArray(neigh.rank,1,"0: -3");
//		neigh = hit_layNeighbor(lay,0,-2);
//		dumpArray(neigh.rank,1,"0: -2");
//		neigh = hit_layNeighbor(lay,0,-1);
//		dumpArray(neigh.rank,1,"0: -1");
//		neigh = hit_layNeighbor(lay,0, 1);
//		dumpArray(neigh.rank,1,"0:  1");
//		neigh = hit_layNeighbor(lay,0, 2);
//		dumpArray(neigh.rank,1,"0:  2");
//
//		hit_layWrapNeighbors(&lay);
//
//		printf("\nNeighbors Wrap\n");
//		neigh = hit_layNeighbor(lay,0,-3);
//		dumpArray(neigh.rank,1,"0: -3");
//		neigh = hit_layNeighbor(lay,0,-2);
//		dumpArray(neigh.rank,1,"0: -2");
//		neigh = hit_layNeighbor(lay,0,-1);
//		dumpArray(neigh.rank,1,"0: -1");
//		neigh = hit_layNeighbor(lay,0, 1);
//		dumpArray(neigh.rank,1,"0:  1");
//		neigh = hit_layNeighbor(lay,0, 2);
//		dumpArray(neigh.rank,1,"0:  2");


		/*************************************************/
		lay = lay1b;
		dumpLayoutSig(lay,"topPlain -> layBlocksF");

		printf("\nNeighbors\n");
		neigh = hit_layNeighbor(lay,0,-4);
		dumpArray(neigh.rank,1,"0: -4");
		neigh = hit_layNeighbor(lay,0,-3);
		dumpArray(neigh.rank,1,"0: -3");
		neigh = hit_layNeighbor(lay,0,-2);
		dumpArray(neigh.rank,1,"0: -2");
		neigh = hit_layNeighbor(lay,0,-1);
		dumpArray(neigh.rank,1,"0: -1");
		neigh = hit_layNeighbor(lay,0, 0);
		dumpArray(neigh.rank,1,"0:  0");
		if(!arrayCmp(neigh.rank,lay.topo.self.rank,4)) printf("ERROR: I am not myself\n");
		neigh = hit_layNeighbor(lay,0, 1);
		dumpArray(neigh.rank,1,"0:  1");
		neigh = hit_layNeighbor(lay,0, 2);
		dumpArray(neigh.rank,1,"0:  2");
		neigh = hit_layNeighbor(lay,1, 1);
		dumpArray(neigh.rank,1,"1:  1");
		neigh = hit_layNeighbor(lay,1, 2);
		dumpArray(neigh.rank,1,"1:  2");

		hit_layWrapNeighbors(&lay);

		printf("\nNeighbors Wrap\n");
		neigh = hit_layNeighbor(lay,0,-4);
		dumpArray(neigh.rank,1,"0: -4");
		neigh = hit_layNeighbor(lay,0,-3);
		dumpArray(neigh.rank,1,"0: -3");
		neigh = hit_layNeighbor(lay,0,-2);
		dumpArray(neigh.rank,1,"0: -2");
		neigh = hit_layNeighbor(lay,0,-1);
		dumpArray(neigh.rank,1,"0: -1");
		neigh = hit_layNeighbor(lay,0, 0);
		dumpArray(neigh.rank,1,"0:  0");
		if(!arrayCmp(neigh.rank,lay.topo.self.rank,4)) printf("ERROR: I am not myself\n");
		neigh = hit_layNeighbor(lay,0, 1);
		dumpArray(neigh.rank,1,"0:  1");
		neigh = hit_layNeighbor(lay,0, 2);
		dumpArray(neigh.rank,1,"0:  2");
		neigh = hit_layNeighbor(lay,1, 1);
		dumpArray(neigh.rank,1,"1:  1");
		neigh = hit_layNeighbor(lay,1, 2);
		dumpArray(neigh.rank,1,"1:  2");


		/*************************************************/
		lay = lay2a;
		dumpLayoutSig(lay,"topArray2D -> layBlocksL");



		/*************************************************/
		lay = lay2b;
		dumpLayoutSig(lay,"topArray2D -> layCyclic");

	}



	return 0;
}



int test2(){

	int self;
	int nprocs = 9;

	// Make a shape
	HitShape sh = hit_shape(2, hit_sig(-5,10,2),hit_sig(1,139,1));

	for(self=0;self<nprocs;self++){

		// Set the Physical topology info
		HIT_TOPOLOGY_INFO.numProcs = nprocs;
		HIT_TOPOLOGY_INFO.selfRank = self;

		// Create the virtual topologies
		HitTopology topo;
		HitLayout lay;

		topo = hit_topology(plug_topPlain);
		lay = hit_layout(plug_layBlocksF, topo, sh);

		dumpLayoutSig(lay,"topPlain -> layBlocks");

	}


	return 0;
}

