/*
* Test of hit_layout
*
* Hit v1.0
* (c) 2010, Arturo González Escribano
*           Javier Fresno Bausela
*           Carlos de Blas Cartón
*/

#include <stdio.h>
#include <stdlib.h>
#include <hitmap.h>
#include <math.h>

void dumpTopologyDims(HitTopology topo){

	int i;
	int nProcs = 1;
	for(i=0;i<topo.numDims;i++){
		nProcs *= topo.card[i];
	}

	printf("Topology dims: %d x %d x %d x %d, %d Processors\n",
			topo.card[0],
			topo.card[1],
			topo.card[2],
			topo.card[3],
			nProcs);

}

void dumpRanks(HitRanks ranks){

	printf("Ranks: %d x %d x %d x %d\n",
			ranks.rank[0],
			ranks.rank[1],
			ranks.rank[2],
			ranks.rank[3]);

}

void dumpLayoutGroup( HitLayout lay ) {

	int i;
	HitLayoutGroup layg = lay.info.layoutGroup;

	printf("active: %d\n",lay.active);
	printf("leader: %d\n",lay.leader);
	printf("group: (%d,%d,%d,%d)\n",lay.group.rank[0],lay.group.rank[1],
									lay.group.rank[2],lay.group.rank[3]);

	printf("numElementsTotal: %d\n",layg.numElementsTotal);
	printf("assignedGroups: ");
	printf(" %d",layg.assignedGroups[0]);
	for(i=1;i<layg.numElementsTotal;i++){
		printf(", %d",layg.assignedGroups[i]);
	}
	printf("\n");

	printf("numGroups: %d\n",layg.numGroups);
	for(i=0;i<layg.numGroups;i++){
		printf("\tgroup: %d: leader %d, numProcs %d\n",i,layg.groups[i].leader,layg.groups[i].numProcs);
	}

	/*
	printf("assignedActual: ");
	printf(" %d",layg.assignedActual[0]);
	for(i=1;i<layg.numElements;i++){
		printf(", %d",layg.assignedActual[0]);
	}
	printf("\n");
	*/
}



int main(){


	//double weights[7] = {.1,.2,.3,.4,.5,.6,.7};
	double weights[7] = {.1,.8,.3,.4,.5,.3,.7};
	HitShape shape = hit_shape(1,hit_sig(0,5,1));

	HIT_TOPOLOGY_INFO.numProcs = 41;
	HIT_TOPOLOGY_INFO.selfRank = 1;

	HitTopology	topo;
	topo = hit_topology( plug_topArray2D );

	dumpTopologyDims(topo);
	dumpRanks(topo.self);

	//HitLayout lay = hit_layout(plug_layIndependentLB, topo, shape, weights);
	HitLayout lay = hit_layout(plug_layContiguous, topo, shape, weights);

	dumpLayoutGroup(lay);

	return 0;

}
