/*
* Test of hit_pattern
*
* Hit v1.0
* (c) 2010, Arturo González Escribano
*           Javier Fresno Bausela
*           Carlos de Blas Cartón
*/
#include <stdio.h>
#include <hitmap.h>
#include <time.h>
#include <unistd.h>


//#define printfd(...) printf(__VA_ARGS__)
#define printfd(...) { if( HIT_TOPOLOGY_INFO.selfRank == 0 ) { printf(__VA_ARGS__); fflush(stdout); }}

hit_tileNewType(int);

// All prints
void printfAllSec( const char* format, ... ) {

	int i = 0;

	for(i=0;i<HIT_TOPOLOGY_INFO.numProcs;i++){

		if(HIT_TOPOLOGY_INFO.selfRank == i){

			fprintf( stdout, "%2d: ",i);
			va_list args;
			va_start( args, format );
			vfprintf( stdout, format, args );
			va_end( args );

		}
		MPI_Barrier(*( (MPI_Comm*) HIT_TOPOLOGY_INFO.lowLevel));
		fflush(stdout);
		struct timespec ts = {0,10000};
		nanosleep(&ts,&ts);
	}
 }

// Dump a integer array
void dumpArray(int * array,int dims, const char * format, ...){

	char name[1024];

	va_list args;
	va_start( args, format );
	vsprintf( name, format, args );
	va_end( args );


	if(dims == 0){
		printfd("%s [X]\n",name);
		return;
	}

	int i;
	printfd("%s [%d",name,array[0]);
	for(i=1;i<dims;i++) printfd(",%d",array[i]);
	printfd("]\n");

}

// All dumps a integer array
void dumpArrayAllSec(int * array,int dims, const char * format, ...){

	char name[1024];

	va_list args;
	va_start( args, format );
	vsprintf( name, format, args );
	va_end( args );

	char string[1024];
	char * s = string;

	if(dims == 0){
		s += sprintf(s,"%s [X]\n",name);
		return;
	}

	int i;
	s += sprintf(s,"%s [%d",name,array[0]);
	for(i=1;i<dims;i++) s += sprintf(s,",%d",array[i]);
	s += sprintf(s,"]\n");

	printfAllSec("%s",string);

}

// Dump a HitTopology
void dumpTopology( HitTopology topo, const char * name ) {

	printfd("\n Topology: %s\n",name);
	printfd("---------------------------\n");
	printfd("NumDims:\t%d\n", topo.numDims );

	dumpArray(topo.card,topo.numDims,"Card");
	dumpArray(topo.self.rank,topo.numDims,"Self");

	printfd("Lineal Rank:\t%d\n", topo.linearRank );
	printfd("Active:\t%d\n", topo.active );
}



void dumpShape(HitShape sh){

	int i;
	printfd("[%d:%d:%d",sh.sig[0].begin,sh.sig[0].end,sh.sig[0].stride);
	for(i=1;i<sh.numDims;i++) printfd(",%d:%d:%d",sh.sig[i].begin,sh.sig[i].end,sh.sig[i].stride);
	printfd("] \t cards: [%d",hit_sigCard(sh.sig[0]));
	for(i=1;i<sh.numDims;i++) printfd(",%d",hit_sigCard(sh.sig[i]));
	printfd("]\n");
}

void dumpShapeAllSec(HitShape sh){

	char string[1024];
	char * s = string;

	int i;
	s += sprintf(s,"[%d:%d:%d",sh.sig[0].begin,sh.sig[0].end,sh.sig[0].stride);
	for(i=1;i<sh.numDims;i++) s += sprintf(s,",%d:%d:%d",sh.sig[i].begin,sh.sig[i].end,sh.sig[i].stride);
	s += sprintf(s,"] \t cards: [%d",hit_sigCard(sh.sig[0]));
	for(i=1;i<sh.numDims;i++) s += sprintf(s,",%d",hit_sigCard(sh.sig[i]));
	s += sprintf(s,"]\n");

	printfAllSec("%s",string);

}


void dumpLayoutSig( HitLayout lay , const char * name ) {

	printfd("\n LayoutSig: %s\n",name);
	printfd("---------------------------\n");
	printfd("Active:\t%d, ", lay.active );
	printfd("NumDims:\t%d, ", lay.info.layoutSig.shape.numDims );
	dumpArray(lay.numActives,lay.info.layoutSig.shape.numDims,"numActives");
	dumpArray(lay.topo.self.rank,lay.topo.numDims,"topo self");
	dumpArray(lay.group.rank,lay.info.layoutSig.shape.numDims,"group (also know as self Active)");
	printfd("Restrict:\t%d\n", lay.info.layoutSig.restrictToDim );
	printfd("My shape: "); dumpShape(lay.info.layoutSig.shape);
}


int arrayCmp(int* a, int* b, int dims){

	//printfd("** Compare ** \n");
	//dumpArray(a,dims,"a");
	//dumpArray(b,dims,"b");

	int i;
	for(i=0;i<dims;i++){
		if(a[i] != b[i]) return 0;
	}
	return 1;
}


int main(int argc, char **argv) {
       // Init Hitmap Comm
	hit_comInit(&argc,&argv);

	// Make a shape
	HitShape sh = hit_shape(1, hit_sig(0,9,1));
	printfd("Original Shape");
	dumpShape(sh);

	// Topology
	HitTopology topo = hit_topology(plug_topPlain);

	dumpTopology(topo,"plain");
	printfd("\n");

	printfAllSec("Topology active: %d\n",topo.active);
	printfd("\n");
	dumpArrayAllSec(topo.self.rank,1,"my coords");
	printfd("\n");

	if(hit_NProcs!=1) {
                if( HIT_TOPOLOGY_INFO.selfRank == 0 ) printf("needed 1 process\n");
                hit_comFinalize();
                return 0;
        }


	// Layout
	printfd("layout start\n");
	HitLayout lay = hit_layout(plug_layBlocksF,topo,sh);
	hit_layWrapNeighbors(&lay);
	printfd("layout end\n");

	dumpLayoutSig(lay,"BlocksF");
	printfd("\n");


	printfAllSec("Layout active: %d\n",lay.active);
	printfd("\n");

	HitRanks activeRank = hit_layRanksToActiveRanksInternal(lay,topo.self);
	dumpArrayAllSec(activeRank.rank,1,"my coords layout");
	printfd("\n");

	HitShape mysh = hit_layShape(lay);

	dumpShapeAllSec(mysh);
	printfd("\n");

	printfAllSec("Topology rank: %d, Layout rank: %d\n",topo.pTopology.selfRank,lay.pTopology[0].selfRank);
	printfd("\n");

	if(lay.active){
		printfAllSec("Creating Tile\n");
	} else {
		printfAllSec("I don't do anything\n");
	}
	printfd("\n");

	// Creating Tile
	HitTile_int tile;
	hit_tileDomainShapeAlloc(&tile,sizeof(int),HIT_NONHIERARCHICAL,mysh);
	
	int i=0;
	hit_tileFill(&tile,&i);

        HitPattern pat1=hit_pattern(HIT_PAT_ORDERED);        
        hit_patternAdd(&pat1,hit_comSend(lay, hit_ranks1(1), &tile, HIT_INT));
        hit_patternAdd(&pat1,hit_comSend(lay, hit_ranks1(2), &tile, HIT_INT));
        hit_patternAdd(&pat1,hit_comSend(lay, hit_ranks1(3), &tile, HIT_INT));
	HitPattern pat2=hit_pattern(HIT_PAT_ORDERED);        
        hit_patternAdd(&pat2,hit_comSend(lay, hit_ranks1(4), &tile, HIT_INT));
        hit_patternAdd(&pat2,hit_comSend(lay, hit_ranks1(5), &tile, HIT_INT));

	hit_patternCompose(&pat1,&pat2);
	

	if (hit_Rank == 0) {
	   printf("pattern 1 has %d communications: \n",pat1.numComm);
           HitPatS * elem;
	   for (elem = pat1.first; elem != NULL; elem = elem->next) {
               printf("%d ",elem->commSpec.sendTo);
           }
	   printf("\n");


	   printf("pattern 2 has %d communications: \n",pat2.numComm);
	   for (elem = pat2.first; elem != NULL; elem = elem->next) {
               printf("%d ",elem->commSpec.sendTo);
           }
	   printf("\n");

	} 


	hit_comFinalize(); 
	return 0;
}

