/*
* Test of hit reduce operations
*
* Hit v1.0
* (c) 2010, Arturo González Escribano
*           Javier Fresno Bausela
*           Carlos de Blas Cartón
*/

//does not work: segmentation fault

#include <stdio.h>
#include <stdlib.h>
#include <hitmap.h>
#include <unistd.h>

//#define printfd(...) printf(__VA_ARGS__)
#define printfd(...) if( HIT_TOPOLOGY_INFO.selfRank == 0 ) { printf(__VA_ARGS__); fflush(stdout); }

HitTopology topo;
HitLayout lay;

// All prints
void printfAllSec( const char* format, ... ) {

	int i = 0;

	for(i=0;i<HIT_TOPOLOGY_INFO.numProcs;i++){

		if(HIT_TOPOLOGY_INFO.selfRank == i){

			fprintf( stdout, "%2d: ",i);
			va_list args;
			va_start( args, format );
			vfprintf( stdout, format, args );
			va_end( args );

		}
		hit_comBarrier(lay);
		fflush(stdout);
	}
 }

// Dump a double array
void dumpArray(double * array,int dims, char * name){

	if(dims == 0){
		printfd("%s [X]\n",name);
		return;
	}

	int i;
	printfd("%s [%lf",name,array[0]);
	for(i=1;i<dims;i++) printfd(",%lf",array[i]);
	printfd("]\n");

}

// All dumps a double array
void dumpArrayAllSec(double * array,int dims, char * name){

	char string[1024];
	char * s = string;

	if(dims == 0){
		s += sprintf(s,"%s [X]\n",name);
		return;
	}

	int i;
	s += sprintf(s,"%s [%lf",name,array[0]);
	for(i=1;i<dims;i++) s += sprintf(s,",%lf",array[i]);
	s += sprintf(s,"]\n");

	printfAllSec("%s",string);

}

void dumpShape(HitShape sh){

	int i;
	printfd("[%d:%d:%d",sh.sig[0].begin,sh.sig[0].end,sh.sig[0].stride);
	for(i=1;i<sh.numDims;i++) printfd(",%d:%d:%d",sh.sig[i].begin,sh.sig[i].end,sh.sig[i].stride);
	printfd("] \t cards: [%d",hit_sigCard(sh.sig[0]));
	for(i=1;i<sh.numDims;i++) printfd(",%d",hit_sigCard(sh.sig[i]));
	printfd("]\n");
}

void dumpShapeAllSec(HitShape sh){


	char string[1024];
	char * s = string;

	int i;
	s += sprintf(s,"[%d:%d:%d",sh.sig[0].begin,sh.sig[0].end,sh.sig[0].stride);
	for(i=1;i<sh.numDims;i++) s += sprintf(s,",%d:%d:%d",sh.sig[i].begin,sh.sig[i].end,sh.sig[i].stride);
	s += sprintf(s,"] \t cards: [%d",hit_sigCard(sh.sig[0]));
	for(i=1;i<sh.numDims;i++) s += sprintf(s,",%d",hit_sigCard(sh.sig[i]));
	s += sprintf(s,"]\n");

	printfAllSec("%s",string);

}

int arrayCmp(float* a, float* b, int dims){

	int i;
	for(i=0;i<dims;i++){
		if(a[i] != b[i]) return 0;
	}
	return 1;
}


/*
 
 MPI_COMBINER_NAMEDa named predefined datatype
 MPI_COMBINER_CONTIGUOUS MPI_TYPE_CONTIGUOUS
 MPI_COMBINER_HVECTOR MPI_TYPE_HVECTOR from C or C++
 MPI_COMBINER_STRUCT MPI_TYPE_STRUCT from C or C++
 */

//in this function I try to get all the structure of a type
int printdatatype( MPI_Datatype datatype )  
{ 
    int *array_of_ints; 
    MPI_Aint *array_of_adds; 
    MPI_Datatype *array_of_dtypes; 
    int num_ints, num_adds, num_dtypes, combiner; 
    int i; 
	
    MPI_Type_get_envelope( datatype,  
						  &num_ints, &num_adds, &num_dtypes, &combiner ); 
	
	printf("\n%d-%d-%d\n",num_ints, num_adds, num_dtypes);
	
    switch (combiner) { 
		case MPI_COMBINER_NAMED: 
			printf( "Datatype is named:" ); 
			if      (datatype == MPI_INT)    printf( "MPI_INT\n" ); 
			else if (datatype == MPI_DOUBLE) printf( "MPI_DOUBLE\n" ); 
			else printf("basic type\n"); 
				return 0; 
			break; 
		case MPI_COMBINER_STRUCT: //N+1-N-N N is number of datatypes, stored in ints[0]
			printf( "Datatype is struct containing" ); 
			array_of_ints   = (int *)malloc( num_ints * sizeof(int) ); 
			array_of_adds   =  
			(MPI_Aint *) malloc( num_adds * sizeof(MPI_Aint) ); 
			array_of_dtypes = (MPI_Datatype *) 
            malloc( num_dtypes * sizeof(MPI_Datatype) ); 
			MPI_Type_get_contents( datatype, num_ints, num_adds, num_dtypes, 
								  array_of_ints, array_of_adds, array_of_dtypes ); 
			printf( " %d datatypes:\n", array_of_ints[0] ); 
			for (i=0; i<array_of_ints[0]; i++) { 
				printf( "blocklength %d, displacement %ld, type:\n",  
					   array_of_ints[i+1], array_of_adds[i] ); 
				if (printdatatype( array_of_dtypes[i] )) { 

				} 
			} 
			free( array_of_ints ); 
			free( array_of_adds ); 
			free( array_of_dtypes ); 
			break;
		case MPI_COMBINER_HVECTOR:  //2-1-1   count in ints[0] & blocklength in ints[1] & stride in adds[0] & oldtype in dtypes[0]
			printf( "Datatype is a hvector: " ); 
			array_of_ints   = (int *)malloc( num_ints * sizeof(int) ); 
			array_of_adds   =  
			(MPI_Aint *) malloc( num_adds * sizeof(MPI_Aint) ); 
			array_of_dtypes = (MPI_Datatype *) 
            malloc( num_dtypes * sizeof(MPI_Datatype) ); 
			MPI_Type_get_contents( datatype, num_ints, num_adds, num_dtypes, 
								  array_of_ints, array_of_adds, array_of_dtypes ); 
			printf( " count %d, blocklength %d, stride %ld, type:\n",  
					   array_of_ints[0], array_of_ints[1], array_of_adds[0] ); 
			if (printdatatype( array_of_dtypes[0] )) { 

			} 
			free( array_of_ints ); 
			free( array_of_adds ); 
			free( array_of_dtypes ); 
			break; 
		case MPI_COMBINER_CONTIGUOUS:  //1-0-1   count in ints[0] & oldtype in dtypes[0]
			printf( "Datatype is contiguous:" ); 
			array_of_ints   = (int *)malloc( num_ints * sizeof(int) ); 
			array_of_adds   =  
			(MPI_Aint *) malloc( num_adds * sizeof(MPI_Aint) ); 
			array_of_dtypes = (MPI_Datatype *) 
            malloc( num_dtypes * sizeof(MPI_Datatype) ); 
			MPI_Type_get_contents( datatype, num_ints, num_adds, num_dtypes, 
								  array_of_ints, array_of_adds, array_of_dtypes ); 
			printf( " count %d:\n", array_of_ints[0] ); 
			if (printdatatype( array_of_dtypes[0] )) { 

			} 
			free( array_of_ints ); 
			free( array_of_adds ); 
			free( array_of_dtypes ); 
			break; 
		default: 
			printf( "Unrecognized combiner type\n" ); 
    } 
    return 1; 
} 


//in this function I try to get all the offsets of a type
int offsetDatatype( MPI_Datatype datatype, int offset, size_t tam )  
{ 
    int *array_of_ints; 
    MPI_Aint *array_of_adds; 
    MPI_Datatype *array_of_dtypes; 
    int num_ints, num_adds, num_dtypes, combiner; 
    int i,j,stride,jump,ac,ac2;
	MPI_Aint ext;
	
    MPI_Type_get_envelope( datatype,  
						  &num_ints, &num_adds, &num_dtypes, &combiner ); 
		
    switch (combiner) { 
		case MPI_COMBINER_NAMED: 
			printf( "Bassic offset: %d\n",offset );  
			return 0; 
			break; 
		case MPI_COMBINER_STRUCT: //N+1-N-N N is number of datatypes, stored in ints[0]
			printf( "Struct\n" ); 
			array_of_ints   = (int *)malloc( num_ints * sizeof(int) ); 
			array_of_adds   =  
			(MPI_Aint *) malloc( num_adds * sizeof(MPI_Aint) ); 
			array_of_dtypes = (MPI_Datatype *) 
            malloc( num_dtypes * sizeof(MPI_Datatype) ); 
			MPI_Type_get_contents( datatype, num_ints, num_adds, num_dtypes, 
								  array_of_ints, array_of_adds, array_of_dtypes ); 
			//our blocklengths are always one: internal loop not tested
			for (i=0; i<array_of_ints[0]; i++) { 
				ac=array_of_adds[i]/tam + offset;
				MPI_Type_extent(array_of_dtypes[i],&ext);
				jump = ext / tam;
				for(j=0;j<array_of_ints[i+1];j++) {
					offsetDatatype( array_of_dtypes[i],ac,tam);
					ac+=jump;
				}
			} 
			free( array_of_ints ); 
			free( array_of_adds ); 
			free( array_of_dtypes ); 
			break; 
		case MPI_COMBINER_HVECTOR:  //2-1-1   count in ints[0] & blocklength in ints[1] & stride in adds[0] & oldtype in dtypes[0]
			array_of_ints   = (int *)malloc( num_ints * sizeof(int) ); 
			array_of_adds   =  
			(MPI_Aint *) malloc( num_adds * sizeof(MPI_Aint) ); 
			array_of_dtypes = (MPI_Datatype *) 
            malloc( num_dtypes * sizeof(MPI_Datatype) ); 
			MPI_Type_get_contents( datatype, num_ints, num_adds, num_dtypes, 
								  array_of_ints, array_of_adds, array_of_dtypes ); 
			//our blocklengths are always one: internal loop not tested
			MPI_Type_extent(array_of_dtypes[0],&ext);
			printf( "Hvector %d count %d stride\n", array_of_ints[0], array_of_adds[0]/tam); 
			jump = ext / tam;
			ac=offset, ac2;
			stride = array_of_adds[0] / tam;
			for(i=0;i<array_of_ints[0];i++) {
				ac2=ac;
				for(j=0;j<array_of_ints[1];j++) {
					printf( "Hvector %d starts in %d\n",i+1,ac2 ); 
					offsetDatatype( array_of_dtypes[0],ac2, tam );
					ac2+=jump;
				}
				ac+=stride;
			}
			free( array_of_ints ); 
			free( array_of_adds ); 
			free( array_of_dtypes ); 
			break; 
		case MPI_COMBINER_CONTIGUOUS:  //1-0-1   count in ints[0] & oldtype in dtypes[0]
			printf( "Contiguous\n" ); 
			array_of_ints   = (int *)malloc( num_ints * sizeof(int) ); 
			array_of_adds   =  
			(MPI_Aint *) malloc( num_adds * sizeof(MPI_Aint) ); 
			array_of_dtypes = (MPI_Datatype *) 
            malloc( num_dtypes * sizeof(MPI_Datatype) ); 
			MPI_Type_get_contents( datatype, num_ints, num_adds, num_dtypes, 
								  array_of_ints, array_of_adds, array_of_dtypes );
			MPI_Type_extent(array_of_dtypes[0],&ext);
			jump = ext / tam;
			ac=offset;
			for(i=0;i<array_of_ints[0];i++) {
				printf( "Contiguous %d --> %d\n",i+1,ac ); 
				offsetDatatype( array_of_dtypes[0],ac,tam );
				ac+=jump;
			}
			free( array_of_ints ); 
			free( array_of_adds ); 
			free( array_of_dtypes ); 
			break; 
		default: 
			printf( "Unrecognized combiner type\n" ); 
    } 
    return 1; 
} 

hit_tileNewType(double);

int main(int argc, char *argv[]){

	// Init Hitmap Comm
	hit_comInit(&argc,&argv);
	HitTile_double v1, v2,res,reslocal;
	int i,j,rank;
	double counter=0;	
	
	
	if(hit_NProcs!=4) {
		if( HIT_TOPOLOGY_INFO.selfRank == 0 ) printf("needs 4 processes\n");
		hit_comFinalize();
		return 0;
	}

	// Make a shape
	HitShape sh = hit_shape(2, hit_sig(0,7,1),hit_sig(0,7,1));
	//printfd("Original Shape");
	//dumpShape(sh);

	// Topology
	topo = hit_topology(plug_topArray2D);

	// Layout
	lay = hit_layout(plug_layCyclic,topo,sh,HIT_LAYOUT_NODIM);

	HitShape mysh = hit_layShape(lay);

	//dumpShapeAllSec(mysh);
	//printfd("\n");

	//op construction
	HitOp suma;
	hit_comOp(hit_comOpSumDouble,suma);
	
	
	hit_tileDomainShapeAlloc(&v1, sizeof(double), HIT_NONHIERARCHICAL, sh);
	hit_tileDomainShapeAlloc(&res, sizeof(double), HIT_NONHIERARCHICAL, sh);
	
	
	//if( HIT_TOPOLOGY_INFO.selfRank == 0 ) hit_tileDumpTile(v1);	
	
	//		hit_comBarrier(lay);

	
	
	for (i=0; i<hit_tileDimCard(v1,0); i++)
		for (j=0; j<hit_tileDimCard(v1,1); j++) {
				hit_tileSet(v1,2,counter,i,j);
				//if( HIT_TOPOLOGY_INFO.selfRank == 0 ) printf("v1[%d,%d]=%3.0lf ", i,j,counter);
				counter++;
			}
	//if( HIT_TOPOLOGY_INFO.selfRank == 0 ) printf("\n");

	
	//hit_tileDumpTile(res);

	//		hit_comBarrier(lay);


	hit_tileSelect(&v2, &v1, mysh);
	hit_tileSelect(&reslocal, &res, mysh);
	//hit_tileDumpTile(v2);
	
	//		hit_comBarrier(lay);


	
	/*
	if( HIT_TOPOLOGY_INFO.selfRank == 0 ) printf("\nResult selecting positions: v2\n");

	
	for(rank=0;rank<hit_NProcs;rank++) { 
		if( HIT_TOPOLOGY_INFO.selfRank == rank ) printf("\n rank %d: Result selecting positions: v2\n",rank);
		if( HIT_TOPOLOGY_INFO.selfRank == rank ) {
			for (i=0; i<v2.card[0]; i++) 
					for (j=0; j<v2.card[1]; j++) {
						double * p = &hit_tileGet(v2,2,i,j);
						printf("(%d) v2[%d,%d]=%3.0lf ",hit_Rank,i,j,*p); fflush(stdout);
					}
			printf("\n");
		}
	
	 		hit_comBarrier(lay);

		
	}
	*/
	hit_comBarrier(lay);

	
	if( HIT_TOPOLOGY_INFO.selfRank == 0 )  printf("\nNext step: commit Reduction\n");

	HitRanks r={{0,0,-1,-1}};
	

	//HitCom reduce = hit_comReduce(lay, r, &v2, &res, HIT_DOUBLE, suma);

	
	HitCom reduce = hit_comReduceSelect(lay, r, &v2, hit_shape(2, hit_sig(0,2,1),hit_sig(0,2,1)), HIT_COM_TILECOORDS, &reslocal, hit_shape(2, hit_sig(0,2,1),hit_sig(0,2,1)), HIT_COM_TILECOORDS, HIT_DOUBLE, suma);
	
	if( HIT_TOPOLOGY_INFO.selfRank == 0 ) printf("\nReduction comitted\n");
	

	//		hit_comBarrier(lay);

	
	//test reduce.datatype
	
	MPI_Aint ext,lb,ub;
	int tam;
	
	//		hit_comBarrier(lay);

	
	MPI_Type_extent(reduce.typeSend,&ext);
	MPI_Type_lb(reduce.typeSend,&lb);
	MPI_Type_ub(reduce.typeSend,&ub);
	MPI_Type_size(reduce.typeSend,&tam);

	if( HIT_TOPOLOGY_INFO.selfRank == 0 ) printf("--> reduction_type: extent=%d; lb=%d; ub=%d; size=%d, double=%d, size/double=%d, extent/double=%d\n",ext,lb,ub,tam,sizeof(double),tam/sizeof(double),ext/sizeof(double));
		
	if( HIT_TOPOLOGY_INFO.selfRank == 0 ) offsetDatatype( reduce.typeSend, 0, sizeof(double) ); 

	//if( HIT_TOPOLOGY_INFO.selfRank == 0 ) printf("%d = %d\n",*((MPI_Comm *)lay.pTopology[0].lowLevel),MPI_COMM_WORLD);
	
	//reduce.comm = MPI_COMM_WORLD;

	hit_comDoReduce(&reduce);
		
	/* RESULTS
	 rank 0:
	 (0) v2[0,0]=  0 (0) v2[0,1]=  2 (0) v2[0,2]=  4 (0) v2[1,0]= 16 (0) v2[1,1]= 18 (0) v2[1,2]= 20 (0) v2[2,0]= 32 (0) v2[2,1]= 34 (0) v2[2,2]= 36 
	 
	 rank 1: Result selecting positions: v2
	 (1) v2[0,0]=  8 (1) v2[0,1]= 10 (1) v2[0,2]= 12 (1) v2[1,0]= 24 (1) v2[1,1]= 26 (1) v2[1,2]= 28 (1) v2[2,0]= 40 (1) v2[2,1]= 42 (1) v2[2,2]= 44
	 
	 rank 2: Result selecting positions: v2
	 (2) v2[0,0]=  1 (2) v2[0,1]=  3 (2) v2[0,2]=  5 (2) v2[1,0]= 17 (2) v2[1,1]= 19 (2) v2[1,2]= 21 (2) v2[2,0]= 33 (2) v2[2,1]= 35 (2) v2[2,2]= 37
	 
	 rank 3: Result selecting positions: v2
	 (3) v2[0,0]=  9 (3) v2[0,1]= 11 (3) v2[0,2]= 13 (3) v2[1,0]= 25 (3) v2[1,1]= 27 (3) v2[1,2]= 29 (3) v2[2,0]= 41 (3) v2[2,1]= 43 (3) v2[2,2]= 45
	 
	 add:
	 (+) v2[0,0]=  18 (+) v2[0,1]= 26 (+) v2[0,2]= 34 (+) v2[1,0]= 82 (+) v2[1,1]= 90 (+) v2[1,2]= 98 (+) v2[2,0]= 146 (+) v2[2,1]= 154 (+) v2[2,2]= 162
	 */
	
	rank=0;

		if( HIT_TOPOLOGY_INFO.selfRank == rank ) printf("\n rank %d: Result res\n",rank);
		if( HIT_TOPOLOGY_INFO.selfRank == rank ) {
			for (i=0; i<64; i++) {
					double * p = &reslocal.data[i];
					printf("(%d) res[%d]=%3.0lf ",hit_Rank,i,*p); fflush(stdout);
				}
			printf("\n");
		}
	if( HIT_TOPOLOGY_INFO.selfRank == rank ) {
		for (i=0; i<reslocal.card[0]; i++) 
			for (j=0; j<reslocal.card[1]; j++) {
				double * p = &hit_tileGet(reslocal,2,i,j);
				printf("(%d) res[%d,%d]=%3.0lf ",hit_Rank,i,j,*p); fflush(stdout);
			}
		printf("\n");
	}

	
	
	hit_tileFree(v1);
	
	hit_tileFree(res);
	
	hit_comOpFree(suma);
	hit_comFinalize();
	
	return 0;
}

