/*
* Test of hit_topology
*
* Hit v1.0
* (c) 2010, Arturo González Escribano
*           Javier Fresno Bausela
*           Carlos de Blas Cartón
*/
#include <stdio.h>
#include <stdlib.h>
#include <hitmap.h>


// Dump a integer array
void dumpArray(int * array,int dims, char * name){

	int i;
	printf("%s [%d",name,array[0]);
	for(i=1;i<dims;i++) printf(",%d",array[i]);
	printf("]\n");

}

// Dump a HitTopology
void dumpTopo( HitTopology topo, char * name ) {

	printf("\n Topology: %s\n",name);
	printf("---------------------------\n");
	printf("NumDims:\t%d\n", topo.numDims );

	dumpArray(topo.card,topo.numDims,"Card");
	dumpArray(topo.self.rank,topo.numDims,"Self");

	printf("Lineal Rank:\t%d\n", topo.linearRank );
	printf("Active:\t%d\n", topo.active );
}

// Number of processors sets
#define PROCS_SETS 5
int procs_sets[PROCS_SETS] = {10,16,31,40,65};


// Main function
int main() {

	int  set, nprocs, self;
	/* Select the set of processors */
	for(set=0;set<PROCS_SETS;set++){

		nprocs = procs_sets[set];

		/* Take 3 processors from the set */
		for(self=0;self<=nprocs;self+=((nprocs-1)/2)){

			printf("\n\n Processors Number: %d, Self: %d\n",nprocs,self);
			printf("==================================\n");

			// Set the Physical topology info
			HIT_TOPOLOGY_INFO.numProcs = nprocs;
			HIT_TOPOLOGY_INFO.selfRank = self;

			// Create the virtual topologies
			HitTopology topo;
			topo = hit_topology( plug_topPlain );
			dumpTopo( topo , "plain" );
			topo = hit_topology( plug_topArray2D );
			dumpTopo( topo , "array2D" );
			topo = hit_topology( plug_topArray2DComplete );
			dumpTopo( topo , "array2dComplete" );
			topo = hit_topology( plug_topSquare );
			dumpTopo( topo , "square" );
			topo = hit_topology( plug_topArray3D );
			dumpTopo( topo , "array3D" );
			topo = hit_topology( plug_topArray4D );
			dumpTopo( topo , "array4D" );
		}
	}
	return 0;
}
