/**
 * Test for the bitmap shape.
 * 
 * @file testBitmap.c
 * @date 5-12-2011
 * @author Javier Fresno Bausela
 */

/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2007-2011, Javier Fresno Bausela
 *
 * All Rights Reserved.
 */

#include "hit_test.h"
#include <stdio.h>
#include <stdlib.h>
#include <hitmap.h>
#include <hit_com.h>
#include <unistd.h>
#include <time.h>


// Define the HitTile_int type
hit_tileNewType(int);

// Main function
int main(int argc, char ** argv) {

	// Init hitmap communication module.
	hit_comInit(&argc,&argv);

	// Shape and tile variables.
	HitShape shape_global = HIT_BITMAP_SHAPE_NULL;
	HitShape shape, ext_shape;
	HitTile_int global;
	HitTile_int local;



	if(hit_Rank == 0){

		hit_bShapeAddEdge2(&shape_global,9,5);
		hit_bShapeAddEdge2(&shape_global,3,5);
		hit_bShapeAddEdge2(&shape_global,2,5);
		hit_bShapeAddEdge2(&shape_global,7,9);
		hit_bShapeAddEdge2(&shape_global,2,4);

		dumpBShape(shape_global);
		

		// Tile allocation
		hit_gbTileDomainShapeAlloc(&global, int, shape_global, HIT_VERTICES);
		
		hit_gbTileGraphVertexAt(global,9) = 9;
		hit_gbTileGraphVertexAt(global,3) = 3;
		hit_gbTileGraphVertexAt(global,5) = 5;
		hit_gbTileGraphVertexAt(global,2) = 2;
		hit_gbTileGraphVertexAt(global,7) = 7;
		hit_gbTileGraphVertexAt(global,4) = 4;
		

	}
	
	
	// Create the topology object.
	HitTopology topo = hit_topology(plug_topPlain);

	// Distribute the graph among the processors.
	HitLayout lay = hit_layout(plug_layBitmap,topo,&shape_global);
	


	// Get the shapes for the local and local extended graphs.
	shape = hit_layShape(lay);
	ext_shape = hit_bShapeExpand(shape,shape_global,1);
	
	// Allocate local shape
	hit_gbTileDomainShapeAlloc(&local, int, ext_shape, HIT_VERTICES);
	
	// Clean the values
	int vertex;
	hit_bShapeVertexIterator(vertex,ext_shape) {
		hit_gbTileVertexAt(local,vertex) = (hit_Rank+1)*1000+vertex;
	}
	

	// Print the local shape.
	ALL_SEQUENTIAL_BEGIN(var)
	printf("Layout List:\n");
	dumpListLayout(lay,"bitmap",NULL);
	printf("Proc %d shape:\n",hit_Rank);
	dumpBShape(shape);
	printf("Extended Shape:\n");
	dumpBShape(ext_shape);
	ALL_SEQUENTIAL_END(var)
	

	// Communicate the tile.
	HitCom comA = hit_comSparseScatter(lay, &global, &local, HIT_INT);
	hit_comDo(&comA);


	//HitCom comB = hit_comSparseUpdate(lay, &local, HIT_INT);
	HitCom comB = hit_comSparseUpdateBitmap(lay, &local, HIT_INT);
	hit_comDo(&comB);
	
	
	// Print the values of the vertices.
	ALL_SEQUENTIAL_BEGIN(var2)
	hit_cShapeVertexIterator(vertex,ext_shape) {
		int gvertex = hit_bShapeVertexToGlobal(ext_shape,vertex);
		printf("Proc: %d, item %d, value %d\n", hit_Rank  ,gvertex,	hit_gbTileGraphVertexAt(local,gvertex));
	}
	ALL_SEQUENTIAL_END(var2)
	
	
	// Finalize hitmap communication module.
	hit_comFinalize();

	return 0;
}
