/**
 *
 * 
 * @file testEdges.c
 * @date
 * @author Javier Fresno Bausela
 */

/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2007-2011, Javier Fresno Bausela
 *
 * All Rights Reserved.
 */

#include "hit_test.h"
#include <stdio.h>
#include <stdlib.h>
#include <hitmap.h>
#include <hit_com.h>


// Define the HitTile_int type
hit_tileNewType(int);



void cshape() {

	// Shape and tile variables.
	HitShape shape = HIT_CSR_SHAPE_NULL;
	HitTile_int tile;

	// Shape definition
	hit_cShapeAddEdge2(&shape,9,5);
	hit_cShapeAddEdge2(&shape,3,5);
	hit_cShapeAddEdge2(&shape,2,5);
	hit_cShapeAddEdge2(&shape,7,9);
	hit_cShapeAddEdge2(&shape,2,4);
	
	dumpCShape(shape);
		
	// Tile allocation.
	hit_gcTileDomainShapeAlloc(&tile, int, shape, HIT_EDGES);
	
	
	// Direct access to the edge values.
	int num = 0;
	hit_gcTileGraphEdgeAt(tile, 9,5) = num++;
	hit_gcTileGraphEdgeAt(tile, 3,5) = num++;
	hit_gcTileGraphEdgeAt(tile, 2,5) = num++;
	hit_gcTileGraphEdgeAt(tile, 7,9) = num++;
	hit_gcTileGraphEdgeAt(tile, 2,4) = num++;
	hit_gcTileGraphEdgeAt(tile, 5,9) = num++;
	hit_gcTileGraphEdgeAt(tile, 5,3) = num++;
	hit_gcTileGraphEdgeAt(tile, 5,2) = num++;
	hit_gcTileGraphEdgeAt(tile, 9,7) = num++;
	hit_gcTileGraphEdgeAt(tile, 4,2) = num++;
	
	// Iterators.
	int vertex, edge;
	hit_cShapeVertexIterator(vertex,shape){
	
		hit_cShapeEdgeIterator(edge,shape,vertex){
		
			int dst = hit_cShapeEdgeTarget(shape,edge);
			printf("%d->%d (%d->%d), value %d\n",vertex,dst,
				hit_cShapeVertexToGlobal(shape,vertex),
				hit_cShapeVertexToGlobal(shape,dst),
				hit_gcTileEdgeIteratorAt(tile,vertex,edge));
		
		}
	
	}
	
}

void bshape() {

	// Shape and tile variables.
	HitShape shape = HIT_BITMAP_SHAPE_NULL;
	HitTile_int tile;

	// Shape definition
	hit_bShapeAddEdge2(&shape,9,5);
	hit_bShapeAddEdge2(&shape,3,5);
	hit_bShapeAddEdge2(&shape,2,5);
	hit_bShapeAddEdge2(&shape,7,9);
	hit_bShapeAddEdge2(&shape,2,4);
	
	dumpBShape(shape);
		
	// Tile allocation.
	hit_gbTileDomainShapeAlloc(&tile, int, shape, HIT_EDGES);
	
	
	// Direct access to the edge values.
	int num = 0;
	hit_gbTileGraphEdgeAt(tile, 9,5) = num++;
	hit_gbTileGraphEdgeAt(tile, 3,5) = num++;
	hit_gbTileGraphEdgeAt(tile, 2,5) = num++;
	hit_gbTileGraphEdgeAt(tile, 7,9) = num++;
	hit_gbTileGraphEdgeAt(tile, 2,4) = num++;
	hit_gbTileGraphEdgeAt(tile, 5,9) = num++;
	hit_gbTileGraphEdgeAt(tile, 5,3) = num++;
	hit_gbTileGraphEdgeAt(tile, 5,2) = num++;
	hit_gbTileGraphEdgeAt(tile, 9,7) = num++;
	hit_gbTileGraphEdgeAt(tile, 4,2) = num++;
	
	// Iterators.
	int vertex, edge;
	hit_bShapeVertexIterator(vertex,shape){
	
		hit_bShapeEdgeIterator(edge,shape,vertex){
		
			int dst = hit_bShapeEdgeTarget(shape,edge);
			printf("%d->%d (%d->%d), value %d\n",vertex,dst,
				hit_bShapeVertexToGlobal(shape,vertex),
				hit_bShapeVertexToGlobal(shape,dst),
				hit_gbTileEdgeIteratorAt(tile,vertex,edge));
		
		}
	}
	
}







// Main function
int main() {

	cshape();
	bshape();

	return 0;

}


