/**
 * Test for Layout Metis and Sparse communications.
 * 
 * @file testSparse.c
 * @date 10-11-2011
 * @author Javier Fresno Bausela
 */

/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2007-2011, Javier Fresno Bausela
 *
 * All Rights Reserved.
 */

#include "hit_test.h"
#include <stdio.h>
#include <stdlib.h>
#include <hitmap.h>
#include <hit_com.h>
#include <unistd.h>
#include <time.h>


// Define the HitTile_int type
hit_tileNewType(int);

// Main function
int main(int argc, char ** argv) {

	// Init hitmap communication module.
	hit_comInit(&argc,&argv);

	// Shape and tile variables.
	HitShape shape_global = HIT_CSR_SHAPE_NULL;
	HitShape shape, ext_shape;
	HitTile_int global;
	HitTile_int local;

	// Only root has the global values
	if(hit_Rank == 0){

		hit_cShapeAddEdge2(&shape_global,9,5);
		hit_cShapeAddEdge2(&shape_global,3,5);
		hit_cShapeAddEdge2(&shape_global,2,5);
		hit_cShapeAddEdge2(&shape_global,7,9);
		hit_cShapeAddEdge2(&shape_global,2,4);


		// Root prints the sparse.
		dumpCShape(shape_global); sleep(1);

		// Tile allocation
		hit_gcTileDomainShapeAlloc(&global, int, shape_global, HIT_VERTICES);

		// Tile initial values.
		hit_gcTileGraphVertexAt(global,9) = 9;
		hit_gcTileGraphVertexAt(global,3) = 3;
		hit_gcTileGraphVertexAt(global,5) = 5;
		hit_gcTileGraphVertexAt(global,2) = 2;
		hit_gcTileGraphVertexAt(global,7) = 7;
		hit_gcTileGraphVertexAt(global,4) = 4;
	}

	// Create the topology object.
	HitTopology topo = hit_topology(plug_topPlain);

	// Distribute the graph among the processors.
	HitLayout lay = hit_layout(plug_layMetis,topo,&shape_global);
	
	// Get the shapes for the local and local extended graphs.
	shape = hit_layShape(lay);
	ext_shape = hit_cShapeExpand(shape,shape_global,1);
	
	// Allocate local shape
	hit_gcTileDomainShapeAlloc(&local, int, ext_shape, HIT_VERTICES);
	
	// Clean the values
	int vertex;
	hit_cShapeVertexIterator(vertex,ext_shape) {
		hit_gcTileVertexAt(local,vertex) = (hit_Rank+1)*1000+vertex;
	}
	
	
	// Print the local shape.
	ALL_SEQUENTIAL_BEGIN(var)
	printf("Proc %d\n",hit_Rank);
	printf("         Shape: "); dumpCShape(shape);
	printf("Extended Shape: "); dumpCShape(ext_shape);
	ALL_SEQUENTIAL_END(var)

	// Communicate the tile.
	HitCom comA = hit_comSparseScatter(lay, &global, &local, HIT_INT);
	hit_comDo(&comA);

	HitCom comB = hit_comSparseUpdate(lay, &local, HIT_INT);
	hit_comDo(&comB);


	// Print the values of the vertices.
	ALL_SEQUENTIAL_BEGIN(var2)
	hit_cShapeVertexIterator(vertex,ext_shape) {
		int gvertex = hit_cShapeVertexToGlobal(ext_shape,vertex);
		printf("Proc: %d, item %d, value %d\n", hit_Rank  ,gvertex,	hit_gcTileGraphVertexAt(local,gvertex));
	}
	ALL_SEQUENTIAL_END(var2)

	
	// Free all resources.
	hit_comFree(comA);
	hit_comFree(comB);
	hit_tileFree(local);
	if(hit_Rank == 0) hit_tileFree(global);
	hit_layFree(lay);
	hit_topFree( topo );
	hit_shapeFree(shape_global);

	
	// Finalize hitmap communication module.
	hit_comFinalize();

	return 0;
}
