
/**
 * 
 * @file testSparseExpand.c
 * @date Oct 2011
 * @author Javier Fresno Bausela
 */

/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2007-2011, Arturo Gonzalez Escribano
 *                          Javier Fresno Bausela
 *                          Carlos de Blas Carton
 *                          Yuri Torres de la Sierra
 * All Rights Reserved.
 */

#include "hit_test.h"
#include <stdio.h>
#include <stdlib.h>
#include <hitmap.h>
#include <hit_com.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>



// Define the HitTile_int type
hit_tileNewType(int);

#define BIGINPUT

#ifdef BIGINPUT
#define N_VERTICES	 10000
#define N_EDGES		100000
#define N_SELECTION	 5000
#else
#define N_VERTICES	 5
#define N_EDGES		15
#define N_SELECTION	 1
#endif

#define N_EXPAND		2
#define N_ITER			1
#define CSR_FILE	"testSparseExpandFile.csr"

#define CSR_FILE1	"testSparseExpandFile.out1.csr"
#define CSR_FILE2	"testSparseExpandFile.out2.csr"


/**
 * Get time function
 * @return The current time.
 */
double gettime(){

	struct timeval tv;
	gettimeofday(&tv, (void *) 0);
	return tv.tv_sec + 1.0e-6 * tv.tv_usec;
}

int file_exists(const char * filename){
	FILE * file = fopen(filename, "r");
    if (file){
        fclose(file);
        return 1;
    }
    return 0;
}

// Main function
int main() {

	// Timers
	double ta, tb;
	double t = 0;
	int i;
	
	
	
	// Random sparse graph
	HitShape origin = HIT_CSR_SHAPE_NULL;
	
#ifdef BIGINPUT
	if(!file_exists(CSR_FILE)){
#endif	
	
		for(i=0; i<N_EDGES; i++){
			int a = random() % N_VERTICES;
			int b = random() % N_VERTICES;
	
			hit_cShapeAddEdge(&origin,a,b);
		}
	
#ifdef BIGINPUT	
		hit_fileCSRWrite(CSR_FILE,origin);
		
	} else {
		origin = hit_fileCSRRead(CSR_FILE);
	}
#else

	//Print shape.
	printf("Original shape\n");
	dumpSparseShape(origin);
#endif
	
	
	ta = gettime();
	
	// Random Select
	int * selection = malloc(sizeof(int) * N_SELECTION);
	i = 0;
	while(i<N_SELECTION){
	
		int s = random() % N_VERTICES;
		int j;
		int found = 0;
		
		for(j=0; j<i; j++){
			if(selection[j] == s){
				found = 1;
				break;
			};
		}
		
		if(!found){
			selection[i] = s;
			i++;
		}
	}
	
#ifndef BIGINPUT
	// Print selection
	printf("Selection: ");
	for(i=0; i<N_SELECTION -1; i++){
		printf("%d, ",selection[i]);
	}
	printf("%d\n",selection[N_SELECTION -1]);
#endif
	
	tb = gettime();
	printf("Time c select: %f\n",tb-ta); fflush(stdout);
	
	
	
	
	HitShape shape;
	
	// Make selection
	t = 0;
	for(i=0; i<N_ITER; i++){
		ta = gettime();
		shape = hit_cShapeSelect(origin,N_SELECTION ,selection);
		//printf("\nShape selected\n");
		//dumpSparseShape(shape);
		tb = gettime();
		if(i<N_ITER-1) hit_shapeFree(shape);
		t += tb-ta;
	}
	printf("Time select:   %f\n",t/N_ITER); fflush(stdout);

	
	 HitShape shapeE1;
	
	// Expand 1
	t = 0;
	for(i=0; i<N_ITER; i++){
		ta = gettime();
		shapeE1 = hit_cShapeExpand(shape,origin,N_EXPAND);
		//printf("Shape expand 1\n");
		//dumpSparseShape(shapeE1);
		tb = gettime();
		if(i<N_ITER-1) hit_shapeFree(shapeE1);
		t += tb-ta;
	}
	printf("Time expand 1: %f\n",t/N_ITER); fflush(stdout);
	

#ifndef BIGINPUT
	//printf("\nShapeE1\n");
	dumpSparseShape(shapeE1);

#endif
	
	return 0;
}
