
/**
 * Sparse Selection and Expand operations test.
 * 
 * @file testSparseSelectExpand.c
 * @date Oct 2011
 * @author Javier Fresno Bausela
 */

/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2007-2011, Arturo Gonzalez Escribano
 *                          Javier Fresno Bausela
 *                          Carlos de Blas Carton
 *                          Yuri Torres de la Sierra
 * All Rights Reserved.
 */

#include "hit_test.h"
#include <stdio.h>
#include <stdlib.h>
#include <hitmap.h>
#include <hit_com.h>
#include <unistd.h>
#include <time.h>


// Define the HitTile_int type
hit_tileNewType(int);

// Main function
int main() {

	HitShape origin = HIT_CSR_SHAPE_NULL;
	
	/*
	 * 
	 * 
	 *       9 - 8
	 *       |   |
	 *   1 - 2 - 3 - 4 - 5
	 *    \  | 
	 *       6 - 7 
	 */

	
	hit_cShapeAddEdge(&origin,2,3);
	hit_cShapeAddEdge(&origin,3,4);
	hit_cShapeAddEdge(&origin,4,5);
	hit_cShapeAddEdge(&origin,1,6);
	hit_cShapeAddEdge(&origin,2,6);
	hit_cShapeAddEdge(&origin,6,7);
	hit_cShapeAddEdge(&origin,2,9);
	hit_cShapeAddEdge(&origin,3,8);
	hit_cShapeAddEdge(&origin,8,9);
	hit_cShapeAddEdge(&origin,1,2);
	
	// Original shape.
	printf("Original shape\n\t");
	dumpCShape(origin);
	
	
	// Selection
	int selection[2] = {1,2};
	HitShape select1 = hit_cShapeSelect(origin,2,selection);
	printf("Selection 1 and 2\n\t");
	dumpCShape(select1);
	
	
	// Expand
	HitShape expand = hit_cShapeExpand(select1,origin,1);
	printf("Expand 1\n\t");
	dumpCShape(expand);
	
	
	// Expand2
	HitShape expand2 = hit_cShapeExpand(select1,origin,2);
	printf("Expand 2\n\t");
	dumpCShape(expand2);
	
	// Expand 1+1
	HitShape expand2a = hit_cShapeExpand(select1,origin,1);
	HitShape expand2b = hit_cShapeExpand(expand2a,origin,1);
	printf("Expand 1+1\n\t");
	dumpCShape(expand2b);
	
	
	
	hit_cShapeAddEdge(&origin,3,2);
	hit_cShapeAddEdge(&origin,4,3);
	hit_cShapeAddEdge(&origin,5,4);
	hit_cShapeAddEdge(&origin,6,1);
	hit_cShapeAddEdge(&origin,6,2);
	hit_cShapeAddEdge(&origin,7,6);
	hit_cShapeAddEdge(&origin,9,2);
	hit_cShapeAddEdge(&origin,8,3);
	hit_cShapeAddEdge(&origin,9,8);
	hit_cShapeAddEdge(&origin,2,1);
	
	
	// Original shape.
	printf("Original shape modified\n\t");
	dumpCShape(origin);
	
	// Selection
	printf("Selection of 7,9,5\n\t");
	int selection3[3] = {7,9,5};
	HitShape select3 = hit_cShapeSelect(origin,3,selection3);
	dumpCShape(select3);
	
	// Selection
	printf("Selection of 7,5\n\t");
	int selection4[2] = {7,5};
	HitShape select4 = hit_cShapeSelect(select3,2,selection4);
	dumpCShape(select4);
	
	
	// Expand 4
	HitShape expand4 = hit_cShapeExpand(select4,origin,2);
	printf("Expand 7,5 2\n\t");
	dumpCShape(expand4);
	
	// Expand 10
	HitShape expand10 = hit_cShapeExpand(select4,origin,10);
	printf("Expand 7,5 10 times\n\t");
	dumpCShape(expand10);
	

	return 0;
}
