#include<hitmap.h>



/**********************************************************************************************/
/************************************* AUXILIAR FUNCTIONS *************************************/
/**********************************************************************************************/
hit_tileNewType (double);
hit_tileNewType (float);
hit_tileNewType (int);


#define	HIT_TILE_NULL_STATIC_TRASGO	{ 0, 0, HIT_SHAPE_NULL_STATIC, { 0,0,0,0 }, 0, { 0,0,0,0,0 }, { 0,0,0,0 }, NULL, NULL, NULL, NULL, NULL, HIT_MS_NULL, HIT_NONHIERARCHICAL, NULL, NULL, HIT_MULTILEVEL_UNDEFINED_BEGINS, HIT_MULTILEVEL_UNDEFINED_SIZES, HIT_LAYOUT_NULL }


/* PRINT INFORMATION OF THE SHAPE */
void
dumpShape (HitShape sh)
{
  int i;
  printf ("[%d:%d:%d", hit_shapeSig (sh, 0).begin, hit_shapeSig (sh, 0).end,
	  hit_shapeSig (sh, 0).stride);
  for (i = 1; i < hit_shapeDims (sh); i++)
    printf (",%d:%d:%d", hit_shapeSig (sh, i).begin, hit_shapeSig (sh, i).end,
	    hit_shapeSig (sh, i).stride);
  printf ("] \t cards: [%d", hit_sigCard (hit_shapeSig (sh, 0)));
  for (i = 1; i < hit_shapeDims (sh); i++)
    printf (",%d", hit_sigCard (hit_shapeSig (sh, i)));
  printf ("]\n");
}


/* KNOW MY INDEX ARRAY PARTITON */
void knowMyIndex( HitLayout lay, HitShape in,  HitShape *out){

	// the shape it is the intersection
	HitShape intersect= hit_shapeIntersect(in, hit_layShape(lay));
	*out=intersect;

}

/* Hit PATTERN: SPECIFIC PATTERNS FOR SPECIAL COMPLEX COMMUNICATIONS */

/*
 * @ana Abr 2017
 * hit_patternLayRedistribute: Redistribute a tile allocated with a given layout,
 * 		as indicated by another layout.
 * 		Restrictions:
 * 		1) The two layouts should have been built using the same topology
 * 		2) The tile should have been allocated using the local shape from the first layout
 */
#define HIT_PAT_REDISTRIBUTE_TAG	15001
HitPattern
hit_patternLayRedistributeNew (HitLayout lay1, HitLayout lay2, void *tileP1,
			       void *tileP2, HitShape range1, HitShape range2, HitType baseType)
{
  int i;

  // 1. CHECK THAT THE TOPOLOGY IS THE SAME IN BOTH LAYOUTS 
  if (!hit_ptopoCmp (lay1.topo.pTopology, lay2.topo.pTopology))
    hit_errInternal (__FUNCTION__, "Layouts with different topologies", "",
		     __FILE__, __LINE__);

  // 2. SKIP IF MY PROC IS NOT ACTIVE IN ANY LAYOUT 
  if (!hit_layImActive (lay1) && !hit_layImActive (lay2))
    return HIT_PATTERN_NULL;

  // 3. DECLARE THE NEW COMM. PATTERN 
  HitPattern allToAll = hit_pattern (HIT_PAT_UNORDERED);
  int myRank = lay1.topo.linearRank;
  int numProcs = hit_topCard (lay1.topo);

  // 4. ONLY COMPUTE SENDS IF I'M ACTIVE IN THE FIRST LAYOUT (I HAVE DATA) 
  if (hit_layImActive (lay1))
    {
      HitShape localShp = hit_layShape (lay1);
#ifdef DEBUG_OPE
	printf("\n Local Shape ");
	dumpShape(localShp);
#endif
	localShp=hit_shapeIntersect (localShp, range1);

#ifdef DEBUG_OPE
	printf("\n Local Range Shape ");
	dumpShape(localShp);
#endif

      // FOR ALL PROCESSORS IN THE TOPOLOGY 
      //              (FUNCTIONS INTERNALLY CHECK ACTIVE STATUS IN TOPOLOGY AND LAYOUT) 
      //              REORDER THE ORDER OF EXPLORATION TO CREATE A SKEWED PATTERN AND
      //              AVOID POTENTIAL COMM. BOTTLENECKS      
      for (i = 0; i < numProcs; i++)
	{
	  int foreignId = (myRank + i) % numProcs;

	  HitRanks foreignProc = hit_topRanksInternal (lay2.topo, foreignId);

	  HitShape foreignShp1 = hit_layShapeOther (lay1, foreignProc);
	  foreignShp1=hit_shapeIntersect (foreignShp1, range1);


	  HitShape foreignShp2 = hit_layShapeOther (lay2, foreignProc);
	  foreignShp2=hit_shapeIntersect (foreignShp2, range2);
#ifdef DEBUG_OPE
	  printf("\n Remote Shape %d", i);
	  dumpShape(foreignShp2);
#endif

 	  {	
	  int j; 
	  for(j=0; j< hit_layNumDims(lay2); j++){
		hit_shapeSig(foreignShp2,j).begin = hit_shapeSig(foreignShp2,j).begin - hit_shapeSig(range2,j).begin+ hit_shapeSig(range1,j).begin;
		hit_shapeSig(foreignShp2,j).end = hit_shapeSig(foreignShp2,j).end - hit_shapeSig(range2,j).begin+ hit_shapeSig(range1,j).begin;
		}
	  }
#ifdef DEBUG_OPE
	  printf("\n Remote Shape adapted %d", i);
	  dumpShape(foreignShp2);
#endif

	  // INTERSECTION: NON-EMPTY IMPLIES COMMUNICATION 
	  // FOR COPY SIG. LAYOUTS: SKIP COMM. IF THE FOREIGN PROC. ALREADY HAS THIS SHAPE
	  HitShape overlapShp2 = hit_shapeIntersect (localShp, foreignShp2);
	  HitShape alreadyThere =
	    hit_shapeIntersect (overlapShp2, foreignShp1);

	  if (!hit_shapeCmp (overlapShp2, HIT_SHAPE_NULL)
	      && (i == 0 || hit_shapeCmp (alreadyThere, HIT_SHAPE_NULL)))
	    {

#ifdef DEBUG_OPE
		printf ("\n Soy %d envia a %d", myRank, foreignId);
		dumpShape (overlapShp2);
#endif

	        hit_patternAdd (&allToAll,
			      hit_comSendSelectTag (lay2, foreignProc, tileP1,
						    overlapShp2,
						    HIT_COM_ARRAYCOORDS,
						    baseType,
						    HIT_PAT_REDISTRIBUTE_TAG));
		
	    }
	}
    }

  // 6. ONLY COMPUTE RECEIVES IF I'M ACTIVE IN THE SECOND LAYOUT (I WILL HAVE DATA) 
  if (hit_layImActive (lay2))
    {
      HitShape localShp1 = hit_layShape (lay1);
      localShp1=hit_shapeIntersect (localShp1, range1);

      HitShape localShp2 = hit_layShape (lay2);
      localShp2=hit_shapeIntersect (localShp2, range2);

#ifdef DEBUG_OPE
      printf("\n Local Shape2 ");
      dumpShape(localShp2);
#endif

      // FOR ALL PROCESSORS IN THE TOPOLOGY 
      //              (FUNCTIONS INTERNALLY CHECK ACTIVE STATUS IN TOPOLOGY AND LAYOUT) 
      //              REORDER THE ORDER OF EXPLORATION TO CREATE A SKEWED PATTERN AND
      //              AVOID POTENTIAL COMM. BOTTLENECKS       
      for (i = 0; i < numProcs; i++)
	{
	  int foreignId = (myRank + i) % numProcs;

	  HitRanks foreignProc = hit_topRanksInternal (lay1.topo, foreignId);
	  HitShape foreignShp1 = hit_layShapeOther (lay1, foreignProc);
	  foreignShp1=hit_shapeIntersect (foreignShp1, range1);
 	  {	
	  int j; 
	  for(j=0; j< hit_layNumDims(lay2); j++){
		hit_shapeSig(foreignShp1,j).begin = hit_shapeSig(foreignShp1,j).begin - hit_shapeSig(range1,j).begin+ hit_shapeSig(range2,j).begin;
		hit_shapeSig(foreignShp1,j).end = hit_shapeSig(foreignShp1,j).end - hit_shapeSig(range1,j).begin+ hit_shapeSig(range2,j).begin;
		}
	  }

#ifdef DEBUG_OPE
	  printf("\n Remote Shape adapted %d", i);
	  dumpShape(foreignShp1);
#endif
	  // INTERSECTION: NON-EMPTY IMPLIES COMMUNICATION 
	  // FOR COPY SIG. LAYOUTS: SKIP COMM. IF THE FOREIGN PROC. ALREADY HAS THIS SHAPE 
	  HitShape overlapShp = hit_shapeIntersect (localShp2, foreignShp1);
	  HitShape alreadyHere = hit_shapeIntersect (overlapShp, localShp1);

	  if (!hit_shapeCmp (overlapShp, HIT_SHAPE_NULL)
	      && (i == 0 || hit_shapeCmp (alreadyHere, HIT_SHAPE_NULL)))
	    {

#ifdef DEBUG_OPE
		printf ("\n Soy %d recibo de %d", myRank, foreignId);
		dumpShape (overlapShp);
#endif

	      hit_patternAdd (&allToAll,
			      hit_comRecvSelectTag (lay1, foreignProc, tileP2,
						    overlapShp,
						    HIT_COM_ARRAYCOORDS,
						    baseType,
						    HIT_PAT_REDISTRIBUTE_TAG));

	    }
	}
    }

  // 7. RETURN 
  return allToAll;
}



HitLayout hit_layout_plug_layIndependentLB_tile(HitTopology topo, HitShape elements, double *weights){

	/*double weights1[2];
	weights1[0]=hit_tileElemAt1(weights, 0);
	weights1[1]=hit_tileElemAt1(weights, 1);*/

	HitLayout rsult= hit_layout_plug_layIndependentLB(topo, elements,weights);
	return rsult;
}

HitLayout hit_layout_plug_layIndependentLB_1D(HitTopology topo, HitShape elements, double *weights){

	HitLayout rsult= hit_layout_plug_layIndependentLB(topo, elements,weights);
	return rsult;
}


void distribute_mask(HitTile_double data, HitTile_int mask, int *indices, int num_mask){

  HitTile_double aux;
  hit_tileSelect (&aux, &(data), HIT_SHAPE_WHOLE);
  hit_tileAlloc (&(aux));
  hit_tileUpdateFromAncestor (&aux);
  int i,dim, n_mask,j=0;
  int (*indicesAux)[num_mask+1]=malloc(hit_tileDims(aux) * sizeof *indicesAux);
  int (*indicesAux2)[num_mask+1]=malloc(hit_tileDims(aux) * sizeof *indicesAux2);

  //inicio indices a 0
  for (dim=0; dim < hit_tileDims(aux) ;dim++){
    for(i=0; i< num_mask;i++){
	//indices[dim][i] =0;
	indicesAux[dim][i] =0;
	}
   }

  //Solo para una dimension de momento!!
  //busco el numero de mascara y sumo 1
  indicesAux[0][0] =0;
  for(n_mask=0; n_mask<num_mask; n_mask++){

     hit_tileForDimDomain( data, 0, i ){
	if( hit_tileElemAt( mask, 1,i)==n_mask){
		//hit_tileElemAt( aux, 1,j)=hit_tileElemAt( data, 1,i);
		j++;
	}
	else{
	     // hit_tileElemAt( aux, 1,j)=hit_tileElemAt( tileMat, 1,i);
	}	
     }
     indicesAux[0][n_mask+1] =j;
     //printf(" %d ",indicesAux[0][n_mask+1] );

   }

   for(n_mask=0; n_mask<num_mask+1; n_mask++){
     indicesAux2[0][n_mask]=indicesAux[0][n_mask];
     indices[0+n_mask]=indicesAux[0][n_mask];
   }

   //Redistribuyo Copio de nuevo a data
   for(n_mask=0; n_mask<num_mask; n_mask++){
     hit_tileForDimDomain( data, 0, i ){
	     if( hit_tileElemAt( mask, 1,i)==n_mask){
		hit_tileElemAt( data, 1,indicesAux[0][n_mask])=hit_tileElemAt( aux, 1,i);
		indicesAux[0][n_mask]++;
	      }
     }
    }

	//indices
#ifdef DEBUG_OPE
	printf("\n indiceslocales \n");
	for(n_mask=0; n_mask<=num_mask; n_mask++){
		printf(" %d, ", indices[0+n_mask]);
	}
#endif

  	hit_tileFree ((aux));
	free(indicesAux);
	free(indicesAux2);
}



#define STRINGIFY(a) #a
#define hit_t(A) HIT_## A
#define LOWERTOUPPER(x) ((x - 'a') + 'A')

#define NORMALIZE(TOK) NORMALIZE_ ## TOK

#define NORMALIZE_int HIT_INT
#define NORMALIZE_float HIT_FLOAT
#define NORMALIZE_double HIT_DOUBLE

/**********************************************************************************************/
/**********************************************************************************************/
/**********************************************************************************************/




/*****************************************************/
/**********          Operators        ****************/
/*****************************************************/

/** FIRST OPERATOR: ArrayRemapRange **/

#define ArrayRemapRange(tileP1, tileP2, range1, range2, baseType, flag) \
		{								\
		HitTile_##baseType _TT_padre;					\
		hit_tileDomainShape( &_TT_padre, baseType, range2 );		\
		HitTile_##baseType _TT_new = HIT_TILE_NULL_STATIC_TRASGO;  	\
		HitTile_##baseType _TT_m = HIT_TILE_NULL_STATIC_TRASGO;		\
		ArrayRemapRange_dev(tileP1, tileP2, range1, range2, baseType , flag)		\
		}

#define ArrayRemapRange_dev(tileP1, tileP2, range1, range2, baseType, flag) 		\
		HitShape _TT_shape_new = HIT_SHAPE_NULL;				\
		HitShape _TT_shape_m = HIT_SHAPE_NULL;					\
		if(flag==1) {								\
		   hit_tileFree(tileP2);						\
		   if(hit_shapeSigDims(range2)==1) {					\
			HitTopology _TT_topo = hit_topology( plug_topPlain );		\
			HitLayout layAux = hit_layout ( plug_layBlocks,			\
					         	_TT_topo, 			\
							range2);			\
			hit_tileSelect( &tileP2, &_TT_padre, hit_layShape (layAux) );		\
			hit_tileAlloc( &tileP2 );						\
  			tileP2.Layout= layAux;							\
		  	hit_layFree(layAux);							\
		  	hit_topFree(_TT_topo);							\
		   }										\
		   if(hit_shapeSigDims(range2)==2) {						\
			HitTopology _TT_topo = hit_topology( plug_topArray2DComplete );		\
			HitLayout layAux = hit_layout ( plug_layBlocks,				\
					         _TT_topo, 	\
						hit_tileShape( _TT_padre ));			\
			hit_tileSelect( &tileP2, &_TT_padre, hit_layShape (layAux) );		\
			hit_tileAlloc( &tileP2 );						\
  			tileP2.Layout= layAux;							\
		  	hit_layFree(layAux);							\
		  	hit_topFree(_TT_topo);							\
		   }								\
										\
		}								\
										\
		int i;								\
		for(i=0;i< hit_shapeSigDims(range1);i++) {			\
			int diff = hit_tileDimCard(tileP2,i) - hit_shapeSigCard(hit_layShape(tileP2.Layout),i);		\
			hit_shapeDimExpand (range1, i, HIT_SHAPE_END, diff);						\
			hit_shapeDimExpand (range2, i, HIT_SHAPE_END, diff);						\
		}								\
		knowMyIndex (tileP2.Layout, range2, &_TT_shape_new);		\
		knowMyIndex (tileP1.Layout, range1, &_TT_shape_m);		\
										\
		hit_tileSelectArrayCoords (&_TT_new, &tileP2, _TT_shape_new);	\
		hit_tileSelectArrayCoords (&_TT_m, &tileP1, _TT_shape_m);	\
										\
		HitPattern redis= hit_pattern (HIT_PAT_UNORDERED);		\
		redis = hit_patternLayRedistributeNew				\
				      ((tileP1).Layout, (tileP2).Layout,   	\
					&_TT_m, &_TT_new, range1, range2, NORMALIZE(baseType) );			\
		hit_patternDoOnce (redis);										\




/** THIRD OPERATOR: ArrayDivide **/

#define ArrayDivide_dev(grouping, num_maxMask_data_3, weights, data, new_data, mask, _TT_indiceGlobal_data_3, baseType )	\
   { 														\
      int _TT_i=0;												\
	HitShape _TT_Elementsgrouping = hit_shape (1, hit_sigStd (num_maxMask_data_3));				\
      	HitTopology _TT_activesTopogrouping =									\
			hit_layActivesTopology ((*data).Layout);						\
      	grouping =												\
		hit_layout (plug_layIndependentLB_tile, _TT_activesTopogrouping,				\
		  	  _TT_Elementsgrouping, weights);							\
      	hit_topFree (_TT_activesTopogrouping);									\
														\
      /* Indexes */												\
      int (*_TT_indice_data_3) =										\
	(int *) malloc (hit_tileDims (*data) * (num_maxMask_data_3 + 1) * sizeof (int));			\
														\
      /* New Tiles */												\
      HitTile_##baseType (*_TT_tiledata_3) =									\
	malloc (num_maxMask_data_3 * sizeof (HitTile_##baseType));						\
														\
														\
      _TT_indice_data_3[0] = 0;											\
														\
      /* REDISTRIBUTE DATA */											\
      distribute_mask ((*data), (mask), _TT_indice_data_3, num_maxMask_data_3);					\
														\
	/* TODO una reducion */											\
      _TT_indiceGlobal_data_3[0] = 0;										\
      for (_TT_i = 0; _TT_i < num_maxMask_data_3 + 1; _TT_i++){							\
														\
	  HitTile _TT_indice =											\
	    hit_tileSingle (_TT_indice_data_3[0 + _TT_i], int);							\
	  _TT_indiceGlobal_data_3[0 + _TT_i] = 0;								\
														\
	  HitTile _TT_totalindice =										\
	    hit_tileSingle (_TT_indiceGlobal_data_3[0 + _TT_i], int);						\
														\
	  hit_comDoOnce (hit_comReduce ((*data).Layout,								\
					HIT_RANKS_NULL,								\
					&_TT_indice,								\
					&_TT_totalindice,							\
					HIT_INT, HIT_OP_SUM_INT));						\
	}													\
														\
      for (_TT_i = 0; _TT_i < num_maxMask_data_3; _TT_i++){							\
														\
	  HitShape _TT_Shape = hit_shapeSubset (hit_layFullShape ((*data).Layout),				\
						hit_shape (1, 							\
							   hit_sig						\
							   (_TT_indiceGlobal_data_3[0 + _TT_i],			\
							    _TT_indiceGlobal_data_3[0 + _TT_i + 1] -1,		\
							    1)));						\
	  HitTopology _TT_topo = hit_layGroupTopo (grouping,							\
							  hit_lgr_elementGroup (grouping, _TT_i));													\
	  HitLayout lay_aux =											\
	    hit_layout (plug_layBlocks, _TT_topo, _TT_Shape);	\
														\
	  hit_tileDomainShapeAlloc (&new_data[_TT_i], double, hit_layShape (lay_aux));				\
														\
	  new_data[_TT_i].Layout = lay_aux;									\
														\
	  hit_tileSelect (&_TT_tiledata_3[_TT_i], &(*data), 							\
				hit_shape (1,									\
					   hit_sig (_TT_indice_data_3[0 + _TT_i],				\
						    _TT_indice_data_3[0 + _TT_i + 1] - 1,			\
						    1)));							\
														\
	  hit_patternDoOnce (hit_patternRedistribute ((*data).Layout,						\
						      &_TT_tiledata_3[_TT_i],					\
						      &new_data[_TT_i], NORMALIZE(baseType),			\
						      1));							\
	  hit_layFree (lay_aux);										\
	}													\
														\
														\
        hit_tileFree(*data);											\
   	free(_TT_tiledata_3);											\
	free(_TT_indice_data_3);										\
    }														\




/** FOUTH OPERATOR: ArrayMerge **/
#define ArrayMerge_dev(grouping,num_maxMask_data_3, data, old_data, _TT_indiceGlobal_data_3, baseType) 		\
{														\
        int _TT_i=0, Flag=0;											\
        for (_TT_i = 0; _TT_i < num_maxMask_data_3; _TT_i++){							\
		if (hit_layHasElem (grouping, _TT_i)) {								\
			Flag=Flag+1;										\
		}												\
	}													\
	if(Flag == num_maxMask_data_3)										\
	{													\
	  hit_tileAlloc( & (*data )); 										\
	  for (_TT_i = 0; _TT_i < num_maxMask_data_3; _TT_i++)							\
	    {													\
	  													\
	      HitShape Shape1 = hit_shapeSubset (hit_layFullShape (((*data)).					\
								   Layout),					\
						 hit_shape (1,							\
							    hit_sig						\
							    (_TT_indiceGlobal_data_3[0 + _TT_i],		\
							     _TT_indiceGlobal_data_3[0 + _TT_i + 1] -1, 	\
							     1)));						\
	      HitTile_##baseType tmp1 = HIT_TILE_NULL_STATIC_TRASGO;						\
	      hit_tileSelectArrayCoords (&tmp1, &((*data)),							\
					 hit_shapeIntersect (hit_tileShape (((*data))),Shape1));		\
														\
	      hit_patternDoOnce (hit_patternRedistribute (((*data)).Layout,					\
							  &old_data[_TT_i], &tmp1,				\
							  NORMALIZE(baseType), 0));				\
	  	hit_tileFree(old_data[_TT_i]);									\
	    }													\
	}													\
	else {													\
     	 for (_TT_i = 0; _TT_i < num_maxMask_data_3; _TT_i++){							\
		if (hit_layHasElem (grouping, _TT_i)) {								\
		  hit_tileAlloc( & (*data )); 									\
		  HitTile_##baseType tmp1 = HIT_TILE_NULL_STATIC_TRASGO; 					\
		  hit_tileSelectArrayCoords (&tmp1, &((*data)), HIT_SHAPE_WHOLE);				\
														\
		  hit_patternDoOnce (hit_patternRedistribute (((*data)).Layout,					\
							      &old_data[_TT_i],					\
							      &tmp1, NORMALIZE(baseType), 1));			\
		  												\
		  hit_tileFree(old_data[_TT_i]);								\
		}												\
	}}													\
														\
	free(old_data);												\
	free(_TT_indiceGlobal_data_3);										\
}








