
#include <assert.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


double objs[] = {
  0. ,0., -100.5, 10000., 0., 0., 0., 0.25,
  0.272166, 0.272166, 0.544331, .027777,
  0.643951, 0.172546, 0., .027777,
  0.172546, 0.643951, 0., .027777,
 -0.371785, 0.099620, 0.544331, .027777,
 -0.471405, 0.471405, 0., .027777,
 -0.643951,-0.172546, 0., .027777,
  0.099620,-0.371785, 0.544331, .027777,
 -0.172546,-0.643951, 0., .027777,
  0.471405,-0.471405, 0., .027777,
  4., 3., 2., 1., -4., 4., -3., 1., 5.
};

double *intersect(double x, double y, double z, double dx, double dy, double dz, double *maxp)
{
  int i;
  double *o = objs, *oo = 0;
  double max = *maxp;
  double xx, yy, zz, b, t;

  for (i = 0; i < 11; i++)
    {
      xx = *o++ - x; yy = *o++ - y; zz = *o++ - z;
      b = xx * dx + yy * dy + zz * dz;
      if ((t = b * b - xx * xx - yy * yy - zz * zz + *o++) < 0 || (t = b-sqrt(t)) < 1e-6 || t > max)
	continue;
      oo = o - 4;
      max = t;
    }
  *maxp = max;
  return oo;
}

double shade(double x, double y, double z, double dx, double dy, double dz, int de)
{
  double max = 1e6, c = 0, r, k, *o;
  double nx, ny, nz, ldx, ldy, ldz, rdx, rdy, rdz;
  int i;

  if (!(o = intersect(x, y, z, dx, dy, dz, &max)))
    return 0;
  x += max * dx; y += max * dy; z += max * dz;
  nx = x - *o++; ny = y - *o++; nz = z - *o++;
  r = sqrt(nx * nx + ny * ny + nz * nz);
  nx /= r; ny /= r; nz /= r;
  k = nx * dx + ny * dy + nz * dz;
  rdx = dx - 2 * k * nx; rdy = dy - 2 * k * ny; rdz = dz - 2 * k * nz;
  o = objs + 44;
  for (i = 0; i < 3; i++)
    {
      ldx = *o++ - x; ldy = *o++ - y; ldz = *o++ - z;
      r = sqrt(ldx * ldx + ldy * ldy + ldz * ldz);
      ldx /= r; ldy /= r; ldz /= r;
      if (intersect(x, y, z, ldx, ldy, ldz, &r))
	continue;
      if ((r = ldx * nx + ldy * ny + ldz * nz) < 0)
	continue;
      c += r;
      if ((r = rdx * ldx + rdy * ldy + rdz * ldz) > 0)
	c += 2 * pow(r, 15.);
    }
  if (de < 10)
    c += .5 * shade(x, y, z, rdx, rdy, rdz, de + 1);
  return c;
}


void calc_tile(char *picture, int size)
{
    double dx, dy, dz, c, r;
    int x, i,y;
	
    for (y = 0; y < size; y++)
    {
	    i = y * size;
	    for (x = 0; x < size; x++)
	    {
		double xx = x / (float)(size-1);
		double yy = 1. - y / (float)(size-1);
		dx = -0.847569 - xx * 1.30741 - yy * 1.19745;
		dy = -1.98535  + xx * 2.11197 - yy * 0.741279;
		dz = -2.72303                 + yy * 2.04606;
		r = sqrt(dx * dx + dy * dy + dz * dz);
		c = 100 * shade(2.1, 1.3, 1.7, dx / r, dy / r, dz / r, 0);
		if (c < 0)
		    c = 0;
		else if (c > 255)
		    c = 255;
		picture[i++] = c;
	    }
    }

}

int main(int argc, char** argv)
{
    FILE *fd;
    char *picture, *whole_picture;
    int error, rank, num_procs, picture_size, size;
  //  error = MPI_Init(&argc, &argv); assert(error == 0);
   // error = MPI_Comm_rank(MPI_COMM_WORLD, &rank); assert(error == 0);
   // error = MPI_Comm_size(MPI_COMM_WORLD, &num_procs); assert(error == 0);

    size     = 1000;
    picture_size = size * size;

    if((picture=(char*)malloc(picture_size*sizeof(char)))==NULL)
    {
        fprintf(stderr,"Could not allocate picture memory!\n");
        exit(1);
    }
    if((whole_picture=(char*)malloc(picture_size*sizeof(char)))==NULL)
    {
        fprintf(stderr,"Could not allocate picture memory!\n");
        exit(1);
    }

    calc_tile(picture, size );

 //   MPI_Reduce(picture, whole_picture, picture_size, MPI_BYTE, MPI_BOR, 0, MPI_COMM_WORLD);

    if(rank == 0)
    {
        fd = fopen("par.pnm", "w");
        fprintf(fd,"P5\n%d %d\n255\n", size, size );
        fwrite(whole_picture, picture_size, 1, fd);
    }

    free(picture);
    //MPI_Finalize();
}
